/*

  Antlr lexer and parser for VHDL-AMS (IEEE Std 1076.1-1999).
  
  The grammar fits to
  http://www.vhdl-ams.org/language/language.html 
  with some ideas from
  vhdl.jjt [VHDL-93, VHDL-AMS grammar for JavaCC (c) 1997 Christoph Grimm].
   
  First Author: E.Fehlauer, F.Rogin 
                Copyright (c) 2004 Fraunhofer-Gesellschaft.
                All rights reserved.
                
  This code was update in march 2008 by Juraj Petras and in september 2010
  by Dominik Macko, to make it working against C#. Also some bugs and 
  ambiguities were fixes. There was also added code for generating TreeParser
  (imaginary tokens, ...).

  Permission to use, copy, modify, and distribute this 
  software and its documentation for educational, research 
  and non-profit purposes, without fee, and without a 
  written agreement is hereby granted, provided that the 
  above copyright notice, this paragraph and the following 
  three paragraphs appear in all copies. 
  
  Permission to incorporate this software into commercial 
  products may be obtained by contacting
  
    Fraunhofer Institute for Integrated Circuits 
    Branch Lab Design Automation
    Zeunerstrae 38
    D-01069 Dresden
    Email: erhard.fehlauer@eas.iis.fhg.de
  
  This software program and documentation are copyrighted 
  by FhG IIS/EAS. The software program and documentation 
  are supplied "as is", without any accompanying services. 
  FhG IIS/EAS does not warrant that the operation of the 
  program will be uninterrupted or error-free. The end-user 
  understands that the program was developed for research 
  purposes and is advised not to rely exclusively on the 
  program for any reason. 
  
  IN NO EVENT SHALL FHG IIS/EAS BE LIABLE TO ANY PARTY FOR 
  DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL 
  DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE 
  OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF 
  FHG IIS/EAS HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH 
  DAMAGE. FHG IIS/EAS SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE 
  SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND 
  FHG IIS/EAS HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, 
  SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. 

*/

grammar vamsCSharp3;

options {
  language='CSharp2';
  backtrack=true;
  output = AST;
}


//--------------------------------------------------------------------------
// The start symbol is `design_file'.

tokens {
  ARCHITECTURE='architecture';
  ACCESS='access';
  ACROSS='across';
  OPEN='open';
  ALIAS='alias';
  IS='is';
  RETURN='return';
  NEW='new';
  OF='of';
  BEGIN='begin';
  END='end';
  BLOCK='block';
  POSTPONED='postponed';
  PROCESS='process';
  ASSERT='assert';
  ARRAY='array';
  REPORT='report';
  SEVERITY='severity';
  ATTRIBUTE='attribute';
  RANGETOK='range';
  USE='use';
  FOR='for';
  SHARED='shared';
  VARIABLE='variable';
  GROUP='group';
  QUANTITY='quantity';
  BREAK='break';
  WHEN='when';
  CASE='case';
  OTHERS='others';
  COMPONENT='component';
  UNTIL='until';
  ELSE='else';
  CONFIGURATION='configuration';
  CONSTANT='constant';
  TRANSPORT='transport';
  REJECT='reject';
  INERTIAL='inertial';
  TO='to';
  DOWNTO='downto';
  DISCONNECT='disconnect';
  AFTER='after';
  ENTITY='entity';
  PROCEDURE='procedure';
  FUNCTION='function';
  PACKAGE='package';
  TYPE='type';
  SUBTYPE='subtype';
  SIGNAL='signal';
  LABEL='label';
  LITERAL='literal';
  UNITS='units';
  FILE='file';
  NATURE='nature';
  SUBNATURE='subnature';
  TERMINAL='terminal';
  ALL='all';
  EXIT='exit';
  ABS='abs';
  NOT='not';
  GENERATE='generate';
  IF='if';
  GENERIC='generic';
  MAP='map';
  THEN='then';
  ELSIF='elsif';
  IN='in';
  OUT='out';
  BUS='bus';
  WHILE='while';
  LIBRARY='library';
  NULLTOK='null';
  AND='and';
  OR='or';
  NAND='nand';
  NOR='nor';
  XOR='xor';
  XNOR='xnor';
  LOOP='loop';
  INOUT='inout';
  BUFFER='buffer';
  LINKAGE='linkage';
  MOD='mod';
  REM='rem';
  NEXT='next';
  GUARDED='guarded';
  BODY='body';
  PORT='port';
  RECORD='record';
  THROUGH='through';
  WITH='with';
  SELECT='select';
  ON='on';
  WAIT='wait';
  SLL='sll';
  SRL='srl';
  SLA='sla';
  SRA='sra';
  ROL='rol';
  ROR='ror';
  REGISTER='register';
  PROCEDURAL='procedural';
  SPECTRUM='spectrum';
  NOISE='noise';
  LIMIT='limit';
  TOLERANCE='tolerance';
  PURE='pure'; 
  IMPURE='impure';
  UNAFFECTED='unaffected';
  REFERENCE='reference'; 
  //IMAGINARY TOKENS FOR TREES 
  T_design_file;
  T_design_unit;
  T_context_clause;
  T_library_unit;
  T_primary_unit;
  T_secondary_unit;
  
  T_entity_declaration;
  T_entity_header; 
  T_port_clause;
  T_port_list;
  T_interface_list;
  T_interface_element;
  T_interface_declaration;
  T_interface_signal_declaration;
  T_identifier_list;
  T_identifier;
  T_subtype_indication; 
  T_constraint;
  T_range_constraint;
  T_index_constraint; 
  
  T_architecture_body;
  T_name;
  T_entityname;
  T_architecture_declarative_part;
  T_block_declarative_item;
  T_signal_declaration;
  T_component_declaration;
  T_configuration_specification;
  T_component_specification;
  T_instantiation_list;
  T_binding_indication;
  T_entity_aspect;
  T_generic_map_aspect;
  T_association_list;
  T_association_element;
  T_formal_part;
  T_actual_part;
  T_port_map_aspect;
  T_architecture_statement_part;
  T_architecture_statement;
  T_component_instantiation_statement;
  T_label_colon;
  T_instantiated_unit;
  T_generation_scheme;
  T_generate_statement;
  T_parameter_specification;
  T_discrete_range;
  T_range;
  T_condition;
  T_simple_expression; 
  T_generic_clause;
  T_mode;
  T_conditional_signal_assignment; 
  T_conditional_waveforms;
  T_waveform;
  T_waveform_element;
  
  T_access_type_definition;
  T_across_aspect;
  T_actual_designator;
  T_actual_parameter_part;
  T_aggregate;
  T_alias_declaration;
  T_alias_indication;
  T_allocator;
  T_array_nature_definition;
  T_array_type_definition;
  T_assertion;
  T_assertion_statement;
  T_attribute_declaration;
  T_attribute_specification;
  T_base_unit_declaration;
  T_block_configuration;
  T_block_declarative_part;
  T_block_header;
  T_block_specification;
  T_block_statement;
  T_block_statement_part;
  T_branch_quantity_declaration;
  T_break_element;
  T_break_list;
  T_break_selector_clause;
  T_break_statement;
  T_case_statement;
  T_case_statement_alternative;
  T_choice;
  T_choices;
  T_component_configuration;
  T_composite_nature_definition;
  T_composite_type_definition;
  T_concurrent_assertion_statement;
  T_concurrent_break_statement;
  T_concurrent_procedure_call_statement;
  T_concurrent_signal_assignment_statement;
  T_condition_clause;
  T_conditional_waveforms_bi;
  T_configuration_declaration;
  T_configuration_declarative_item;
  T_configuration_declarative_part;
  T_configuration_item;
  T_constant_declaration;
  T_constrained_array_definition;
  T_constrained_nature_definition;
  T_context_item;
  T_delay_mechanism;
  T_designator;
  T_disconnection_specification;
  T_element_association;
  T_element_declaration;
  T_element_subnature_definition;
  T_element_subtype_definition;
  T_entity_class_entry;
  T_entity_class_entry_list;
  T_entity_declarative_item;
  T_entity_declarative_part;
  T_entity_designator;
  T_entity_name_list;
  T_entity_specification;
  T_entity_statement;
  T_entity_statement_part;
  T_enumeration_type_definition;
  T_exit_statement;
  T_expression;
  T_helpexpression;
  T_factor;
  T_file_declaration;
  T_file_logical_name;
  T_file_open_information;
  T_file_type_definition;
  T_formal_parameter_list;
  T_free_quantity_declaration;
  T_function_call;
  T_helpgenerate_statement;
  T_generic_list;
  T_group_constituent;
  T_group_constituent_list;
  T_group_declaration;
  T_group_template_declaration;
  T_guarded_signal_specification;
  T_if_statement;
  T_elsif_statement;
  T_else_statement;
  T_index_specification;
  T_index_subtype_definition;
  T_interface_constant_declaration;
  T_interface_file_declaration;
  T_interface_quantity_declaration;
  T_interface_terminal_declaration;
  T_interface_variable_declaration;
  T_iteration_scheme;
  T_library_clause;
  T_logical_name_list;
  T_loop_statement;
  T_helpname;
  T_nature_declaration;
  T_nature_definition;
  T_nature_element_declaration;
  T_next_statement;
  T_numeric_literal;
  T_object_declaration;
  T_opts;
  T_package_body;
  T_package_body_declarative_item;
  T_package_body_declarative_part;
  T_package_declaration;
  T_package_declarative_item;
  T_package_declarative_part;
  T_physical_literal;
  T_physical_type_definition;
  T_primary;
  T_procedural_declarative_item;
  T_procedural_declarative_part;
  T_procedural_statement_part;
  T_procedure_call;
  T_procedure_call_statement;
  T_process_declarative_item;
  T_process_declarative_part;
  T_process_statement;
  T_process_statement_part;
  T_qualified_expression;
  T_quantity_declaration;
  T_quantity_list;
  T_quantity_specification;
  T_record_nature_definition;
  T_record_type_definition;
  T_relation;
  T_report_statement;
  T_return_statement;
  T_scalar_nature_definition;
  T_scalar_type_definition;
  T_secondary_unit_declaration;
  T_selected_signal_assignment;
  T_selected_waveforms;
  T_sensitivity_clause;
  T_sensitivity_list;
  T_sequence_of_statements;
  T_sequential_statement;
  T_shift_expression;
  T_signal_assignment_statement;
  T_signal_list;
  T_signature;
  T_helpsimple_expression;
  T_simple_simultaneous_statement;
  T_simultaneous_alternative;
  T_simultaneous_case_statement;
  T_simultaneous_if_statement;
  T_simultaneous_procedural_statement;
  T_simultaneous_statement;
  T_simultaneous_statement_part;
  T_source_aspect;
  T_source_quantity_declaration;
  T_step_limit_specification;
  T_subnature_declaration;
  T_subnature_indication;
  T_subprogram_body;
  T_subprogram_declaration;
  T_subprogram_declarative_item;
  T_subprogram_declarative_part;
  T_subprogram_specification;
  T_subprogram_statement_part;
  T_subtype_declaration;
  T_target;
  T_term;
  T_terminal_aspect;
  T_terminal_declaration;
  T_through_aspect;
  T_timeout_clause;
  T_tolerance_aspect;
  T_type_declaration;
  T_type_definition;
  T_unconstrained_array_definition;
  T_unconstrained_nature_definition;
  T_use_clause;
  T_variable_assignment_statement;
  T_variable_declaration;
  T_wait_statement;
  
  T_operator;
}

@parser::namespace { Antlr.Vhdl }

@lexer::namespace { Antlr.Vhdl } 

abstract_literal
  : DECIMAL_LITERAL
  | BASED_LITERAL
  ;

access_type_definition
  : ACCESS subtype_indication -> ^(T_access_type_definition subtype_indication)
  ;

across_aspect
  : identifier_list ( tolerance_aspect )? ( VARASGN expression )? ACROSS -> ^(T_across_aspect identifier_list tolerance_aspect? expression?)
  ;

actual_designator
  : expression -> ^(T_actual_designator expression)
  | OPEN -> ^(T_actual_designator OPEN)
  ;

actual_parameter_part
  : association_list -> ^(T_actual_parameter_part association_list)
  ;

actual_part
  : (name LPAREN)=> name LPAREN actual_designator RPAREN	-> ^(T_actual_part name actual_designator)
  | actual_designator -> ^(T_actual_part actual_designator)
  ;

adding_operator
  : PLUS -> ^(T_operator PLUS)
  | MINUS -> ^(T_operator MINUS)
  | AMPERSAND -> ^(T_operator AMPERSAND)
  ;

aggregate
  : LPAREN element_association ( COMMA element_association )* RPAREN -> ^(T_aggregate element_association+)
  ;

alias_declaration
  : ALIAS alias_designator ( COLON alias_indication )? IS
    name ( signature )? SEMI -> ^(T_alias_declaration alias_designator alias_indication? name signature?)
  ;

alias_designator
  : identifier
  | CHARACTER_LITERAL
  | STRING_LITERAL
  ;

alias_indication
  : (subnature_indication)=> subnature_indication -> ^(T_alias_indication subnature_indication)
  | subtype_indication -> ^(T_alias_indication subtype_indication)
  ;

allocator
  : NEW ( (name APOSTROPHE)=> qualified_expression | subtype_indication ) -> ^(T_allocator qualified_expression? subtype_indication?)
  ;

architecture_body
  : ARCHITECTURE identifier OF entityname IS
    architecture_declarative_part
    BEGIN
    architecture_statement_part
    END ( ARCHITECTURE )? ( identifier )? SEMI -> ^(T_architecture_body identifier entityname architecture_declarative_part architecture_statement_part)
  ;

architecture_declarative_part
  : ( block_declarative_item )* -> ^(T_architecture_declarative_part block_declarative_item*)
  ;

architecture_statement
  : block_statement -> ^(T_architecture_statement block_statement)
  | (( label_colon )? ( POSTPONED )? PROCESS)=>
    process_statement -> ^(T_architecture_statement process_statement)
  | (( label_colon )? ( POSTPONED )? procedure_call SEMI)=>
    concurrent_procedure_call_statement -> ^(T_architecture_statement concurrent_procedure_call_statement)
  | (( label_colon )? ( POSTPONED )? ASSERT)=>
    concurrent_assertion_statement -> ^(T_architecture_statement concurrent_assertion_statement)
  | (( label_colon )? ( POSTPONED )?
      ( conditional_signal_assignment | selected_signal_assignment ))=>
    concurrent_signal_assignment_statement -> ^(T_architecture_statement concurrent_signal_assignment_statement)
  | (label_colon instantiated_unit)=>
    component_instantiation_statement -> ^(T_architecture_statement component_instantiation_statement)
  | (label_colon generation_scheme GENERATE)=>
    generate_statement -> ^(T_architecture_statement generate_statement)
  | (( label_colon )? BREAK ( break_list )? ( sensitivity_clause )?)=>
    concurrent_break_statement -> ^(T_architecture_statement concurrent_break_statement)
  | simultaneous_statement -> ^(T_architecture_statement simultaneous_statement)
  ;

architecture_statement_part
  : ( architecture_statement )* -> ^(T_architecture_statement_part architecture_statement*)
  ; 

array_nature_definition
  : (ARRAY LPAREN index_subtype_definition)=> unconstrained_nature_definition -> ^(T_array_nature_definition unconstrained_nature_definition)
  | constrained_nature_definition -> ^(T_array_nature_definition constrained_nature_definition)
  ;

array_type_definition
  : (unconstrained_array_definition)=> unconstrained_array_definition -> ^(T_array_type_definition unconstrained_array_definition)
  | constrained_array_definition -> ^(T_array_type_definition constrained_array_definition)
  ;

assertion
  : ASSERT condition ( REPORT expression )? ( SEVERITY expression )? -> ^(T_assertion condition expression*)
  ;

assertion_statement
  : ( label_colon )? assertion SEMI -> ^(T_assertion_statement label_colon? assertion)
  ;

association_element
  : ( (formal_part ARROW)=> formal_part ARROW )? actual_part -> ^(T_association_element formal_part? actual_part)
  ;

association_list
  : association_element ( COMMA association_element )* -> ^(T_association_list association_element+)
  ;

attribute_declaration
  : ATTRIBUTE label_colon name SEMI -> ^(T_attribute_declaration label_colon name)
  ;

// Need to add several tokens here, for they are both, VHDLAMS reserved words
// and attribute names.
// (25.2.2004, e.f.)
attribute_designator
  : identifier
  | RANGETOK
  | ACROSS
  | THROUGH
  | REFERENCE
  | TOLERANCE
  ;

attribute_specification
  : ATTRIBUTE attribute_designator OF entity_specification IS expression SEMI -> ^(T_attribute_specification attribute_designator entity_specification expression) 
  ;

base_unit_declaration
  : identifier SEMI -> ^(T_base_unit_declaration identifier)
  ;

//TODO: generic_map_aspect, port_map_aspect - implementovat
binding_indication
  : ( USE entity_aspect )? ( generic_map_aspect )? ( port_map_aspect )? -> ^(T_binding_indication entity_aspect? generic_map_aspect? port_map_aspect?)
  ;

block_configuration
  : FOR block_specification
    ( use_clause )*
    ( configuration_item )*
    END FOR SEMI -> ^(T_block_configuration block_specification use_clause* configuration_item*)
  ;
//TODO: mozno use_clause bude treba implementovat
block_declarative_item
  : (subprogram_declaration)=> subprogram_declaration -> ^(T_block_declarative_item subprogram_declaration)
  | subprogram_body -> ^(T_block_declarative_item subprogram_body)
  | type_declaration -> ^(T_block_declarative_item type_declaration)
  | subtype_declaration -> ^(T_block_declarative_item subtype_declaration)
  | constant_declaration -> ^(T_block_declarative_item constant_declaration)
  | signal_declaration -> ^(T_block_declarative_item signal_declaration)
  | variable_declaration -> ^(T_block_declarative_item variable_declaration)
  | file_declaration -> ^(T_block_declarative_item file_declaration)
  | alias_declaration -> ^(T_block_declarative_item alias_declaration)
  | component_declaration -> ^(T_block_declarative_item component_declaration)
  | attribute_declaration -> ^(T_block_declarative_item attribute_declaration)
  | attribute_specification -> ^(T_block_declarative_item attribute_specification)
  | configuration_specification -> ^(T_block_declarative_item configuration_specification)
  | disconnection_specification -> ^(T_block_declarative_item disconnection_specification)
  | step_limit_specification -> ^(T_block_declarative_item step_limit_specification)
  | use_clause -> ^(T_block_declarative_item use_clause)
  | group_template_declaration -> ^(T_block_declarative_item group_template_declaration)
  | group_declaration -> ^(T_block_declarative_item group_declaration)
  | nature_declaration -> ^(T_block_declarative_item nature_declaration)
  | subnature_declaration -> ^(T_block_declarative_item subnature_declaration)
  | quantity_declaration -> ^(T_block_declarative_item quantity_declaration)
  | terminal_declaration -> ^(T_block_declarative_item terminal_declaration)
  ;

block_declarative_part
  : ( block_declarative_item )* -> ^(T_block_declarative_part block_declarative_item*)
  ;

block_header
  : ( generic_clause ( generic_map_aspect SEMI )? )?
    ( port_clause ( port_map_aspect SEMI )? )? -> ^(T_block_header generic_clause? generic_map_aspect? port_clause? port_map_aspect?)
  ;

block_specification
  : (identifier)=> identifier ( LPAREN index_specification RPAREN )? -> ^(T_block_specification identifier index_specification?)
  | name -> ^(T_block_specification name)
  ;

block_statement
  : label_colon BLOCK ( LPAREN expression RPAREN )? ( IS )?
    block_header
    block_declarative_part BEGIN
    block_statement_part 
    END BLOCK ( identifier )? SEMI -> ^(T_block_statement label_colon expression? block_header block_declarative_part block_statement_part)
  ;

block_statement_part
  : ( architecture_statement )* -> ^(T_block_statement_part architecture_statement*)
  ;

branch_quantity_declaration
  : QUANTITY ( (across_aspect)=> across_aspect )?
    ( through_aspect )? terminal_aspect SEMI -> ^(T_branch_quantity_declaration across_aspect? through_aspect? terminal_aspect)
  ;

break_element
  : ( break_selector_clause )? name ARROW expression -> ^(T_break_element break_selector_clause? name expression)
  ;

break_list
  : break_element ( COMMA break_element )* -> ^(T_break_list break_element+)
  ;

break_selector_clause
  : FOR name USE -> ^(T_break_selector_clause name)
  ;

break_statement
  : ( label_colon )? BREAK ( break_list )? ( WHEN condition )? SEMI -> ^(T_break_statement label_colon? break_list? condition?)
  ;

case_statement
  : ( label_colon )? CASE expression IS
    ( case_statement_alternative )+
    END CASE ( identifier )? SEMI -> ^(T_case_statement label_colon? expression case_statement_alternative+)
  ;

case_statement_alternative
  : WHEN choices ARROW sequence_of_statements -> ^(T_case_statement_alternative choices sequence_of_statements)
  ;

choice
  : (identifier)=> identifier -> ^(T_choice identifier)
  | (discrete_range)=> discrete_range -> ^(T_choice discrete_range)
  | simple_expression -> ^(T_choice simple_expression)
  | OTHERS -> ^(T_choice OTHERS)
  ;

choices
  : choice ( BAR choice )* -> ^(T_choices choice+)
  ;
  
//TODO
component_configuration
  : FOR component_specification
    ( binding_indication SEMI )?
    ( block_configuration )?
    END FOR SEMI -> ^(T_component_configuration component_specification binding_indication? block_configuration?)
  ;

component_declaration
  : COMPONENT identifier ( IS )?
    ( generic_clause )?
    ( port_clause )?
    END COMPONENT ( identifier )? SEMI -> ^(T_component_declaration identifier generic_clause?  port_clause?)
  ;

component_instantiation_statement
  : label_colon instantiated_unit
    ( generic_map_aspect )?
    ( port_map_aspect )? SEMI -> ^(T_component_instantiation_statement label_colon instantiated_unit generic_map_aspect? port_map_aspect?)
  ;

component_specification
  : instantiation_list COLON name -> ^(T_component_specification instantiation_list name)
  ;

composite_nature_definition
  : array_nature_definition -> ^(T_composite_nature_definition array_nature_definition)
  | record_nature_definition -> ^(T_composite_nature_definition record_nature_definition)
  ;

composite_type_definition
  : array_type_definition -> ^(T_composite_type_definition array_type_definition)
  | record_type_definition -> ^(T_composite_type_definition record_type_definition)
  ;

concurrent_assertion_statement
  : ( label_colon )? ( POSTPONED )? assertion SEMI -> ^(T_concurrent_assertion_statement label_colon? assertion)
  ;

concurrent_break_statement
  : ( label_colon )? BREAK ( break_list )? ( sensitivity_clause )?
    ( WHEN condition )? SEMI -> ^(T_concurrent_break_statement label_colon? break_list? sensitivity_clause? condition?)
  ;

concurrent_procedure_call_statement
  : ( label_colon )? ( POSTPONED )? procedure_call SEMI -> ^(T_concurrent_procedure_call_statement label_colon? procedure_call)
  ;

concurrent_signal_assignment_statement
  : ( label_colon )? ( POSTPONED )?
    ( conditional_signal_assignment | selected_signal_assignment ) -> ^(T_concurrent_signal_assignment_statement label_colon? conditional_signal_assignment? selected_signal_assignment?)
  ;

condition
  : expression -> ^(T_condition expression)
  ;

condition_clause
  : UNTIL condition -> ^(T_condition_clause condition)
  ;

conditional_signal_assignment
  : target LE opts conditional_waveforms SEMI -> ^(T_conditional_signal_assignment target opts conditional_waveforms)
  ;

conditional_waveforms
  : waveform ( (WHEN condition ELSE)=> conditional_waveforms_bi )?
    ( WHEN condition )? -> ^(T_conditional_waveforms waveform conditional_waveforms_bi? ( WHEN condition )?)
  ;
conditional_waveforms_bi
  : WHEN condition ELSE waveform
    ( (WHEN condition ELSE)=> conditional_waveforms_bi )? -> ^(T_conditional_waveforms_bi WHEN condition ELSE waveform conditional_waveforms_bi?)
  ;

configuration_declaration
  : CONFIGURATION identifier OF name IS
    configuration_declarative_part
    block_configuration
    END ( CONFIGURATION )? ( identifier )? SEMI -> ^(T_configuration_declaration identifier name configuration_declarative_part block_configuration)
  ;

configuration_declarative_item
  : use_clause -> ^(T_configuration_declarative_item use_clause)
  | attribute_specification -> ^(T_configuration_declarative_item attribute_specification)
  | group_declaration -> ^(T_configuration_declarative_item group_declaration)
  ;

configuration_declarative_part
  : ( configuration_declarative_item )* -> ^(T_configuration_declarative_part configuration_declarative_item*)
  ;

configuration_item
  : block_configuration -> ^(T_configuration_item block_configuration)
  | component_configuration -> ^(T_configuration_item component_configuration)
  ;

configuration_specification
  : FOR component_specification binding_indication SEMI -> ^(T_configuration_specification component_specification binding_indication )
  ;

constant_declaration
  : CONSTANT identifier_list COLON subtype_indication
    ( VARASGN expression )? SEMI -> ^(T_constant_declaration identifier_list subtype_indication expression?)
  ;

constrained_array_definition
  : ARRAY index_constraint OF subtype_indication -> ^(T_constrained_array_definition index_constraint subtype_indication)
  ;

constrained_nature_definition
  : ARRAY index_constraint OF subnature_indication -> ^(T_constrained_nature_definition index_constraint subnature_indication)
  ;

constraint 
  : range_constraint -> ^(T_constraint range_constraint)
  | index_constraint -> ^(T_constraint index_constraint) 
  ;

context_clause
  : ( context_item )* -> ^(T_context_clause context_item*)
  ;

context_item
  : library_clause -> ^(T_context_item library_clause)
  | use_clause -> ^(T_context_item use_clause)
  ;

delay_mechanism
  : TRANSPORT
  | ( REJECT expression )? INERTIAL -> ^(T_delay_mechanism expression?)
  ;

design_file
  : ( design_unit )* EOF -> ^(T_design_file design_unit*)
  ;

design_unit
  : context_clause library_unit -> ^(T_design_unit context_clause library_unit)
  ;

designator
  : identifier -> ^(T_designator identifier)
  | STRING_LITERAL -> ^(T_designator STRING_LITERAL)
  ;

direction
  : TO
  | DOWNTO
  ;

disconnection_specification
  : DISCONNECT guarded_signal_specification AFTER expression SEMI -> ^(T_disconnection_specification guarded_signal_specification expression)
  ;

discrete_range
  : (range)=> range// -> ^(T_discrete_range range)
  | subtype_indication// -> ^(T_discrete_range subtype_indication)
  ;

element_association
  : ( (choices ARROW)=> choices ARROW )? expression -> ^(T_element_association choices? expression)
  ;
 
element_declaration
  : identifier_list COLON element_subtype_definition SEMI -> ^(T_element_declaration identifier_list element_subtype_definition)
  ;

element_subnature_definition
  : subnature_indication -> ^(T_element_subnature_definition subnature_indication)
  ;

element_subtype_definition
  : subtype_indication -> ^(T_element_subtype_definition subtype_indication)
  ;

entity_aspect
  : ENTITY entityname ( LPAREN entityname RPAREN )? -> ^(T_entity_aspect ENTITY entityname entityname?)
  | CONFIGURATION name -> ^(T_entity_aspect CONFIGURATION name) 
  | OPEN -> ^(T_entity_aspect OPEN)
  ;
 
entity_class
  : ENTITY
  | ARCHITECTURE
  | CONFIGURATION
  | PROCEDURE
  | FUNCTION
  | PACKAGE
  | TYPE
  | SUBTYPE
  | CONSTANT
  | SIGNAL
  | VARIABLE
  | COMPONENT
  | LABEL
  | LITERAL
  | UNITS
  | GROUP
  | FILE
  | NATURE
  | SUBNATURE
  | QUANTITY
  | TERMINAL
  ;

entity_class_entry
  : entity_class ( BOX )? -> ^(T_entity_class_entry entity_class)
  ;

entity_class_entry_list
  : entity_class_entry ( COMMA entity_class_entry )* -> ^(T_entity_class_entry_list entity_class_entry+)
  ;

entity_declaration
  : ENTITY identifier IS entity_header
    entity_declarative_part
    ( BEGIN entity_statement_part )?
    END ( ENTITY )? ( identifier )? SEMI -> ^(T_entity_declaration identifier entity_header entity_declarative_part entity_statement_part?);
 

entity_declarative_item
  : (subprogram_declaration)=> subprogram_declaration -> ^(T_entity_declarative_item subprogram_declaration)
  | subprogram_body -> ^(T_entity_declarative_item subprogram_body)
  | type_declaration -> ^(T_entity_declarative_item type_declaration)
  | subtype_declaration -> ^(T_entity_declarative_item subtype_declaration)
  | constant_declaration -> ^(T_entity_declarative_item constant_declaration)
  | signal_declaration -> ^(T_entity_declarative_item signal_declaration)
  | variable_declaration -> ^(T_entity_declarative_item variable_declaration)
  | file_declaration -> ^(T_entity_declarative_item file_declaration)
  | alias_declaration -> ^(T_entity_declarative_item alias_declaration)
  | attribute_declaration -> ^(T_entity_declarative_item attribute_declaration)
  | attribute_specification -> ^(T_entity_declarative_item attribute_specification)
  | disconnection_specification -> ^(T_entity_declarative_item disconnection_specification)
  | step_limit_specification -> ^(T_entity_declarative_item step_limit_specification)
  | use_clause -> ^(T_entity_declarative_item use_clause)
  | group_template_declaration -> ^(T_entity_declarative_item group_template_declaration)
  | group_declaration -> ^(T_entity_declarative_item group_declaration)
  | nature_declaration -> ^(T_entity_declarative_item nature_declaration)
  | subnature_declaration -> ^(T_entity_declarative_item subnature_declaration)
  | quantity_declaration -> ^(T_entity_declarative_item quantity_declaration)
  | terminal_declaration -> ^(T_entity_declarative_item terminal_declaration)
  ;

entity_declarative_part
  : ( entity_declarative_item )* -> ^(T_entity_declarative_part entity_declarative_item*)
  ;

entity_designator
  : entity_tag ( signature )? -> ^(T_entity_designator entity_tag signature?)
  ;

entity_header
  : ( generic_clause )?
    ( port_clause )?  -> ^(T_entity_header generic_clause?  port_clause?)
  ;

entity_name_list
  : entity_designator ( COMMA entity_designator )* -> ^(T_entity_name_list entity_designator entity_designator*)
  | OTHERS -> ^(T_entity_name_list OTHERS)
  | ALL -> ^(T_entity_name_list ALL)
  ;

entity_specification
  : entity_name_list COLON entity_class -> ^(T_entity_specification entity_name_list entity_class)
  ;

entity_statement
  : (( label_colon )? ( POSTPONED )? ASSERT)=> concurrent_assertion_statement -> ^(T_entity_statement concurrent_assertion_statement)
  | (( label_colon )? ( POSTPONED )? PROCESS)=> process_statement -> ^(T_entity_statement process_statement)
  | concurrent_procedure_call_statement -> ^(T_entity_statement concurrent_procedure_call_statement)
  ;

entity_statement_part
  : ( entity_statement )* -> ^(T_entity_statement_part entity_statement*)
  ;

entity_tag
  : identifier
  | CHARACTER_LITERAL
  | STRING_LITERAL
  ;

enumeration_literal
  : identifier
  | CHARACTER_LITERAL
  ;

enumeration_type_definition
  : LPAREN enumeration_literal ( COMMA enumeration_literal )* RPAREN -> ^(T_enumeration_type_definition enumeration_literal+)
  ;

exit_statement
  : ( label_colon )? EXIT ( identifier )? ( WHEN condition )? SEMI -> ^(T_exit_statement label_colon? identifier? condition?)
  ;

// NOTE that NAND/NOR are in (...)* now (used to be in (...)?).
// (21.1.2004, e.f.)
expression
  : relation ( (options{greedy=true;}: logical_operator relation) )*// -> ^(T_expression relation (logical_operator relation)*)
  ;
  
//pridal som helpexpression, aby sa to dalo dat do stromu 
helpexpression
  : (options{greedy=true;}: logical_operator relation) -> ^(T_helpexpression logical_operator relation)
  ;

factor
  : primary ( options{greedy=true;}: DOUBLESTAR primary )? -> ^(T_factor primary (^(T_operator DOUBLESTAR) primary)?)
  | ABS primary -> ^(T_factor ^(T_operator ABS) primary)
  | NOT primary -> ^(T_factor ^(T_operator NOT) primary)
  ;

file_declaration
  : FILE identifier_list COLON subtype_indication
    ( file_open_information )? SEMI -> ^(T_file_declaration identifier_list subtype_indication file_open_information?)
  ;

file_logical_name
  : expression -> ^(T_file_logical_name expression)
  ;

file_open_information
  : ( OPEN expression )? IS file_logical_name -> ^(T_file_open_information expression? file_logical_name)
  ;

file_type_definition
  : FILE OF name -> ^(T_file_type_definition name)
  ;

formal_parameter_list
  : interface_list -> ^(T_formal_parameter_list interface_list)
  ;

formal_part
  : name ( LPAREN  name RPAREN )? -> ^(T_formal_part name name?)
  ;

free_quantity_declaration
  : QUANTITY identifier_list COLON subtype_indication
    ( VARASGN expression )? SEMI -> ^(T_free_quantity_declaration identifier_list subtype_indication expression?)
  ;

function_call
  : name ( LPAREN actual_parameter_part RPAREN )?// -> ^(T_function_call name actual_parameter_part?)
  ;

generate_statement
  : label_colon generation_scheme
    GENERATE
    ( helpgenerate_statement BEGIN )?
    ( architecture_statement )*
    END GENERATE ( identifier )? SEMI -> ^(T_generate_statement label_colon generation_scheme helpgenerate_statement? architecture_statement*)
  ;
  
//vytvorene koli jednoduchsuemu vlozeniu do stromu  
helpgenerate_statement
  : ( block_declarative_item )* -> ^(T_helpgenerate_statement block_declarative_item*)
  ;

generation_scheme
  : FOR parameter_specification -> ^(T_generation_scheme parameter_specification)
  | IF condition -> ^(T_generation_scheme condition)
  ;

generic_clause
  : GENERIC LPAREN generic_list RPAREN SEMI -> ^(T_generic_clause generic_list)
  ;

generic_list
  : interface_list -> ^(T_generic_list interface_list)
  ;

generic_map_aspect
  : GENERIC MAP LPAREN association_list RPAREN -> ^(T_generic_map_aspect association_list)
  ;

group_constituent
  : name -> ^(T_group_constituent name)
  | CHARACTER_LITERAL -> ^(T_group_constituent CHARACTER_LITERAL)
  ;

group_constituent_list
  : group_constituent ( COMMA group_constituent )* -> ^(T_group_constituent_list group_constituent+)
  ;

group_declaration
  : GROUP label_colon name
    LPAREN group_constituent_list RPAREN SEMI -> ^(T_group_declaration label_colon name group_constituent_list)
  ;

group_template_declaration
  : GROUP identifier IS LPAREN entity_class_entry_list RPAREN SEMI -> ^(T_group_template_declaration identifier entity_class_entry_list)
  ;

guarded_signal_specification
  : signal_list COLON name -> ^(T_guarded_signal_specification signal_list name)
  ;

identifier
  : BASIC_IDENTIFIER
  | EXTENDED_IDENTIFIER
  ;

identifier_list
  : identifier ( COMMA identifier )* -> ^(T_identifier_list identifier+)
  ;

if_statement
  : ( label_colon )? IF condition THEN
    sequence_of_statements
    ( elsif_statement )* 
    ( else_statement )?
    END IF ( identifier )? SEMI -> ^(T_if_statement label_colon? condition sequence_of_statements elsif_statement* else_statement?)
  ;

//vytvorene, aby sa to dalo lahsie do stromu
fragment
elsif_statement
  : ELSIF condition THEN sequence_of_statements -> ^(T_elsif_statement condition sequence_of_statements)
  ;

//vytvorene, aby sa to dalo lahsie do stromu
fragment
else_statement
  : ELSE sequence_of_statements -> ^(T_else_statement sequence_of_statements)
  ;

index_constraint
  : LPAREN discrete_range ( COMMA discrete_range )* RPAREN -> ^(T_index_constraint discrete_range+)
  ;

index_specification
  : (discrete_range)=> discrete_range -> ^(T_index_specification discrete_range)
  | expression -> ^(T_index_specification expression)
  ;

index_subtype_definition
  : name RANGETOK BOX -> ^(T_index_subtype_definition name)
  ;

instantiated_unit
  : ( COMPONENT )? name -> ^(T_instantiated_unit COMPONENT? name)
  | ENTITY name ( LPAREN identifier RPAREN )? -> ^(T_instantiated_unit ENTITY name identifier?)
  | CONFIGURATION name -> ^(T_instantiated_unit CONFIGURATION name)
  ;

instantiation_list
  : identifier ( COMMA identifier )* -> ^(T_instantiation_list identifier+)
  | OTHERS -> ^(T_instantiation_list OTHERS)
  | ALL -> ^(T_instantiation_list ALL)
  ;

interface_constant_declaration
  : ( CONSTANT )? identifier_list COLON ( IN )? subtype_indication 
    ( VARASGN expression )? -> ^(T_interface_constant_declaration identifier_list subtype_indication expression?)
  ;

//My Change: (interface_constant_declaration)=> interface_constant_declaration   NA  (interface_constant_declaration)=> interface_signal_declaration  
//Reason: naozv: in bit;  pre mod in defaultne pasuje aj interface_constant_declaration, preto som zmenil na signal declaration ako prioritna
interface_declaration
  : (interface_signal_declaration)=> interface_signal_declaration -> ^(T_interface_declaration interface_signal_declaration)
  | (interface_constant_declaration)=> interface_constant_declaration -> ^(T_interface_declaration interface_constant_declaration)
  | interface_variable_declaration -> ^(T_interface_declaration interface_variable_declaration)
  | interface_file_declaration -> ^(T_interface_declaration interface_file_declaration)
  | interface_terminal_declaration -> ^(T_interface_declaration interface_terminal_declaration)
  | interface_quantity_declaration -> ^(T_interface_declaration interface_quantity_declaration)
  ;

interface_element
  : interface_declaration -> ^(T_interface_element interface_declaration)
  ;

interface_file_declaration
  : FILE identifier_list COLON subtype_indication -> ^(T_interface_file_declaration identifier_list subtype_indication)
  ;

interface_list
  : interface_element ( SEMI interface_element )* -> ^(T_interface_list interface_element+)
  ; 

interface_quantity_declaration
  : QUANTITY identifier_list COLON ( IN | OUT )? subtype_indication
    ( VARASGN expression )? -> ^(T_interface_quantity_declaration identifier_list subtype_indication expression?)
  ;

interface_signal_declaration
  : ( SIGNAL )? identifier_list COLON ( mode )? subtype_indication
    ( BUS )? ( VARASGN expression )? -> ^(T_interface_signal_declaration  identifier_list mode? subtype_indication expression?)
  ;

interface_terminal_declaration
  : TERMINAL identifier_list COLON subnature_indication -> ^(T_interface_terminal_declaration identifier_list subnature_indication)
  ;

interface_variable_declaration
  : ( VARIABLE )? identifier_list COLON
    ( mode )? subtype_indication ( VARASGN expression )? -> ^(T_interface_variable_declaration identifier_list mode? subtype_indication expression?)
  ;

iteration_scheme
  : WHILE condition -> ^(T_iteration_scheme WHILE condition)
  | FOR parameter_specification -> ^(T_iteration_scheme FOR parameter_specification)
  ;

label_colon
  : identifier COLON -> ^(T_label_colon identifier)
  ;

library_clause
  : LIBRARY logical_name_list SEMI -> ^(T_library_clause logical_name_list)
  ;

library_unit
  : ( ARCHITECTURE | PACKAGE BODY )=> secondary_unit -> ^(T_library_unit secondary_unit)
  | primary_unit -> ^(T_library_unit primary_unit)
  ;

literal
  : NULLTOK
  | BIT_STRING_LITERAL
  | (DBLQUOTE)=> STRING_LITERAL
  | (enumeration_literal)=> enumeration_literal
  | numeric_literal
  ;

logical_name
  : identifier
  ;

logical_name_list
  : logical_name ( COMMA logical_name )* -> ^(T_logical_name_list logical_name+)
  ;

logical_operator
  : AND -> ^(T_operator AND)
  | OR -> ^(T_operator OR)
  | NAND -> ^(T_operator NAND)
  | NOR -> ^(T_operator NOR)
  | XOR -> ^(T_operator XOR)
  | XNOR -> ^(T_operator XNOR)
  ;

loop_statement
  : ( label_colon )? ( iteration_scheme )?
    LOOP
    sequence_of_statements 
    END LOOP ( identifier )? SEMI -> ^(T_loop_statement label_colon? iteration_scheme? sequence_of_statements)
  ;

mode
  : IN -> ^(T_mode IN)
  | OUT -> ^(T_mode OUT)
  | INOUT -> ^(T_mode INOUT)
  | BUFFER -> ^(T_mode BUFFER)
  | LINKAGE -> ^(T_mode LINKAGE)
  ; 

multiplying_operator
  : MUL -> ^(T_operator MUL)
  | DIV -> ^(T_operator DIV)
  | MOD -> ^(T_operator MOD)
  | REM -> ^(T_operator REM)
  ;


// was
//   name
//     : simple_name
//     | operator_symbol
//     | selected_name
//     | indexed_name
//     | slice_name
//     | attribute_name
//     ;
// changed to avoid left-recursion to name (from selected_name, indexed_name,
// slice_name, and attribute_name, respectively)
// (2.2.2004, e.f.)

//My change: name som rozdelil do name a helpname, aby som to lahsie dal do stromu;
/*name
  : ( identifier | STRING_LITERAL )
    ( options{greedy=true;}: helpname )* -> ^(T_name identifier? STRING_LITERAL? helpname*)
  ;
  
fragment
helpname
  : DOT suffix -> ^(T_helpname DOT suffix)
  | APOSTROPHE aggregate -> ^(T_helpname aggregate)
  | ( signature )? APOSTROPHE attribute_designator -> ^(T_helpname signature? attribute_designator)
  | (LPAREN expression ( COMMA expression )* RPAREN)=>
    LPAREN expression ( COMMA expression )* RPAREN -> ^(T_helpname expression+)
  | (LPAREN actual_parameter_part RPAREN)=> 
    LPAREN actual_parameter_part RPAREN -> ^(T_helpname actual_parameter_part)
  | LPAREN discrete_range ( COMMA discrete_range )* RPAREN -> ^(T_helpname discrete_range+)
  ;*/
name 
  : helpname -> ^(T_name helpname)
  ;
  
hname
  : helpname
  ;
  
fragment
helpname
  : ( identifier | STRING_LITERAL )
    ( options{greedy=true;}:
      (
          DOT suffix
        | APOSTROPHE aggregate
        | ( signature )? APOSTROPHE attribute_designator
        | (LPAREN expression ( COMMA expression )* RPAREN)=>
            LPAREN expression ( COMMA expression )* RPAREN
        | (LPAREN actual_parameter_part RPAREN)=> 
            LPAREN actual_parameter_part RPAREN
        | LPAREN discrete_range ( COMMA discrete_range )* RPAREN
      ) 
    )*
  ;

//My change - added entityname, becase with just name it was complicated to reach architecutre name: name LPAREN identifier PPAREN
entityname
  : identifier (DOT suffix)* -> ^(T_entityname identifier (DOT suffix)*)
  ;

nature_declaration
  : NATURE identifier IS nature_definition SEMI -> ^(T_nature_declaration identifier nature_definition)
  ;

nature_definition
  : scalar_nature_definition -> ^(T_nature_definition scalar_nature_definition)
  | composite_nature_definition -> ^(T_nature_definition composite_nature_definition)
  ;

nature_element_declaration
  : identifier_list COLON element_subnature_definition -> ^(T_nature_element_declaration identifier_list element_subnature_definition)
  ;

next_statement
  : ( label_colon )? NEXT ( identifier )? ( WHEN condition )? SEMI -> ^(T_next_statement label_colon? identifier? condition?)
  ;

numeric_literal
  : (abstract_literal)=> abstract_literal// -> ^(T_numeric_literal abstract_literal)
  | physical_literal// -> ^(T_numeric_literal physical_literal)
  ;

object_declaration
  : constant_declaration -> ^(T_object_declaration constant_declaration)
  | signal_declaration -> ^(T_object_declaration signal_declaration)
  | variable_declaration -> ^(T_object_declaration variable_declaration)
  | file_declaration -> ^(T_object_declaration file_declaration)
  | terminal_declaration -> ^(T_object_declaration terminal_declaration)
  | quantity_declaration -> ^(T_object_declaration quantity_declaration)
  ;

opts
  : ( GUARDED )? ( delay_mechanism )? -> ^(T_opts delay_mechanism?)
  ;

package_body
  : PACKAGE BODY identifier IS
    package_body_declarative_part
    END ( PACKAGE BODY )? ( identifier )? SEMI -> ^(T_package_body identifier package_body_declarative_part)
  ;

package_body_declarative_item
  : (subprogram_declaration)=> subprogram_declaration -> ^(T_package_body_declarative_item subprogram_declaration)
  | subprogram_body -> ^(T_package_body_declarative_item subprogram_body)
  | type_declaration -> ^(T_package_body_declarative_item type_declaration)
  | subtype_declaration -> ^(T_package_body_declarative_item subtype_declaration)
  | constant_declaration -> ^(T_package_body_declarative_item constant_declaration)
  | variable_declaration -> ^(T_package_body_declarative_item variable_declaration)
  | file_declaration -> ^(T_package_body_declarative_item file_declaration)
  | alias_declaration -> ^(T_package_body_declarative_item alias_declaration)
  | use_clause -> ^(T_package_body_declarative_item use_clause)
  | group_template_declaration -> ^(T_package_body_declarative_item group_template_declaration)
  | group_declaration -> ^(T_package_body_declarative_item group_declaration)
  ;

package_body_declarative_part
  : ( package_body_declarative_item )* -> ^(T_package_body_declarative_part package_body_declarative_item*)
  ;

package_declaration
  : PACKAGE identifier IS
    package_declarative_part
    END ( PACKAGE )? ( identifier )? SEMI -> ^(T_package_declaration identifier package_declarative_part)
  ;

package_declarative_item
  : subprogram_declaration -> ^(T_package_declarative_item subprogram_declaration)
  | type_declaration -> ^(T_package_declarative_item type_declaration)
  | subtype_declaration -> ^(T_package_declarative_item subtype_declaration)
  | constant_declaration -> ^(T_package_declarative_item constant_declaration)
  | signal_declaration -> ^(T_package_declarative_item signal_declaration)
  | variable_declaration -> ^(T_package_declarative_item variable_declaration)
  | file_declaration -> ^(T_package_declarative_item file_declaration)
  | alias_declaration -> ^(T_package_declarative_item alias_declaration)
  | component_declaration -> ^(T_package_declarative_item component_declaration)
  | attribute_declaration -> ^(T_package_declarative_item attribute_declaration)
  | attribute_specification -> ^(T_package_declarative_item attribute_specification)
  | disconnection_specification -> ^(T_package_declarative_item disconnection_specification)
  | use_clause -> ^(T_package_declarative_item use_clause)
  | group_template_declaration -> ^(T_package_declarative_item group_template_declaration)
  | group_declaration -> ^(T_package_declarative_item group_declaration)
  | nature_declaration -> ^(T_package_declarative_item nature_declaration)
  | subnature_declaration -> ^(T_package_declarative_item subnature_declaration)
  | terminal_declaration -> ^(T_package_declarative_item terminal_declaration)
  ;

package_declarative_part
  : ( package_declarative_item )* -> ^(T_package_declarative_part package_declarative_item*)
  ;

parameter_specification
  : identifier IN discrete_range -> ^(T_parameter_specification identifier discrete_range)
  ;

physical_literal
  : ( abstract_literal )? name// -> ^(T_physical_literal abstract_literal? name)
  ;

physical_type_definition
  : range_constraint UNITS base_unit_declaration
    ( secondary_unit_declaration )* 
    END UNITS ( identifier )? -> ^(T_physical_type_definition range_constraint base_unit_declaration secondary_unit_declaration*)
  ;

port_clause
  : PORT  LPAREN port_list RPAREN SEMI -> ^(T_port_clause port_list) 
  ;

port_list
  : interface_list -> ^(T_port_list interface_list)
  ;

port_map_aspect
  : PORT MAP LPAREN association_list RPAREN -> ^(T_port_map_aspect association_list)
  ;

primary
  : (function_call)=> function_call// -> ^(T_primary function_call)
  | (name APOSTROPHE)=> qualified_expression -> ^(T_primary qualified_expression)
  | (LPAREN expression RPAREN)=> LPAREN expression RPAREN -> ^(T_primary LPAREN expression RPAREN)
  | literal// -> ^(T_primary literal)
  | allocator// -> ^(T_primary allocator)
  | aggregate// -> ^(T_primary aggregate)
  ;

primary_unit
  : entity_declaration -> ^(T_primary_unit entity_declaration)
  | configuration_declaration -> ^(T_primary_unit configuration_declaration)
  | package_declaration -> ^(T_primary_unit package_declaration)
  ;

procedural_declarative_item
  : (subprogram_declaration)=> subprogram_declaration -> ^(T_procedural_declarative_item subprogram_declaration)
  | subprogram_body -> ^(T_procedural_declarative_item subprogram_body)
  | type_declaration -> ^(T_procedural_declarative_item type_declaration)
  | subtype_declaration -> ^(T_procedural_declarative_item subtype_declaration)
  | constant_declaration -> ^(T_procedural_declarative_item constant_declaration)
  | variable_declaration -> ^(T_procedural_declarative_item variable_declaration)
  | alias_declaration -> ^(T_procedural_declarative_item alias_declaration)
  | attribute_declaration -> ^(T_procedural_declarative_item attribute_declaration)
  | attribute_specification -> ^(T_procedural_declarative_item attribute_specification)
  | use_clause -> ^(T_procedural_declarative_item use_clause)
  | group_template_declaration -> ^(T_procedural_declarative_item group_template_declaration)
  | group_declaration -> ^(T_procedural_declarative_item group_declaration)
  ;

procedural_declarative_part
  : ( procedural_declarative_item )* -> ^(T_procedural_declarative_part procedural_declarative_item*)
  ;

procedural_statement_part
  : ( sequential_statement )* -> ^(T_procedural_statement_part sequential_statement*)
  ;

procedure_call
  : name ( LPAREN actual_parameter_part RPAREN )? -> ^(T_procedure_call name actual_parameter_part?)
  ;

procedure_call_statement
  : ( label_colon )? procedure_call SEMI -> ^(T_procedure_call_statement label_colon? procedure_call)
  ;

process_declarative_item
  : (subprogram_declaration)=> subprogram_declaration -> ^(T_process_declarative_item subprogram_declaration)
  | subprogram_body -> ^(T_process_declarative_item subprogram_body)
  | type_declaration -> ^(T_process_declarative_item type_declaration)
  | subtype_declaration -> ^(T_process_declarative_item subtype_declaration)
  | constant_declaration -> ^(T_process_declarative_item constant_declaration)
  | variable_declaration -> ^(T_process_declarative_item variable_declaration)
  | file_declaration -> ^(T_process_declarative_item file_declaration)
  | alias_declaration -> ^(T_process_declarative_item alias_declaration)
  | attribute_declaration -> ^(T_process_declarative_item attribute_declaration)
  | attribute_specification -> ^(T_process_declarative_item attribute_specification)
  | use_clause -> ^(T_process_declarative_item use_clause)
  | group_template_declaration -> ^(T_process_declarative_item group_template_declaration)
  | group_declaration -> ^(T_process_declarative_item group_declaration)
  ;

process_declarative_part
  : ( process_declarative_item )* -> ^(T_process_declarative_part process_declarative_item*)
  ;

process_statement
  : ( label_colon )? ( POSTPONED )? PROCESS
    ( LPAREN sensitivity_list RPAREN )? ( IS )?
    process_declarative_part
    BEGIN
    process_statement_part 
    END ( POSTPONED )? PROCESS ( identifier )? SEMI -> ^(T_process_statement label_colon? sensitivity_list? process_declarative_part process_statement_part)
  ;

process_statement_part
  : ( sequential_statement )* -> ^(T_process_statement_part sequential_statement*)
  ;

qualified_expression
  : name APOSTROPHE ( (aggregate)=> aggregate | LPAREN expression RPAREN ) -> ^(T_qualified_expression name aggregate? expression?)
  ;

quantity_declaration
  : (free_quantity_declaration)=> free_quantity_declaration -> ^(T_quantity_declaration free_quantity_declaration)
  | (branch_quantity_declaration)=> branch_quantity_declaration -> ^(T_quantity_declaration branch_quantity_declaration)
  | source_quantity_declaration -> ^(T_quantity_declaration source_quantity_declaration)
  ;

quantity_list
  : name ( COMMA name )* -> ^(T_quantity_list name+)
  | OTHERS -> ^(T_quantity_list OTHERS)
  | ALL -> ^(T_quantity_list ALL)
  ;

quantity_specification
  : quantity_list COLON name -> ^(T_quantity_specification quantity_list name)
  ;

range
  : (simple_expression direction simple_expression)=>
    simple_expression direction simple_expression -> ^(T_range simple_expression direction simple_expression)
  | name -> ^(T_range name)
  ;

range_constraint
  : RANGETOK range -> ^(T_range_constraint RANGETOK range )
  ;

record_nature_definition
  : RECORD ( nature_element_declaration )+
    END RECORD ( identifier )? -> ^(T_record_nature_definition nature_element_declaration+)
  ;

record_type_definition
  : RECORD ( element_declaration )+
    END RECORD ( identifier )? -> ^(T_record_type_definition element_declaration+)
  ;

relation
  : shift_expression
    ( options{greedy=true;}: relational_operator shift_expression )?// -> ^(T_relation shift_expression (relational_operator shift_expression)?)
  ;

relational_operator
  : EQ -> ^(T_operator EQ)
  | NEQ -> ^(T_operator NEQ)
  | LOWERTHAN -> ^(T_operator LOWERTHAN)
  | LE -> ^(T_operator LE)
  | GREATERTHAN -> ^(T_operator GREATERTHAN)
  | GE -> ^(T_operator GE)
  ;

report_statement
  : ( label_colon )? REPORT expression ( SEVERITY expression )? SEMI -> ^(T_report_statement label_colon? expression expression?)
  ;

return_statement
  : ( label_colon )? RETURN ( expression )? SEMI -> ^(T_return_statement label_colon? expression?)
  ;

scalar_nature_definition
  : name ACROSS name THROUGH name REFERENCE -> ^(T_scalar_nature_definition name name name)
  ;

scalar_type_definition
  : (range_constraint UNITS)=> physical_type_definition -> ^(T_scalar_type_definition physical_type_definition)
  | enumeration_type_definition -> ^(T_scalar_type_definition enumeration_type_definition)
  | range_constraint -> ^(T_scalar_type_definition range_constraint)
  ;
  
secondary_unit
  : architecture_body -> ^(T_secondary_unit architecture_body)
  | package_body -> ^(T_secondary_unit package_body)
  ;

secondary_unit_declaration
  : identifier EQ physical_literal SEMI -> ^(T_secondary_unit_declaration identifier physical_literal)
  ;

selected_signal_assignment
  : WITH expression SELECT target LE opts selected_waveforms SEMI -> ^(T_selected_signal_assignment expression target opts selected_waveforms)
  ;

selected_waveforms
  : waveform WHEN choices ( COMMA waveform WHEN choices )* -> ^(T_selected_waveforms (waveform choices)+)
  ;

sensitivity_clause
  : ON sensitivity_list -> ^(T_sensitivity_clause sensitivity_list)
  ;

sensitivity_list
  : name ( COMMA name )* -> ^(T_sensitivity_list name+)
  ;

sequence_of_statements
  : ( sequential_statement )* -> ^(T_sequence_of_statements sequential_statement*)
  ;

sequential_statement
  : (( label_colon )? WAIT)=> wait_statement -> ^(T_sequential_statement wait_statement)
  | (( label_colon )? ASSERT)=> assertion_statement -> ^(T_sequential_statement assertion_statement)
  | (( label_colon )? REPORT)=> report_statement -> ^(T_sequential_statement report_statement)
  | (( label_colon )? target LE)=> signal_assignment_statement -> ^(T_sequential_statement signal_assignment_statement)
  | (( label_colon )? target VARASGN)=> variable_assignment_statement -> ^(T_sequential_statement variable_assignment_statement)
  | (( label_colon )? IF)=> if_statement -> ^(T_sequential_statement if_statement)
  | (( label_colon )? CASE)=> case_statement -> ^(T_sequential_statement case_statement)
  | (( label_colon )? ( iteration_scheme )? LOOP)=> loop_statement -> ^(T_sequential_statement loop_statement)
  | (( label_colon )? NEXT)=> next_statement -> ^(T_sequential_statement next_statement)
  | (( label_colon )? EXIT)=> exit_statement -> ^(T_sequential_statement exit_statement)
  | (( label_colon )? RETURN)=> return_statement -> ^(T_sequential_statement return_statement)
  | (( label_colon )? NULLTOK SEMI)=> ( label_colon )? NULLTOK SEMI -> ^(T_sequential_statement label_colon? NULLTOK)
  | (( label_colon )? BREAK)=> break_statement -> ^(T_sequential_statement break_statement)
  | procedure_call_statement -> ^(T_sequential_statement procedure_call_statement)
  ;

shift_expression
  : simple_expression
    ( options{greedy=true;}: shift_operator simple_expression )?// -> ^(T_shift_expression simple_expression (shift_operator simple_expression)?)
  ;

shift_operator
  : SLL -> ^(T_operator SLL)
  | SRL -> ^(T_operator SRL)
  | SLA -> ^(T_operator SLA)
  | SRA -> ^(T_operator SRA)
  | ROL -> ^(T_operator ROL)
  | ROR -> ^(T_operator ROR)
  ;

signal_assignment_statement
  : ( label_colon )?
    target LE ( delay_mechanism )? waveform SEMI -> ^(T_signal_assignment_statement label_colon? target delay_mechanism? waveform)
  ;

signal_declaration
  : SIGNAL identifier_list COLON
    subtype_indication ( signal_kind )? ( VARASGN expression )? SEMI -> ^(T_signal_declaration identifier_list subtype_indication signal_kind? expression?)
  ;

signal_kind
  : REGISTER
  | BUS
  ;

signal_list
  : name ( COMMA name )* -> ^(T_signal_list name+)
  | OTHERS -> ^(T_signal_list OTHERS)
  | ALL -> ^(T_signal_list ALL)
  ;

signature
  : LBRACKET ( name ( COMMA name )* )? ( RETURN name )? RBRACKET -> ^(T_signature (name+)? (RETURN name)?)
  ;

// NOTE that sign is applied to first operand only (LRM does not permit
// `a op -b' - use `a op (-b)' instead).
// (3.2.2004, e.f.)

//My Change - pridal som  simple_expression na helpsimple_expression aby som to dal jednoducho do stromu
simple_expression 
  : helpsimple_expression ( options{greedy=true;}: adding_operator term )*// -> ^(T_simple_expression helpsimple_expression (adding_operator term)*)
  ;

helpsimple_expression
  : ( PLUS | MINUS )? term// -> ^(T_helpsimple_expression PLUS? MINUS? term)
  ;

simple_simultaneous_statement
  : ( label_colon )?
    simple_expression ASSIGN simple_expression ( tolerance_aspect )? SEMI -> ^(T_simple_simultaneous_statement label_colon? simple_expression simple_expression tolerance_aspect?)
  ;

simultaneous_alternative
  : WHEN choices ARROW simultaneous_statement_part -> ^(T_simultaneous_alternative choices simultaneous_statement_part)
  ;

simultaneous_case_statement
  : ( label_colon )? CASE expression USE
    ( simultaneous_alternative )+ 
    END CASE ( identifier )? SEMI -> ^(T_simultaneous_case_statement label_colon? expression simultaneous_alternative+)
  ;

simultaneous_if_statement
  : ( label_colon )? IF condition USE
    simultaneous_statement_part
    ( ELSIF condition USE simultaneous_statement_part )*
    ( ELSE simultaneous_statement_part )?
    END USE ( identifier )? SEMI -> ^(T_simultaneous_if_statement label_colon? condition simultaneous_statement_part (ELSIF condition simultaneous_statement_part)* (ELSE simultaneous_statement_part)?)
  ;

simultaneous_procedural_statement
  : ( label_colon )? PROCEDURAL ( IS )?
    procedural_declarative_part BEGIN
    procedural_statement_part 
    END PROCEDURAL ( identifier )? SEMI -> ^(T_simultaneous_procedural_statement label_colon? procedural_declarative_part procedural_statement_part)
  ;

simultaneous_statement
  : (( label_colon )? simple_expression ASSIGN)=> simple_simultaneous_statement -> ^(T_simultaneous_statement simple_simultaneous_statement)
  | (( label_colon )? IF condition USE)=> simultaneous_if_statement -> ^(T_simultaneous_statement simultaneous_if_statement)
  | (( label_colon )? CASE expression USE)=> simultaneous_case_statement -> ^(T_simultaneous_statement simultaneous_case_statement)
  | (( label_colon )? PROCEDURAL ( IS )?)=> simultaneous_procedural_statement -> ^(T_simultaneous_statement simultaneous_procedural_statement)
  | ( label_colon )? NULLTOK SEMI -> ^(T_simultaneous_statement NULLTOK)
  ;

simultaneous_statement_part
  : ( simultaneous_statement )* -> ^(T_simultaneous_statement_part simultaneous_statement*)
  ;

source_aspect
  : SPECTRUM simple_expression COMMA simple_expression -> ^(T_source_aspect simple_expression simple_expression)
  | NOISE simple_expression -> ^(T_source_aspect simple_expression)
  ;

source_quantity_declaration
  : QUANTITY identifier_list COLON subtype_indication source_aspect SEMI -> ^(T_source_quantity_declaration identifier_list subtype_indication source_aspect)
  ;

step_limit_specification
  : LIMIT quantity_specification WITH expression SEMI -> ^(T_step_limit_specification quantity_specification expression)
  ;

subnature_declaration
  : SUBNATURE identifier IS subnature_indication SEMI -> ^(T_subnature_declaration identifier subnature_indication)
  ;

subnature_indication
  : name ( index_constraint )? 
    ( TOLERANCE expression ACROSS expression THROUGH )? -> ^(T_subnature_indication name index_constraint? (expression expression)?)
  ;

subprogram_body
  : subprogram_specification IS
    subprogram_declarative_part
    BEGIN
    subprogram_statement_part
    END ( subprogram_kind )? ( designator )? SEMI -> ^(T_subprogram_body subprogram_specification subprogram_declarative_part subprogram_statement_part subprogram_kind? designator?)
  ;

subprogram_declaration
  : subprogram_specification SEMI -> ^(T_subprogram_declaration subprogram_specification)
  ;

subprogram_declarative_item
  : (subprogram_declaration)=> subprogram_declaration -> ^(T_subprogram_declarative_item subprogram_declaration)
  | subprogram_body -> ^(T_subprogram_declarative_item subprogram_body)
  | type_declaration -> ^(T_subprogram_declarative_item type_declaration)
  | subtype_declaration -> ^(T_subprogram_declarative_item subtype_declaration)
  | constant_declaration -> ^(T_subprogram_declarative_item constant_declaration)
  | variable_declaration -> ^(T_subprogram_declarative_item variable_declaration)
  | file_declaration -> ^(T_subprogram_declarative_item file_declaration)
  | alias_declaration -> ^(T_subprogram_declarative_item alias_declaration)
  | attribute_declaration -> ^(T_subprogram_declarative_item attribute_declaration)
  | attribute_specification -> ^(T_subprogram_declarative_item attribute_specification)
  | use_clause -> ^(T_subprogram_declarative_item use_clause)
  | group_template_declaration -> ^(T_subprogram_declarative_item group_template_declaration)
  | group_declaration -> ^(T_subprogram_declarative_item group_declaration)
  ;

subprogram_declarative_part
  : ( subprogram_declarative_item )* -> ^(T_subprogram_declarative_part subprogram_declarative_item*)
  ;

subprogram_kind
  : PROCEDURE
  | FUNCTION
  ;

subprogram_specification
  : PROCEDURE designator ( LPAREN formal_parameter_list RPAREN )? -> ^(T_subprogram_specification PROCEDURE designator formal_parameter_list?)
  | ( PURE | IMPURE )? FUNCTION designator
    ( LPAREN formal_parameter_list RPAREN )? RETURN name -> ^(T_subprogram_specification FUNCTION designator formal_parameter_list? name)
  ;

subprogram_statement_part
  : ( sequential_statement )* -> ^(T_subprogram_statement_part sequential_statement*)
  ;

subtype_declaration
  : SUBTYPE identifier IS subtype_indication SEMI -> ^(T_subtype_declaration identifier subtype_indication)
  ;

// VHDLAMS 1076.1-1999 declares first name as optional. Here, second name
// is made optional to prevent antlr nondeterminism.
// (9.2.2004, e.f.)
subtype_indication
  : name ( name )? ( constraint )? ( (TOLERANCE)=> tolerance_aspect )? -> ^(T_subtype_indication name+ constraint? tolerance_aspect?)
  ;

suffix
  : identifier
  | CHARACTER_LITERAL
  | STRING_LITERAL
  | ALL
  ;

target
  : name -> ^(T_target name)
  | aggregate -> ^(T_target aggregate)
  ;

term
  : factor ( options{greedy=true;}: multiplying_operator factor )* -> ^(T_term factor (multiplying_operator factor)*)
  ;

terminal_aspect
  : name ( TO name )? -> ^(T_terminal_aspect name name?)
  ;

terminal_declaration
  : TERMINAL identifier_list COLON subnature_indication SEMI -> ^(T_terminal_declaration identifier_list subnature_indication)
  ;

through_aspect
  : identifier_list ( tolerance_aspect )? ( VARASGN expression )? THROUGH -> ^(T_through_aspect identifier_list tolerance_aspect? expression?)
  ;

timeout_clause
  : FOR expression -> ^(T_timeout_clause expression)
  ;

tolerance_aspect
  : TOLERANCE expression -> ^(T_tolerance_aspect expression)
  ;

type_declaration
  : TYPE identifier ( IS type_definition )? SEMI -> ^(T_type_declaration identifier type_definition?)
  ;

type_definition
  : scalar_type_definition -> ^(T_type_definition scalar_type_definition)
  | composite_type_definition -> ^(T_type_definition composite_type_definition)
  | access_type_definition -> ^(T_type_definition access_type_definition)
  | file_type_definition -> ^(T_type_definition file_type_definition)
  ;

unconstrained_array_definition
  : ARRAY LPAREN index_subtype_definition ( COMMA index_subtype_definition )*
    RPAREN OF subtype_indication -> ^(T_unconstrained_array_definition index_subtype_definition+ subtype_indication)
  ;

unconstrained_nature_definition
  : ARRAY LPAREN index_subtype_definition ( COMMA index_subtype_definition )*
    RPAREN OF subnature_indication -> ^(T_unconstrained_nature_definition index_subtype_definition+ subnature_indication)
  ;

// NOTE that name is used here, so need to check if result of name is 
// a seleceted_name (i.e. prefix DOT suffix).
// (30.3.2004, e.f.)
use_clause
  : USE name ( COMMA name )* SEMI -> ^(T_use_clause name+)
  ;

variable_assignment_statement
  : ( label_colon )? target VARASGN expression SEMI -> ^(T_variable_assignment_statement label_colon? target expression)
  ;

variable_declaration
  : ( SHARED )? VARIABLE identifier_list COLON
    subtype_indication ( VARASGN expression )? SEMI -> ^(T_variable_declaration identifier_list subtype_indication expression?)
  ;

wait_statement
  : ( label_colon )? WAIT ( sensitivity_clause )? 
    ( condition_clause )? ( timeout_clause )? SEMI -> ^(T_wait_statement label_colon? sensitivity_clause? condition_clause? timeout_clause?)
  ;

waveform
  : waveform_element ( COMMA waveform_element )* -> ^(T_waveform waveform_element (COMMA waveform_element)*)
  | UNAFFECTED -> ^(T_waveform UNAFFECTED)
  ;

waveform_element
  : expression ( AFTER expression )? -> ^(T_waveform_element expression (AFTER expression )?)
  ;

//--------------------------------------------------------------------------




TAB
  : ( '\t' ) + { $channel = HIDDEN; }
  ;
SPACE
  : ( ' ' )+ { $channel = HIDDEN; }
  ;
NEWLINE
  : '\n'  { $channel = HIDDEN; }
  ;

// comment without NEWLINE
// (2.2.2004, e.f.)
COMMENT
  : '--' ( ~'\n' )*  { $channel = HIDDEN; }
  ;

DOUBLESTAR
  : '**'
  ;
ASSIGN
  : '=='
  ;
LE
  : '<='
  ;
GE
  : '>='
  ;
ARROW
  : '=>'
  ;

NEQ
  : '/='
  ;
VARASGN
  : ':='
  ;
BOX
  : '<>'
  ;

DBLQUOTE
  : '\"'
  ;
SEMI
  : ';'
  ;
COMMA
  : ','
  ;
AMPERSAND
  : '&'
  ;
APOSTROPHE
  : '\''
  ;
LPAREN
  : '('
  ;
RPAREN
  : ')'
  ;
LBRACKET
  : '['
  ;
RBRACKET
  : ']'
  ;
COLON
  : ':'
  ;
MUL
  : '*'
  ;
DIV
  : '/'
  ;
PLUS
  : '+'
  ;
MINUS
  : '-'
  ;
LOWERTHAN
  : '<'
  ;
GREATERTHAN
  : '>'
  ;
EQ
  : '='
  ;
BAR
  : '|'
  ;

DOT
  : '.'
  ;
BACKSLASH
  : '\\'
  ;

BASIC_IDENTIFIER
  : 'a'..'z' ( 'a'..'z' | '0'..'9' | '_' )*
  ;

EXTENDED_IDENTIFIER
  : '\\' ( 'a'..'z' | '0'..'9' | '&' | '\'' | '(' | ')'
    | '+' | ',' | '-' | '.' | '/' | ':' | ';' | '<' | '=' | '>' | '|'
    | ' ' | OTHER_SPECIAL_CHARACTER | '\\'
    | '#' | '[' | ']' | '_' )+ '\\'
  ;

BIT_STRING_LITERAL
  : ( 'b' | 'o' | 'x' ) '\"' BASED_INTEGER '\"'
  ;

STRING_LITERAL
  : '\"' ( 'a'..'z' | '0'..'9' | '&' | '\'' | '(' | ')'
    | '+' | ',' | '-' | '.' | '/' | ':' | ';' | '<' | '=' | '>' | '|'
    | ' ' | OTHER_SPECIAL_CHARACTER | '\\' | '*' 
    | '#' | '[' | ']' | '_' )* '\"'
  ;

// NOTE that BASED_LITERAL is implicitely defined by referencing in the
// parser grammar.
// (2.2.2004, e.f.)
DECIMAL_LITERAL
  : INTEGER
    (
      ( ( '.' INTEGER )? ( EXPONENT )? )
    |
      ( '#' BASED_INTEGER ( '.' BASED_INTEGER )? '#' ( EXPONENT )? )
      
    )
  ;

CR
  : '\r' { $channel = HIDDEN; }
  ;
  
fragment
OTHER_SPECIAL_CHARACTER
  : '!' | '$' | '%' | '@' | '?' | '^' | '`' | '{' | '}' | '~'
  ;

fragment
INTEGER
  : '0'..'9' ( '_' | '0'..'9' )*
  ;

fragment
EXPONENT
  : 'e' ( '+' | '-' )? INTEGER
  ;

fragment
BASED_INTEGER
  : ( '0'..'9' | 'a'..'f' ) ( '_' | '0'..'9' | 'a'..'f' )*
  ;

CHARACTER_LITERAL
   : '\''
     (
        (. '\'')=> . '\''
        | { $type = APOSTROPHE; }
     )
   ;

BASED_LITERAL
   :  ( '#' EXTENDED_DIGIT ( '.' EXTENDED_DIGIT )? '#' ( EXPONENT )? )
   ;
   
EXTENDED_DIGIT
   : INTEGER | LETTER
   ;
   
fragment
LETTER	
  :  'a'..'z' | 'A'..'Z'
  ;
