using System;
using System.Collections.Generic;
using System.Text;
using Antlr.Vhdl;
using System.IO;
using Antlr.Runtime.Tree;

namespace VHDL2XML
{
    /// <summary>
    /// Predstavuje info o konkretnej architekture v kode.
    /// </summary>
    public class ArchitectureDeclaration
    {
        #region properties

        bool _IsStructure = false;
        /// <summary>
        /// True, if architecture is defined by structer. Else its behavioural. 
        /// Automatically set when ParseInstances()is called.
        /// </summary>
        public bool IsStructure
        {
            get { return _IsStructure; }
            set { _IsStructure = value; }
        }

        public string FullName
        {
            get { return VHDLToXML.DotConcateName(ArchitectureLibraryName,"",ArchitectureName);}
        }

        public string EntityFullName
        {
            get { return VHDLToXML.DotConcateName(EntityLibraryName,"", EntityName); }
        }

        //private string _ArchitecturePackageName = "";
        ///// <summary>
        ///// meno balika, v ktorom je architektura
        ///// </summary>
        //public string ArchitecturePackageName
        //{
        //    get { return _ArchitecturePackageName; }
        //    set { _ArchitecturePackageName = value.Trim(); }
        //}

        private string _ArchitectureLibraryName = "";
        /// <summary>
        /// meno balika, v ktorom je architektura
        /// </summary>
        public string ArchitectureLibraryName
        {
            get { return _ArchitectureLibraryName; }
            set { _ArchitectureLibraryName = value.Trim(); }
        }


        private string _ArchitectureName = "";
        /// <summary>
        /// Meno architektury
        /// </summary>
        public string ArchitectureName
        {
            get { return _ArchitectureName; }
            set
            {
                _ArchitectureName = value.Trim();
                ArchitectureLibraryName = VHDLToXML.ParseName(ref _ArchitectureName);
            }
        }

        //private string _EntityPackageName = "";

        ////TODO: pri praci s entitami, a hladani ich instancii, zohladnit aj tento package entity
        ////TODO: vseobecne pridat rozpoznavanie packagov a library
        //public string EntityPackageName
        //{
        //    get { return _EntityPackageName; }
        //    set { _EntityPackageName = value; }
        //}

        private string _EntityLibraryName = "";

        //TODO: pri praci s entitami, a hladani ich instancii, zohladnit aj tento package entity
        //TODO: vseobecne pridat rozpoznavanie packagov a library
        public string EntityLibraryName
        {
            get { return _EntityLibraryName; }
            set { _EntityLibraryName = value; }
        }


        private string _EntityName = "";
        /// <summary>
        /// Meno entity, ku ktorej patri dana architektura.
        /// </summary>
        public string EntityName
        {
            get { return _EntityName; }
            set
            {
                _EntityName = value.Trim();
                EntityLibraryName = VHDLToXML.ParseName(ref _EntityName);
            }
        }

        private EntityDeclaration _EntityReference = null;

        public EntityDeclaration EntityReference
        {
            get { return _EntityReference; }
            set { _EntityReference = value; }
        }

        List<ComponentDeclaration> _ComponentDeclarations = new List<ComponentDeclaration>();

        /// <summary>
        /// List of components declarated in the architecture.
        /// </summary>
        public List<ComponentDeclaration> ComponentDeclarations
        {
            get { return _ComponentDeclarations; }
            set { _ComponentDeclarations = value; }
        }


        List<ComponentConfigSpec> _ComponentConfigSpecifications = new List<ComponentConfigSpec>();

        public List<ComponentConfigSpec> ComponentConfigSpecifications
        {
            get { return _ComponentConfigSpecifications; }
            set { _ComponentConfigSpecifications = value; }
        }


        List<Signal> _Signals = new List<Signal>();

        /// <summary>
        /// Signals defined in architecture declaration.
        /// </summary>
        public List<Signal> Signals
        {
            get { return _Signals; }
            set { _Signals = value; }
        }

        List<Entity_Component_Instance> _Instances = new List<Entity_Component_Instance>();

        public List<Entity_Component_Instance> Instances
        {
            get { return _Instances; }
            set { _Instances = value; }
        }


        List<ArchitectureDeclaration> _ArchitecturesInDesign = new List<ArchitectureDeclaration>();
        /// <summary>
        /// Reference to other architectures in design.
        /// </summary>
        public List<ArchitectureDeclaration> ArchitecturesInDesign
        {
            get { return _ArchitecturesInDesign; }
            set { _ArchitecturesInDesign = value; }
        }

        List<EntityDeclaration> _EntitiesInDesign = new List<EntityDeclaration>();
        /// <summary>
        /// Reference to other entities in design.
        /// </summary>
        public List<EntityDeclaration> EntitiesInDesign
        {
            get { return _EntitiesInDesign; }
            set { _EntitiesInDesign = value; }
        }

        List<Connection> _Connections = new List<Connection>();
        /// <summary>
        /// Zoznam konekcii v ramci architektury.
        /// </summary>
        public List<Connection> Connections
        {
            get { return _Connections; }
            set { _Connections = value; }
        }

        private int maxLeft = 900;
        /// <summary>
        /// lava suradnica (x) najlavejsieho prvku
        /// </summary>
        public int MaxLeft
        {
            get { return maxLeft; }
            set { maxLeft = value; }
        }


        List<ITree> _ConditionalSignalAssigments = new List<ITree>();
        /// <summary>
        /// Node s VHDL kodom PortMap (); ku danej instancii
        /// </summary>
        public List<ITree> ConditionalSignalAssigments
        {
            get { return _ConditionalSignalAssigments; }
            set { _ConditionalSignalAssigments = value; }
        }

        #endregion

        #region functions
        /// <summary>
        /// Portmapping instancii a referencie na entity a architektury.
        /// Nastavi sa - IsStructure
        /// architecturesInDesing - zoznam preparsovanych architektur
        /// entitiesInDesing - zoznam parsovanych entit
        /// </summary>
        public void PostParse(List<ArchitectureDeclaration> architecturesInDesing, List<EntityDeclaration> entitiesInDesing)
        {
            //signal value
            /*foreach (Signal s in Signals)
            {
                portvalues.Add(s.Name, -1, " " + s.signalvalue + " ");
            }*/


            ArchitecturesInDesign = architecturesInDesing;
            EntitiesInDesign = entitiesInDesing;
            //najdem entity a architektury pre instancie komponentov resp. entit
            foreach (Entity_Component_Instance instance in Instances)
            {
                FindReferencesForInstance(instance);
            }

            if (Instances.Count > 0) IsStructure = true;
            else IsStructure = false;

            //najdem entitu pre danu architekturu
            foreach (EntityDeclaration entity in EntitiesInDesign)
            {
                if (String.Compare(entity.FullName, this.EntityFullName, true) == 0)
                {
                    EntityReference = entity;

                    //inicializujem connections; t.j. porty entity, na ktore neskor napojim instancie
                    foreach (EntityPort port in entity.Ports)
                    {
                        Connections.Add(new Connection(port));
                    }
                    break;
                }
            }

            //port mapping pre kazdu instanciu;
            for (int i = 0; i < Instances.Count; i++)
            {
                if (Instances[i].VHDLPortMap != null)
                {
                    //pre kazdu polozku portmapu
                    //association_list
                    CommonTree pomtree = (CommonTree)Instances[i].VHDLPortMap.GetChild(0);
                    for (int ii = 0; ii < pomtree.ChildCount; ii++)
                    {
                        String fromName = "";
                        string fromRangeName = "";
                        EntityPort fromPort = null;

                        String toName = "";
                        string toRangeName = "";

                        Signal sig = null;
                        Connection conn = null;

                        int sigindex = -1;

                        //ak je priradenie v portmap manualne, t.j. portmap (a=>b.c=>d)
                        if (pomtree.GetChild(ii).ChildCount > 1)
                        {
                            toName = pomtree.GetChild(ii).GetChild(1).ToMyString();
                            toRangeName = VHDLToXML.ExtractRangeName(ref toName);

                            fromName = pomtree.GetChild(ii).GetChild(0).ToMyString();
                            fromRangeName = VHDLToXML.ExtractRangeName(ref fromName);

                            

                            foreach (EntityPort port in Instances[i].EntityInstance.Ports)
                            {
                                if (String.Compare(port.Name, fromName, true) == 0)
                                {
                                    fromPort = port;

                                    if (port.Direction == portDirection.IN)
                                    {
                                        int pom = -1;
                                        if (fromRangeName != "")
                                            pom = int.Parse(fromRangeName.Replace("(", "").Replace(")", ""));
                                        portvalues.Add(Instances[i].InstanceName + "." + fromName, pom, " " + toName + UpravRangeName(toRangeName) + " ");
                                        //portvalues.Add(Instances[i].GetEntityArchitectureName() + "." + fromName, pom, " " + toName + UpravRangeName(toRangeName) + " ");
                                    }
                                    else
                                    {
                                        int pom = -1;
                                        if (toRangeName != "")
                                            pom = int.Parse(toRangeName.Replace("(", "").Replace(")", ""));
                                        portvalues.Add(toName, pom, " " + Instances[i].InstanceName + "." + fromName + UpravRangeName(fromRangeName) + " ");
                                        //portvalues.Add(toName, pom, " " + Instances[i].GetEntityArchitectureName() + "." + fromName + UpravRangeName(fromRangeName) + " ");
                                        string s = "";
                                        if (Instances[i].ArchitectureInstance != null)
                                            s = Instances[i].ArchitectureInstance.opv.GetValueOf(fromName,-1);
                                        if (s.Trim() != "")
                                        {
                                            foreach (EntityPort p in Instances[i].EntityInstance.Ports)
                                            {
                                                s = s.Replace(" " + p.Name + " ", " " + Instances[i].InstanceName + "." + p.Name + " ");
                                                for (int x = 0; x < p.pinCount; x++)
                                                    s = s.Replace(" " + p.Name + "(" + x.ToString() + ") ", " " + Instances[i].InstanceName + "." + p.Name + "(" + x.ToString() + ") ");
                                            }
                                            //signal upravime
                                            if (Instances[i].ArchitectureInstance != null)
                                                foreach (Signal ssig in Instances[i].ArchitectureInstance.Signals)
                                                {
                                                    s = s.Replace(" " + ssig.Name + " ", " " + Instances[i].InstanceName + "." + ssig.Name + " ");
                                                    for (int x = 0; x < ssig.pinCount; x++)
                                                        s = s.Replace(" " + ssig.Name + "(" + x.ToString() + ") ", " " + Instances[i].InstanceName + "." + ssig.Name + "(" + x.ToString() + ") ");
                                                }
                                            //
                                            if (s == Instances[i].ArchitectureInstance.opv.GetValueOf(fromName, -1))
                                                s = " " + Instances[i].InstanceName + "." + s.Trim() + " ";

                                            portvalues.Add(Instances[i].InstanceName + "." + fromName, -1, s, true);
                                        }

                                        if (port.pinCount > 1)//skusime
                                        {
                                            for (pom = 0; pom < port.pinCount; pom++)
                                            {
                                                s = "";
                                                if (Instances[i].ArchitectureInstance != null)
                                                    s = Instances[i].ArchitectureInstance.opv.GetValueOf(fromName, pom);
                                                if (s.Trim() != "")
                                                {
                                                    foreach (EntityPort p in Instances[i].EntityInstance.Ports)
                                                    {
                                                        s = s.Replace(" " + p.Name + " ", " " + Instances[i].InstanceName + "." + p.Name + " ");
                                                        for (int x = 0; x < p.pinCount; x++)
                                                            s = s.Replace(" " + p.Name + "(" + x.ToString() + ") ", " " + Instances[i].InstanceName + "." + p.Name + "(" + x.ToString() + ") ");
                                                    }
                                                    //signal upravime
                                                    if (Instances[i].ArchitectureInstance != null)
                                                        foreach (Signal ssig in Instances[i].ArchitectureInstance.Signals)
                                                        {
                                                            s = s.Replace(" " + ssig.Name + " ", " " + Instances[i].InstanceName + "." + ssig.Name + " ");
                                                            for (int x = 0; x < ssig.pinCount; x++)
                                                                s = s.Replace(" " + ssig.Name + "(" + x.ToString() + ") ", " " + Instances[i].InstanceName + "." + ssig.Name + "(" + x.ToString() + ") ");
                                                        }
                                                    //
                                                    if (s == Instances[i].ArchitectureInstance.opv.GetValueOf(fromName, pom))
                                                        s = " " + Instances[i].InstanceName + "." + s.Trim() + " ";

                                                    portvalues.Add(Instances[i].InstanceName + "." + fromName, pom, s, true);
                                                }
                                            }
                                        }
                                    }
                                    break;
                                }
                            }
                            if (fromPort == null) throw new Exception("The port with the name \"" + fromName + "\" was not found");
                        }
                        //ak je priradenie polohou
                        else
                        {
                            toName = pomtree.GetChild(ii).GetChild(0).ToMyString();
                            toRangeName = VHDLToXML.ExtractRangeName(ref toName);
                            fromPort = Instances[i].EntityInstance.Ports[ii];

                            int pom = -1;
                            if (fromPort.Direction != portDirection.IN)
                            {
                                /* K O M E N T !
                                if (toRangeName != "")
                                    pom = int.Parse(toRangeName.Replace("(", "").Replace(")", ""));
                                 */
                                portvalues.Add(toName, pom, " " + Instances[i].InstanceName + "." + fromPort.Name + " ");
                                //portvalues.Add(toName, pom, " " + Instances[i].GetEntityArchitectureName() + "." + fromPort.Name + " ");
                                string s = "";
                                if (Instances[i].ArchitectureInstance != null)
                                    s = Instances[i].ArchitectureInstance.opv.GetValueOf(fromPort.Name, -1);
                                if (s.Trim() != "")
                                {
                                    foreach (EntityPort p in Instances[i].EntityInstance.Ports)
                                    {
                                        s = s.Replace(" " + p.Name + " ", " " + Instances[i].InstanceName + "." + p.Name + " ");
                                        for (int x = 0; x < p.pinCount; x++)
                                            s = s.Replace(" " + p.Name + "(" + x.ToString() + ") ", " " + Instances[i].InstanceName + "." + p.Name + "(" + x.ToString() + ") ");
                                    }
                                    //signal upravime
                                    if (Instances[i].ArchitectureInstance != null)
                                        foreach (Signal ssig in Instances[i].ArchitectureInstance.Signals)
                                        {
                                            s = s.Replace(" " + ssig.Name + " ", " " + Instances[i].InstanceName + "." + ssig.Name + " ");
                                            for (int x = 0; x < ssig.pinCount; x++)
                                                s = s.Replace(" " + ssig.Name + "(" + x.ToString() + ") ", " " + Instances[i].InstanceName + "." + ssig.Name + "(" + x.ToString() + ") ");
                                        }
                                    //
                                    if (s == Instances[i].ArchitectureInstance.opv.GetValueOf(fromPort.Name, -1))
                                        s = " " + Instances[i].InstanceName + "." + s.Trim() + " ";

                                    portvalues.Add(Instances[i].InstanceName + "." + fromPort.Name, -1, s, true);
                                }

                                if (fromPort.pinCount > 1)//skusime
                                {
                                    for (pom = 0; pom < fromPort.pinCount; pom++)
                                    {
                                        s = "";
                                        if (Instances[i].ArchitectureInstance != null)
                                            s = Instances[i].ArchitectureInstance.opv.GetValueOf(fromPort.Name, pom);
                                        if (s.Trim() != "")
                                        {
                                            foreach (EntityPort p in Instances[i].EntityInstance.Ports)
                                            {
                                                s = s.Replace(" " + p.Name + " ", " " + Instances[i].InstanceName + "." + p.Name + " ");
                                                for (int x = 0; x < p.pinCount; x++)
                                                    s = s.Replace(" " + p.Name + "(" + x.ToString() + ") ", " " + Instances[i].InstanceName + "." + p.Name + "(" + x.ToString() + ") ");
                                            }
                                            //signal upravime
                                            if (Instances[i].ArchitectureInstance != null)
                                                foreach (Signal ssig in Instances[i].ArchitectureInstance.Signals)
                                                {
                                                    s = s.Replace(" " + ssig.Name + " ", " " + Instances[i].InstanceName + "." + ssig.Name + " ");
                                                    for (int x = 0; x < ssig.pinCount; x++)
                                                        s = s.Replace(" " + ssig.Name + "(" + x.ToString() + ") ", " " + Instances[i].InstanceName + "." + ssig.Name + "(" + x.ToString() + ") ");
                                                }
                                            //
                                            if (s == Instances[i].ArchitectureInstance.opv.GetValueOf(fromPort.Name, pom))
                                                s = " " + Instances[i].InstanceName + "." + s.Trim() + " ";

                                            portvalues.Add(Instances[i].InstanceName + "." + fromPort.Name, pom, s, true);
                                        }
                                    }
                                }
                            }
                            else
                            {
                                portvalues.Add(Instances[i].InstanceName + "." + fromPort.Name, pom, " " + toName + UpravRangeName(toRangeName) + " ");
                                //portvalues.Add(Instances[i].GetEntityArchitectureName() + "." + fromPort.Name, pom, " " + toName + UpravRangeName(toRangeName) + " ");
                            }
                        }


                        //hladam medzi signalmi
                        for (int x = 0; x < Signals.Count; x++)
                        {
                            //ak sa rovnaju mena signalov
                            if (String.Compare(Signals[x].Name, toName, true) == 0)
                            {
                                sigindex = x;
                                //ak je viacbitovi signal tak, ak nie je tak aj tak je RangeName=="" 
                                if (String.Compare(Signals[x].RangeName, toRangeName, true) == 0)
                                {
                                    sig = Signals[x];
                                    break;
                                }
                            }
                        }

                        //ak bol najdeny signal, avsak konkretne pre jeho range value este nebola vytvorena
                        //napr. existuje sig, sig(1), ale sig(2) neexistuje  kde (2) je rangename tak vytvorim novy signal s range (2)
                        //Pozn.: ak vobec neexistuje, tak sigindex==-1 a to znamena, ze je to bud port entity danej architektury alebo globalny signal resp. konstanta a to je osetrene dalej
                        if ((sigindex > -1) && (sig == null))
                        {
                            sig = new Signal();
                            sig.Name = Signals[sigindex].Name;
                            sig.RangeName = toRangeName;
                            sig.Type = Signals[sigindex].Type;
                            Signals.Add(sig);
                        }

                        if (sig != null)
                        {
                            //idem pre kazdu polozku portmapy
                            if (fromPort.Direction == portDirection.IN)
                            {
                                sig.To.Add(new PortReference(Instances[i], EntityType.Instance, fromPort, fromRangeName));
                            }
                            else
                            {
                                //tu sa riesi problem, ze ak je port INOUT, to ci je in or out viem podla ostatnych portov, ktore vstupuju do signalu
                                //preto inout defalutne priradim do From, avsak ak uz je vo From port s direction OUT, inout ide automaticky to Signal.TO
                                //ak je From prazdne, priradim tam vsetko (okrem portDirection IN). Ak vsak pride port OUT, tak zo Signal.From preradim port do Signal.TO (napr. inout)
                                if (sig.From != null)
                                {
                                    if (sig.From.Port.Direction == portDirection.OUT)
                                    {
                                        sig.To.Add(new PortReference(Instances[i], EntityType.Instance, fromPort, fromRangeName));
                                    }
                                    else
                                    {
                                        sig.To.Add(sig.From);
                                        sig.From = new PortReference(Instances[i], EntityType.Instance, fromPort, fromRangeName);
                                    }
                                }
                                else
                                {
                                    sig.From = new PortReference(Instances[i], EntityType.Instance, fromPort, fromRangeName);
                                }
                            }
                        }
                        //ak port nebol najdeny medzi signalmi, este to moze byt port z entity architektury
                        else
                        {
                            sigindex = -1;
                            //hladam medzi konekciami
                            for (int x = 0; x < Connections.Count; x++)
                            {
                                //ak sa rovnaju mena signalov
                                if (String.Compare(Connections[x].Port.Name, toName, true) == 0)
                                {
                                    sigindex = x;
                                    //ak je viacbitovi signal tak, ak nie je tak aj tak je RangeName=="" 
                                    if (String.Compare(Connections[x].RangeName, toRangeName, true) == 0)
                                    {
                                        conn = Connections[x];
                                        break;
                                    }
                                }
                            }



                            //ak bolo najdene connection, avsak konkretne pre jeho range value fromportu danej konekcie este nebola vytvorena
                            //napr. existuje konekcia pre kon.rangename=="" , kon.rangename=="(1)", ale kon.rangename=="(2)" neexistuje  kde (2) je rangename tak vytvorim konekciu na dany port pre novu range a to (2)
                            //Pozn.: ak vobec neexistuje, tak sigindex==-1 a to znamena, ze je to bud port entity danej architektury alebo globalny signal resp. konstanta a to je osetrene dalej
                            if ((sigindex > -1) && (conn == null))
                            {
                                conn = new Connection(Connections[sigindex].Port);
                                conn.RangeName = toRangeName;
                                Connections.Add(conn);
                            }

                            if (conn != null)
                            {
                                conn.To.Add(new PortReference(Instances[i], EntityType.Instance, fromPort, fromRangeName));
                            }
                        }

                        //signal s danym menom nebol najdeny ani medzi definovanymi signalmi ani medzi portami entity architektury
                        //ak je priradeny nejaky globalny signal (ktory nie je def medzi portami entity ani medzi signalmi)
                        //resp. je pridana konstatna hodnota signalu napr. '1'.
                        if ((sig == null) && (conn == null))
                        {
                            EntityPort port = new EntityPort();
                            port.Name = toName;
                            port.Direction = portDirection.GLOBAL;
                            port.Type = "";
                            Connection con = new Connection(port);
                            Connections.Add(con);
                            con.RangeName = toRangeName;

                            //int pom = -1;
                            //if (fromRangeName != "")
                            //    pom = int.Parse(fromRangeName.Replace("(", "").Replace(")", ""));
                            //portvalues.Add(Instances[i].InstanceName + "." + fromName, pom, " " + toName + UpravRangeName(toRangeName) + " ");

                            if (fromPort.Direction == portDirection.IN)
                            {
                                port.Anchor = EAnchor.Right;
                            }
                            else
                            {
                                port.Anchor = EAnchor.Left;
                            }
                            con.To.Add(new PortReference(Instances[i], EntityType.Instance, fromPort, fromRangeName));
                        }
                    }
                }
            }

            //Priame priradenie signalov spravanim sa: toname(rangename)<=fromName(rangename);
            //T_conditional_signal_assignment
            foreach (ITree tree in ConditionalSignalAssigments)
            {
                //target
                string toName = tree.GetChild(0).ToMyString();
                string toRangeName = VHDLToXML.ExtractRangeName(ref toName);
                //conditional_waveforms->waveform->
                string fromName = tree.GetChild(2).ToMyString();
                string fromRangeName = VHDLToXML.ExtractRangeName(ref fromName);

                Signal sigfrom = null;
                Connection connfrom = null;

                int sigindex = -1;

                int pom = -1;
                if (toRangeName != "")
                    pom = int.Parse(toRangeName.Replace("(", "").Replace(")", ""));
                //portvalues.Add(toName, pom, " " + tree.GetChild(2).ToMyString2() + " ");
                portvalues.Add(toName, pom, " " + tree.GetChild(2).ToMyString2());//nakonci nepridavam medzeru, lebo ToMyString2 ju prida na konci
                /*string s = " " + tree.GetChild(2).ToMyString2() + " ";
                foreach (EntityPort p in EntityReference.Ports)
                {
                    s = s.Replace(" " + p.Name + " ", " " + GetEntityArchitectureName() + "." + p.Name + " ");
                }
                portvalues.Add(GetEntityArchitectureName() + "." + toName, pom, s);*/
                /////////////////////////////
                Signal sigto = null;
                Connection connto = null;

                sigindex = -1;

                //hladam medzi signalmi
                for (int x = 0; x < Signals.Count; x++)
                {
                    //ak sa rovnaju mena signalov
                    if (String.Compare(Signals[x].Name, toName, true) == 0)
                    {
                        sigindex = x;
                        //ak je viacbitovi signal tak, ak nie je tak aj tak je RangeName=="" 
                        if (String.Compare(Signals[x].RangeName, toRangeName, true) == 0)
                        {
                            sigto = Signals[x];
                            break;
                        }
                    }
                }

                //ak bol najdeny signal, avsak konkretne pre jeho range value este nebola vytvorena
                //napr. existuje sig, sig(1), ale sig(2) neexistuje  kde (2) je rangename tak vytvorim novy signal s range (2)
                //Pozn.: ak vobec neexistuje, tak sigindex==-1 a to znamena, ze je to bud port entity danej architektury alebo globalny signal resp. konstanta a to je osetrene dalej
                if ((sigindex > -1) && (sigto == null))
                {
                    sigto = new Signal();
                    sigto.Name = Signals[sigindex].Name;
                    sigto.RangeName = toRangeName;
                    sigto.Type = Signals[sigindex].Type;
                    Signals.Add(sigto);
                }

                //ak port nebol najdeny medzi signalmi, este to moze byt port z entity architektury
                if (sigto == null)
                {
                    sigindex = -1;
                    //hladam medzi konekciami
                    for (int x = 0; x < Connections.Count; x++)
                    {
                        //ak sa rovnaju mena signalov
                        if (String.Compare(Connections[x].Port.Name, toName, true) == 0)
                        {
                            sigindex = x;
                            //ak je viacbitovi signal tak, ak nie je tak aj tak je RangeName=="" 
                            if (String.Compare(Connections[x].RangeName, toRangeName, true) == 0)
                            {
                                connto = Connections[x];
                                if (toRangeName == "")
                                    if (IsStructure)
                                        connto.Port.setValue(tree.GetChild(2).ToMyString2(), 0);
                                    else
                                        connto.Port.setBValue(tree.GetChild(2).ToMyString2(), 0);
                                else
                                    if (IsStructure)
                                        connto.Port.setValue(tree.GetChild(2).ToMyString2(), int.Parse(toRangeName.Replace("(", "").Replace(")", "")));
                                    else
                                        connto.Port.setBValue(tree.GetChild(2).ToMyString2(), int.Parse(toRangeName.Replace("(", "").Replace(")", "")));
                                break;
                            }
                        }
                    }


                    //ak bolo najdene connection, avsak konkretne pre jeho range value fromportu danej konekcie este nebola vytvorena
                    //napr. existuje konekcia pre kon.rangename=="" , kon.rangename=="(1)", ale kon.rangename=="(2)" neexistuje  kde (2) je rangename tak vytvorim konekciu na dany port pre novu range a to (2)
                    //Pozn.: ak vobec neexistuje, tak sigindex==-1 a to znamena, ze je to bud port entity danej architektury alebo globalny signal resp. konstanta a to je osetrene dalej
                    if ((sigindex > -1) && (connto == null))
                    {
                        connto = new Connection(Connections[sigindex].Port);
                        connto.RangeName = toRangeName;
                        if (toRangeName != "")
                            if (IsStructure)
                                connto.Port.setValue(tree.GetChild(2).ToMyString2(), int.Parse(toRangeName.Replace("(", "").Replace(")", "")));
                            else
                                connto.Port.setBValue(tree.GetChild(2).ToMyString2(), int.Parse(toRangeName.Replace("(", "").Replace(")", "")));
                        // Connections.Add(connto);
                    }
                }

                //signal s danym menom nebol najdeny ani medzi definovanymi signalmi ani medzi portami entity architektury
                //ak je priradeny nejaky globalny signal (ktory nie je def medzi portami entity ani medzi signalmi)
                //resp. je pridana konstatna hodnota signalu napr. '1'.
                if ((sigto == null) && (connto == null))
                {
                    EntityPort port = new EntityPort();
                    port.Name = toName;
                    port.Direction = portDirection.GLOBAL;
                    port.Type = "";
                    connto = new Connection(port);
                    connto.RangeName = toRangeName;
                    //if (toRangeName != "")
                    //    connto.Port.setValue(tree.GetChild(2).ToMyString2(), int.Parse(toRangeName.Replace("(", "").Replace(")", "")));
                    //Connections.Add(connto);
                }

                ///////////////////////////
                sigindex = -1;
                //hladam medzi signalmi
                for (int x = 0; x < Signals.Count; x++)
                {
                    //ak sa rovnaju mena signalov
                    if (String.Compare(Signals[x].Name, fromName, true) == 0)
                    {
                        sigindex = x;
                        //ak je viacbitovi signal tak, ak nie je tak aj tak je RangeName=="" 
                        if (String.Compare(Signals[x].RangeName, fromRangeName, true) == 0)
                        {
                            sigfrom = Signals[x];
                            break;
                        }
                    }
                }

                //ak bol najdeny signal, avsak konkretne pre jeho range value este nebola vytvorena
                //napr. existuje sig, sig(1), ale sig(2) neexistuje  kde (2) je rangename tak vytvorim novy signal s range (2)
                //Pozn.: ak vobec neexistuje, tak sigindex==-1 a to znamena, ze je to bud port entity danej architektury alebo globalny signal resp. konstanta a to je osetrene dalej
                if ((sigindex > -1) && (sigfrom == null))
                {
                    sigfrom = new Signal();
                    sigfrom.Name = Signals[sigindex].Name;
                    sigfrom.RangeName = fromRangeName;
                    sigfrom.Type = Signals[sigindex].Type;
                    Signals.Add(sigfrom);
                }

                //ak port nebol najdeny medzi signalmi, este to moze byt port z entity architektury
                if (sigfrom == null)
                {
                    sigindex = -1;
                    //hladam medzi konekciami
                    for (int x = 0; x < Connections.Count; x++)
                    {
                        //ak sa rovnaju mena signalov
                        if (String.Compare(Connections[x].Port.Name, fromName, true) == 0)
                        {
                            sigindex = x;
                            //ak je viacbitovi signal tak, ak nie je tak aj tak je RangeName=="" 
                            if (String.Compare(Connections[x].RangeName, fromRangeName, true) == 0)
                            {
                                connfrom = Connections[x];
                                break;
                            }
                        }
                    }


                    //ak bolo najdene connection, avsak konkretne pre jeho range value fromportu danej konekcie este nebola vytvorena
                    //napr. existuje konekcia pre kon.rangename=="" , kon.rangename=="(1)", ale kon.rangename=="(2)" neexistuje  kde (2) je rangename tak vytvorim konekciu na dany port pre novu range a to (2)
                    //Pozn.: ak vobec neexistuje, tak sigindex==-1 a to znamena, ze je to bud port entity danej architektury alebo globalny signal resp. konstanta a to je osetrene dalej
                    if ((sigindex > -1) && (connfrom == null))
                    {
                        connfrom = new Connection(Connections[sigindex].Port);
                        connfrom.RangeName = fromRangeName;
                        Connections.Add(connfrom);
                    }
                }

                //signal s danym menom nebol najdeny ani medzi definovanymi signalmi ani medzi portami entity architektury
                //ak je priradeny nejaky globalny signal (ktory nie je def medzi portami entity ani medzi signalmi)
                //resp. je pridana konstatna hodnota signalu napr. '1'.
                if ((sigfrom == null) && (connfrom == null))
                {
                    EntityPort port = new EntityPort();
                    port.Name = fromName;
                    port.Direction = portDirection.GLOBAL;
                    port.Type = "";
                    connfrom = new Connection(port);
                    connfrom.RangeName = fromRangeName;

                    if (sigto==null) Connections.Add(connfrom);
                }


                ///////////////////////////


                //ak je port entity architektury <= portentityarcnitektury
                if ((connfrom != null) && (connto != null))
                {
                    connfrom.To.Add(new PortReference(null, EntityType.Entity, connto.Port, connto.RangeName));
                    //<=porentityarchitektury  typu globalnysignal or konstanta, nastavi jeho anchor
                    if (connto.Port.Direction == portDirection.GLOBAL)
                    {
                        if (connfrom.Port.Direction == portDirection.IN)
                        {
                            connto.Port.Anchor = EAnchor.Left;
                        }
                        else
                        {
                            connto.Port.Anchor = EAnchor.Right;
                        }
                    }
                    else if (connfrom.Port.Direction == portDirection.GLOBAL)
                    {
                        if (connto.Port.Direction == portDirection.IN)
                        {
                            connfrom.Port.Anchor = EAnchor.Left;
                        }
                        else
                        {
                            connfrom.Port.Anchor = EAnchor.Right;
                        }
                    }
                }
                //ak je poftentityarchitektury<=signal
                else if ((sigfrom != null) && (connto != null))
                {
                    //idem pre kazdu polozku portmapy
                    if (connto.Port.Direction == portDirection.OUT)
                    {
                        sigfrom.To.Add(new PortReference(null, EntityType.Entity, connto.Port, connto.RangeName));
                    }
                    else
                    {
                        //tu sa riesi problem, ze ak je port INOUT, to ci je in or out viem podla ostatnych portov, ktore vstupuju do signalu
                        //preto inout defalutne priradim do From, avsak ak uz je vo From port s direction OUT, inout ide automaticky to Signal.TO
                        //ak je From prazdne, priradim tam vsetko (okrem portDirection IN). Ak vsak pride port OUT, tak zo Signal.From preradim port do Signal.TO (napr. inout)
                        if (sigfrom.From != null)
                        {
                            if (sigfrom.From.Port.Direction == portDirection.IN)
                            {
                                sigfrom.To.Add(new PortReference(null, EntityType.Entity, connto.Port, connto.RangeName));
                            }
                            else
                            {
                                sigfrom.To.Add(sigfrom.From);
                                sigfrom.From = new PortReference(null, EntityType.Entity, connto.Port, connto.RangeName);
                            }
                        }
                        else
                        {
                            sigfrom.From = new PortReference(null, EntityType.Entity, connto.Port, connto.RangeName);
                        }
                    }
                }
                //ak je signal<=poftentityarchitektury
                else if ((sigto != null) && (connfrom != null))
                {
                    //idem pre kazdu polozku portmapy
                    if (connfrom.Port.Direction == portDirection.OUT)
                    {
                        sigto.To.Add(new PortReference(null, EntityType.Entity, connfrom.Port, connfrom.RangeName));
                    }
                    else
                    {
                        //tu sa riesi problem, ze ak je port INOUT, to ci je in or out viem podla ostatnych portov, ktore vstupuju do signalu
                        //preto inout defalutne priradim do From, avsak ak uz je vo From port s direction OUT, inout ide automaticky to Signal.TO
                        //ak je From prazdne, priradim tam vsetko (okrem portDirection IN). Ak vsak pride port OUT, tak zo Signal.From preradim port do Signal.TO (napr. inout)
                        if (sigto.From != null)
                        {
                            if (sigto.From.Port.Direction == portDirection.IN)
                            {
                                sigto.To.Add(new PortReference(null, EntityType.Entity, connfrom.Port, connfrom.RangeName));
                            }
                            else
                            {
                                sigto.To.Add(sigfrom.From);
                                sigto.From = new PortReference(null, EntityType.Entity, connfrom.Port, connfrom.RangeName);
                            }
                        }
                        else
                        {
                            sigto.From = new PortReference(null, EntityType.Entity, connfrom.Port, connfrom.RangeName);
                        }
                    }
                }
                //ak je signal<=signal
                else if ((sigto != null) && (sigfrom != null))
                {

                }
            }

            //correction of signals (if smth=>sig(0) a potom smth2=>sig)
            foreach (Signal sig in Signals)
            {
                if (String.IsNullOrEmpty(sig.RangeName))
                {
                    foreach (Signal sig2 in Signals)
                    {
                        if ((sig.Name == sig2.Name) && (!String.IsNullOrEmpty(sig2.RangeName)))
                        {
                            //TODO: upravit portreference rangename
                            if (sig2.From != null)
                            {
                                sig2.To.AddRange(sig.To);
                                //TODO: mozno bude pridat aj toto a opravit
                                //if (sig.From != null)
                                //{
                                //    if (sig.From.Port.Direction != portDirection.OUT)
                                //    {
                                //        sig2.To.Add(sig.From);
                                //    }
                                //}
                            }
                            else
                            {
                                //moze byt aj null
                                sig2.From = sig.From;
                                //TODO: mozno bude treba pridat aj toto a opravit
                                //foreach (PortReference portref in sig.To)
                                //{
                                //    if (portref.Port.Direction != portDirection.IN)
                                //    {
                                //        sig2.To.Add(portref);
                                //    }
                                //}
                            }
                        }
                    }
                }
            }

            //Signals.Sort();

            //Uloz("D:/actual.assigments");
            //UlozSignals("D:/actual.signals");
            //prirad();
            //Uloz("D:/actual.assigments2");
        }

        /// <summary>
        /// Najde odkazy na architekturu a entitu a priradi danej instancii podla mena entity a mena architektury v instancii.
        /// </summary>
        /// <param name="instance"></param>
        void FindReferencesForInstance(Entity_Component_Instance instance)
        {
            //zistim, ci nie je viac rovnakych entit v navrhu - potom neviem jednoznacne najst entitu podla mena
            for (int i = 0; i < EntitiesInDesign.Count; i++)
            {
                for (int j = 0; j < EntitiesInDesign.Count; j++)
                    if (i != j && String.Compare(EntitiesInDesign[i].FullName, EntitiesInDesign[j].FullName, true) == 0)
                    {
                        throw new Exception("Some entities in design have a same name: " + EntitiesInDesign[i].FullName);
                    }
            }

            //najdem instanciu entity
            for (int i = 0; i < EntitiesInDesign.Count; i++)
            {
                if (String.Compare(EntitiesInDesign[i].FullName, instance.EntityFullName, true) == 0)
                {
                    //TODO: pridat library name, package, name porovnanie
                    instance.EntityInstance = EntitiesInDesign[i];
                    break;
                }
            }

            //ak nie je specifikovana architektura, najdem prvu moznu v designe
            if (String.IsNullOrEmpty(instance.ArchitectureName))
            {
                for (int i = 0; i < ArchitecturesInDesign.Count; i++)
                {
                    //ak je architektura pre danu entitu
                    if (String.Compare(ArchitecturesInDesign[i].EntityFullName, instance.EntityFullName, true) == 0)
                    {
                        //TODO: porovnanie package, libraryname,..
                        instance.ArchitectureInstance = ArchitecturesInDesign[i];
                        instance.ArchitectureName = ArchitecturesInDesign[i].ArchitectureName;
                        instance.ArchitectureLibraryName = ArchitecturesInDesign[i].ArchitectureLibraryName;
                        break;
                    }
                }
            }
            //najdem instanciu architektury podla jej mena
            else
            {
                for (int i = 0; i < ArchitecturesInDesign.Count; i++)
                {
                    //ak je zhodne meno architektury a aj meno entity dane architektury
                    if ((String.Compare(ArchitecturesInDesign[i].EntityFullName, instance.EntityFullName, true) == 0) &&
                        (String.Compare(ArchitecturesInDesign[i].FullName, instance.ArchitectureFullName, true) == 0))
                    {
                        instance.ArchitectureInstance = ArchitecturesInDesign[i];
                        break;
                    }
                }
            }
        }

        #endregion

        #region DominoExtension

        public Assignments opv
        {
            get { return portvalues; }
        }

        public string GetEntityArchitectureName()
        {
            return this._EntityName + "(" + this._ArchitectureName + ")";
        }

        Assignments portvalues = new Assignments();

        public void prirad()
        {
            /*foreach (EntityPort p in EntityReference.Ports)
            {
                int i = 0;
                for (i = 0; i < portvalues.pr.Count; i++)
                    if (portvalues.pr[i].Nazov() == p.Name)
                        break;
                if (i >= portvalues.pr.Count)
                    portvalues.Add(p.Name, -1, "");
            }*/

            foreach (Assignment a in portvalues.pr)
            {
                if (a.Pin() != -1 && !a.Vystuny())
                    Zmen(a.Nazov() + "(" + a.Pin() + ")", a.Hodnota());
            }
            foreach (Assignment a in portvalues.pr)
            {
                if (a.Pin() == -1 && !a.Vystuny())
                    Zmen(a.Nazov(), a.Hodnota());
            }

            /*foreach (Entity_Component_Instance eci in Instances)
                foreach (Assignment a in eci.ArchitectureInstance.portvalues.pr)
                    ;*/
            /*foreach (Assignment a in portvalues.pr)
                foreach (EntityPort p in EntityReference.Ports)
                {
                    a.h = a.h.Replace(" " + p.Name + " ", " " + GetEntityArchitectureName() + "." + p.Name + " ");
                }*/
        }

        public void Zmen(string s1, string s2)
        {
            foreach (Assignment a in portvalues.pr)
            {
                if (a.Vystuny()) continue;
                if (a.Hodnota().Contains(" " + s1 + " ") && s2 != "")
                {
                    a.h = a.h.Replace(" " + s1 + " ", s2);
                    if (a.Pin() != -1 && !a.Vystuny())
                        Zmen(a.Nazov() + "(" + a.Pin() + ")", a.Hodnota());
                    else if (!a.Vystuny())
                        Zmen(a.Nazov(), a.Hodnota());
                }
            }
        }

        public string UpravRangeName(string s)
        {
            //return s.Replace(" ", "").Replace("(", " ( ").Replace(")", " )");
            return s.Replace(" ", "");
        }

        public void Uloz(string file)
        {
            TextWriter tw = new StreamWriter(file, true);
            foreach (Assignment a in portvalues.pr)
            {
                tw.WriteLine(a.Nazov() + "(" + a.Pin() + ") = " + a.Hodnota() + " // " + a.Vystuny().ToString());
            }
            tw.WriteLine("------------------------------");
            tw.Close();
        }

        public void UlozSignals(string file)
        {
            TextWriter tw = new StreamWriter(file, true);
            foreach (Signal s in Signals)
            {
                tw.Write(s.Name + s.RangeName + "[" + s.pinCount.ToString() + "]" + "\t");
                if (s.From != null)
                    tw.Write(s.From.Port.Name + s.From.RangeName + "\t");
                //foreach (PortReference p in s.To)
                  //  tw.Write(p.Port.Name + p.RangeName + " ");
                tw.Write("\n");
            }
            tw.WriteLine("------------------------------");
            tw.Close();
        }

        #endregion
    }
}
