using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using System.Xml;
using System.Xml.XPath;
using System.Drawing;
using System.Xml.Linq;
using Antlr.Vhdl;
using Antlr;
using Antlr.Runtime;
using Antlr.Runtime.Tree;


namespace VHDL2XML
{
    class Designitem
    {
        public string LibraryName;
        public string VhdlCode;
        public string FileName;
    }

    public class VHDLToXML
    {
        List<Designitem> designItems = new List<Designitem>();

        public void AddDesignItem(string filename, string libname, string vhdlcode)
        {
            Designitem d = new Designitem();
            d.LibraryName = libname;
            d.VhdlCode = vhdlcode;
            d.FileName = filename;
            designItems.Add(d);
        }


        public void RemoveDesignItems()
        {
            designItems.Clear();
        }

        //Tu bude v parsingu vygenerovany strom VHDL kodu.
        CommonTree dftree = null;

        #region properties
        private List<ArchitectureDeclaration> _architectures = new List<ArchitectureDeclaration>();

        /// <summary>
        /// List of architectures defined by structure, that were parsed from the VhdlCode.
        /// </summary>
        public List<ArchitectureDeclaration> Architectures
        {
            get { return _architectures; }
            set { _architectures = value; }
        }

        private List<EntityDeclaration> _entities = new List<EntityDeclaration>();

        /// <summary>
        /// List of entity declarations, that were parsed from the VhdlCode.
        /// </summary>
        public List<EntityDeclaration> Entities
        {
            get { return _entities; }
            set { _entities = value; }
        }


        public void Parse()
        {
            foreach (Designitem di in designItems)
            {
                VHDLCode = di.VhdlCode;

                //upravim vstupny kod aby neobsahoval xxxx'xxxx lebo, gramatika to nezhltne. Ale moze obsahovat xxx'x'ssss
                /*for (int i = 0; i < VHDLCode.Length; i++)
                {
                    if (VHDLCode[i] == '\'')
                    {
                        if (i + 2 < VHDLCode.Length)
                        {
                            if (VHDLCode[i + 2] != '\'')
                            {
                                VHDLCode = VHDLCode.Substring(0, i) + VHDLCode.Substring(i + 1, VHDLCode.Length - i - 1);
                            }
                            else
                            {
                                i += 2;
                            }
                        }
                    }
                }
                if (VHDLCode[VHDLCode.Length - 2] == '\'') VHDLCode = VHDLCode.Substring(0, VHDLCode.Length - 2) + VHDLCode.Substring(VHDLCode.Length - 1, 1);*/
                try
                {
                    //parsujem
                    vamsCSharp3Lexer lex = new vamsCSharp3Lexer(
                                       new ANTLRStringStream(
                                           VHDLCode.ToLower()
                                       )
                                  );
                    
                    CommonTokenStream tokens = new CommonTokenStream(lex);
                    vamsCSharp3Parser parser = new vamsCSharp3Parser(tokens);
                    vamsCSharp3Parser.design_file_return df = parser.design_file();
                    dftree = (CommonTree)df.Tree;
                    RecurzParse(0, di.FileName, di.LibraryName, "", dftree);
                    
                    //ToXMLtree(dftree, "D:/actual.xvhd");////////////////
                }
                catch (Exception eee)
                {
                    throw new Exception("Could not parse file: " + di.FileName + " " +  eee.Message);
                }
            }

            //Postparsing
            foreach (ArchitectureDeclaration architecture in Architectures)
            {
                architecture.PostParse(Architectures, Entities);
            }
        }

        /// <summary>
        /// Rekurzivne prechadzam cez strom a co potrebujem, sparsujem v samostatnej funkcii.
        /// Ak zistim pocas rekurzie, ze som v nejakej library resp. package, ... tak nastavim prislusny parameter funkcie.
        /// </summary>
        private void RecurzParse(int hlbka, string file, string library, string package, ITree tree)
        {
            //TODO: dorobit aj package, library, configuration, signals ...

            //Ak entita
            if (tree.Type == vamsCSharp3Parser.T_entity_declaration)
            {
                Entities.Add(ParseEntity(file, library, package, tree));
            }
            //Ak konfiguracia
            else if (tree.Type == vamsCSharp3Parser.T_configuration_declaration)
            {
                //TODO: deklaracia konfiguracie
            }
            //Ak balicek
            else if (tree.Type == vamsCSharp3Parser.T_package_declaration)
            {
                //TODO: deklaracia balicka
            }
            //Ak architektura
            else if (tree.Type == vamsCSharp3Parser.T_architecture_body)
            {
                ParseArchitecture(file, library, package, tree, 0, null);
            }
            //umelo ubrzdim rekuziu max do hlbky, (hlbsie by uz nemali byt ani architektury ani entity ...)
            else if (hlbka < 6)
            {
                for (int i = 0; i < tree.ChildCount; i++)
                {
                    RecurzParse(hlbka + 1, file, library, package, tree.GetChild(i));
                }
            }
        }

        /// <summary>
        /// Rozparsuje entitu. hlbka-daj 0 (pomocna premenna pri rekurzii)
        /// Ako architecture daj NULL - tie je to pomocna premenna pri rekurzii
        /// /// Ak zistim pocas rekurzie, ze som v nejakej library resp. package, ... tak nastavim prislusny parameter funkcie.
        /// </summary>
        private void ParseArchitecture(string file, string library, string package, ITree tree, int hlbka, ArchitectureDeclaration architecture)
        {
            //ak som na zaciatku architektury (savnem jej meno a pre ktoru entitu je)
            if (tree.Type == vamsCSharp3Parser.T_architecture_body)
            {
                architecture = new ArchitectureDeclaration();
                architecture.ArchitectureName = tree.GetChild(0).ToMyString();
                architecture.ArchitectureLibraryName = library;
                //architecture.ArchitecturePackageName = package;
                string entityName = tree.GetChild(1).ToMyString();
                architecture.EntityName = entityName;
                architecture.EntityLibraryName = VHDLToXML.ParseName(ref entityName, library);
                architecture.ComponentDeclarations = new List<ComponentDeclaration>();
                architecture.Signals = new List<Signal>();
                architecture.ComponentConfigSpecifications = new List<ComponentConfigSpec>();
                Architectures.Add(architecture);
            }//nemoze byt else if, lebo toto je uvodny if v rekurzii

            //ak deklaracia komponentu v architekture
            if (tree.Type == vamsCSharp3Parser.T_component_declaration)
            {
                architecture.ComponentDeclarations.Add(ParserComponent(file, library, package, tree));
            }
            else if (tree.Type == vamsCSharp3Parser.T_signal_declaration)
            {
                architecture.Signals.AddRange(ParseSignalDeclaration(tree));
            }
            else if (tree.Type == vamsCSharp3Parser.T_configuration_specification)
            {
                architecture.ComponentConfigSpecifications.AddRange(ParseComponentConfigSpecification(tree,library));
            }
            else if (tree.Type == vamsCSharp3Parser.T_conditional_signal_assignment)
            {
                architecture.ConditionalSignalAssigments.Add(tree);
            }
            else if (tree.Type == vamsCSharp3Parser.T_component_instantiation_statement)
            {
                //instantiated_unit 
                ITree pomtree = tree.GetChild(1).GetChild(0);
                //instancia entity
                if (pomtree.Type == vamsCSharp3Parser.ENTITY)
                {

                    //je to instancia entity
                    Entity_Component_Instance instance = new Entity_Component_Instance();
                    instance.InstanceName = tree.GetChild(0).ToMyString();
                    instance.IsComponentInstance = false;
                    instance.VHDLPortMap = null;
                    //hodi automaticky je pri priradeni do instance.EntityName vytvorena aj instance.EntityLibraryName, musim maunalne priradit meno library kvoli pracovnej kniznici, ak je napr. prazdna
                    string entityName = tree.GetChild(1).GetChild(1).ToMyString();
                    instance.EntityName = entityName;
                    instance.EntityLibraryName = VHDLToXML.ParseName(ref entityName, library);
                    

                    //je definovana aj architektura pre entitu
                    //ak nie je, meno architektury sa zisti prve najdene neskor pri druhom kole
                    if (tree.GetChild(1).ChildCount > 2)
                    {
                        //hodi automaticky je pri priradeni do instance.ArchitectureName vytvorena aj instance.ArchitectureLibraryName, musim maunalne priradit meno library kvoli pracovnej kniznici, ak je napr. prazdna
                        string architectureName = tree.GetChild(1).GetChild(1).ToMyString();
                        instance.ArchitectureName = architectureName;
                        instance.ArchitectureLibraryName = VHDLToXML.ParseName(ref architectureName, library);
                    }

                    //priradnim prislusny portmap, ktory v postparsingu rozparsujem
                    for (int i = 0; i < tree.ChildCount; i++)
                    {
                        if (tree.GetChild(i).Type == vamsCSharp3Parser.T_port_map_aspect)
                        {
                            instance.VHDLPortMap = tree.GetChild(i);
                        }
                        else if (tree.GetChild(i).Type == vamsCSharp3Parser.T_generic_map_aspect)
                        {
                            //TODO:generic map aspect
                        }
                    }
                    architecture.Instances.Add(instance);
                }
                else if (pomtree.Type == vamsCSharp3Parser.CONFIGURATION)
                {
                    //throw new NotImplementedException();
                }
                //instancia komponentu
                else
                {
                    Entity_Component_Instance instance = new Entity_Component_Instance();
                    instance.InstanceName = tree.GetChild(0).ToMyString();
                    instance.ComponentName = pomtree.Type == vamsCSharp3Parser.COMPONENT ? tree.GetChild(1).GetChild(1).ToMyString() : pomtree.ToMyString();
                    instance.IsComponentInstance = true;
                    instance.VHDLPortMap = null;
                    instance.ComponentInstance = null;

                    //najdem prislusnu deklaraciu komponentu a dam instancii referenciu na nu
                    for (int i = 0; i < architecture.ComponentDeclarations.Count; i++)
                    {
                        if (String.Compare(architecture.ComponentDeclarations[i].Name, instance.ComponentName, true) == 0)
                        {
                            instance.ComponentInstance = architecture.ComponentDeclarations[i];
                            break;
                        }
                    }

                    if (instance.ComponentInstance == null)
                    {
                        throw new Exception("Can not create instance from component with the name: " + instance.ComponentName +". Could not find the declaration of the component. If it is an entity name, please add the word ENTITY in front of the entity name " + instance.ComponentName + ".");
                    }

                    //zistim entityname a architecture name pre instanciu z configuration
                    int others = -1;
                    bool choosed = false;
                    for (int ii = 0; ii < architecture.ComponentConfigSpecifications.Count; ii++)
                    {
                        if (architecture.ComponentConfigSpecifications[ii].ComponentName.ToLower() == instance.ComponentName.ToLower())
                        {
                            if (architecture.ComponentConfigSpecifications[ii].ConfigurationFor == ConfigurationFor.SpecificComponent)
                            {
                                if (instance.InstanceName.ToLower() == architecture.ComponentConfigSpecifications[ii].InstanceName.ToLower())
                                {
                                    instance.EntityName = architecture.ComponentConfigSpecifications[ii].EntityName;
                                    instance.EntityLibraryName = architecture.ComponentConfigSpecifications[ii].EntityLibraryName;
                                    instance.ArchitectureName = architecture.ComponentConfigSpecifications[ii].ArchitectureName;
                                    instance.ArchitectureLibraryName = architecture.ComponentConfigSpecifications[ii].ArchitectureLibraryName;
                                    choosed = true;
                                }
                            }
                            else if (architecture.ComponentConfigSpecifications[ii].ConfigurationFor == ConfigurationFor.ALL)
                            {
                                if (instance.ComponentName.ToLower() == architecture.ComponentConfigSpecifications[ii].ComponentName.ToLower())
                                {
                                    instance.EntityName = architecture.ComponentConfigSpecifications[ii].EntityName;
                                    instance.EntityLibraryName = architecture.ComponentConfigSpecifications[ii].EntityLibraryName;
                                    instance.ArchitectureName = architecture.ComponentConfigSpecifications[ii].ArchitectureName;
                                    instance.ArchitectureLibraryName = architecture.ComponentConfigSpecifications[ii].ArchitectureLibraryName;
                                    choosed = true;
                                }
                            }
                            else if (architecture.ComponentConfigSpecifications[ii].ConfigurationFor == ConfigurationFor.OTHERS)
                            {
                                others = ii;
                            }
                        }
                    }

                    //ak entita a architektura nie je specifikovana medzi configuraciou
                    if (!choosed)
                    {
                        //ak bolo pouzite others pre dany komponent;
                        if (others > -1)
                        {
                            instance.EntityName = architecture.ComponentConfigSpecifications[others].EntityName;
                            instance.EntityLibraryName = architecture.ComponentConfigSpecifications[others].EntityLibraryName;
                            instance.ArchitectureName = architecture.ComponentConfigSpecifications[others].ArchitectureName;
                            instance.ArchitectureLibraryName = architecture.ComponentConfigSpecifications[others].ArchitectureLibraryName;
                        }
                        else
                        {
                            string name = instance.ComponentName;
                            instance.EntityName = name;
                            instance.EntityLibraryName = VHDLToXML.ParseName(ref name, library);
                            //meno architektury sa zisti prve najdene neskor pri druhom kole
                        }
                    }

                    //priradnim prislusny portmap, ktory v postparsingu rozparsujem
                    for (int i = 0; i < tree.ChildCount; i++)
                    {
                        if (tree.GetChild(i).Type == vamsCSharp3Parser.T_port_map_aspect)
                        {
                            instance.VHDLPortMap = tree.GetChild(i);
                        }
                        else if (tree.GetChild(i).Type == vamsCSharp3Parser.T_generic_map_aspect)
                        {
                            //TODO:generic map aspect
                        }
                    }
                    architecture.Instances.Add(instance);
                }
            }
            //umelo ubrzdim rekuziu max do hlbky, (hlbsie by uz nemali byt ani architektury ani entity ...)
            else if (hlbka < 6)
            {
                for (int i = 0; i < tree.ChildCount; i++)
                {
                    ParseArchitecture(file, library, package, tree.GetChild(i), hlbka + 1, architecture);
                }
            }
        }

        /// <summary>
        /// Parses: FOR instanceName : componentname USE ENTITY entityname (architecturename)
        /// FOR instanceName : componentname USE CONFIGURATION name
        /// </summary>
        private List<ComponentConfigSpec> ParseComponentConfigSpecification(ITree tree, string library)
        {
            List<ComponentConfigSpec> configs = new List<ComponentConfigSpec>();
            if (tree.Type == vamsCSharp3Parser.T_configuration_specification)
            {

                string entityName = "";
                string architectureName = "";
                string entityLibraryName = "";
                string architectureLibraryName = "";
                string componentName = "";
                //zistim aku entitu resp. architekturu pouziva ... for a,b,c : adam use entity blabla (architecturename);
                //Gramatika: binding_indication -> entity_aspect 
                ITree pomtree = tree.GetChild(1).GetChild(0);
                if (pomtree.Type == vamsCSharp3Parser.T_entity_aspect)
                {
                    //USE entity
                    if (pomtree.GetChild(0).Type == vamsCSharp3Parser.ENTITY)
                    {
                        entityName = pomtree.GetChild(1).ToMyString();
                        entityLibraryName = VHDLToXML.ParseName(ref entityName,library);
                        //architecture name
                        if (pomtree.ChildCount > 2)
                        {
                            //BUG: neparsuje menoentity(menoarchitektury) resp. parsuje library.menoentity?
                            architectureName = pomtree.GetChild(2).ToMyString();
                            architectureLibraryName = VHDLToXML.ParseName(ref architectureName,library);
                        }
                    }
                    //USE configuration
                    else if (pomtree.GetChild(0).Type == vamsCSharp3Parser.CONFIGURATION)
                    {
                        //TODO: vamsCSharp3Parser.CONFIGURATION
                        //throw new NotImplementedException("Configuration neparsujem.");
                    }
                    else
                    {
                        //OPEN - neimplementujem
                    }
                }
                //TODO: else portmapaspect

                //component name  (T_component_specification)
                //for a,b,c : componentname use entity blabla (architecturename);
                componentName = tree.GetChild(0).GetChild(1).ToMyString();

                //pre ktore instancie je tato konfiguracia urcena
                pomtree = tree.GetChild(0).GetChild(0).GetChild(0);
                //for all : componentname use 
                if (pomtree.Type == vamsCSharp3Parser.ALL)
                {
                    ComponentConfigSpec config = new ComponentConfigSpec();
                    config.ConfigurationFor = ConfigurationFor.ALL;
                    config.ArchitectureName = architectureName;
                    config.ArchitectureLibraryName = architectureLibraryName;
                    config.ComponentName = componentName;
                    config.EntityName = entityName;
                    config.EntityLibraryName = entityLibraryName;
                    configs.Add(config);
                }
                else if (pomtree.Type == vamsCSharp3Parser.OTHERS)
                {
                    ComponentConfigSpec config = new ComponentConfigSpec();
                    config.ConfigurationFor = ConfigurationFor.OTHERS;
                    config.ArchitectureName = architectureName;
                    config.ArchitectureLibraryName = architectureLibraryName;
                    config.ComponentName = componentName;
                    config.EntityName = entityName;
                    config.EntityLibraryName = entityLibraryName;
                    configs.Add(config);
                }
                //for others : componentname use ...
                else
                {
                    pomtree = tree.GetChild(0).GetChild(0);
                    for (int i = 0; i < pomtree.ChildCount; i++)
                    {
                        ComponentConfigSpec config = new ComponentConfigSpec();
                        config.ConfigurationFor = ConfigurationFor.SpecificComponent;
                        config.ArchitectureName = architectureName;
                        config.ArchitectureLibraryName = architectureLibraryName;
                        config.ComponentName = componentName;
                        config.EntityName = entityName;
                        config.EntityLibraryName = entityLibraryName;
                        config.InstanceName = pomtree.GetChild(i).ToMyString();
                        configs.Add(config);
                    }
                }
                return configs;
            }
            else return null;
        }


        /// <summary>
        /// nie je rekurzivna, hned tree musi byt vamsCSharp3Parser.T_signal_declaration
        /// </summary>
        private List<Signal> ParseSignalDeclaration(ITree tree)
        {
            List<Signal> signals = new List<Signal>();
            if (tree.Type == vamsCSharp3Parser.T_signal_declaration)
            {
                //subtype_indication
                string type = tree.GetChild(1).ToMyString2();
                int pincount = 0;
                if (tree.GetChild(1).GetChild(0).ChildCount == 4)
                    pincount = 1 + Math.Abs(int.Parse(tree.GetChild(1).GetChild(0).GetChild(2).GetChild(0).ToMyString2()) - int.Parse(tree.GetChild(1).GetChild(0).GetChild(2).GetChild(2).ToMyString2()));
                else
                    pincount = 1;
                string value = "";
                //expression
                if ((tree.ChildCount == 3 && tree.GetChild(2).Type == vamsCSharp3Parser.T_term)
                    || (tree.ChildCount == 4 && tree.GetChild(3).Type == vamsCSharp3Parser.T_term))
                    value = tree.GetChild(2).ToMyString2();
                //identifier_list 
                ITree pomtree = tree.GetChild(0);
                for (int i = 0; i < pomtree.ChildCount; i++)
                {
                    Signal signal = new Signal();
                    signal.Name = pomtree.GetChild(i).ToMyString();
                    signal.Type = type;
                    signal.signalvalue = value.Trim();
                    signal.pinCount = pincount;
                    signals.Add(signal);
                }
                return signals;
            }
            else return signals;
        }

        /// <summary>
        /// nie je rekurzivna, hned tree musi byt vamsCSharp3Parser.T_component_declaration
        /// </summary>
        private ComponentDeclaration ParserComponent(string file, string library, string package, ITree tree)
        {
            if (tree.Type == vamsCSharp3Parser.T_component_declaration)
            {
                ComponentDeclaration component = new ComponentDeclaration();
                component.LibraryName = library;
                component.PackageName = package;
                component.Name = tree.GetChild(0).ToMyString();
                component.Ports = new List<EntityPort>();

                //pridam porty komponentu
                for (int i = 0; i < tree.ChildCount; i++)
                {
                    if (tree.GetChild(i).Type == vamsCSharp3Parser.T_port_clause)
                    {
                        component.Ports.AddRange(ParseInterfaceSignal(tree.GetChild(i)));
                    }
                    else if (tree.Type == vamsCSharp3Parser.T_generic_clause)
                    {
                        //TODO: generic
                        //throw new NotImplementedException();
                    }
                }

                return component;
            }
            else return null;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="tree"></param>
        /// <param name="recurzports">daj null - sluzi pre rekurziu</param>
        /// <returns></returns>
        private List<EntityPort> ParseInterfaceSignal(ITree tree)
        {
            List<EntityPort> ports = new List<EntityPort>();
            if (tree.Type == vamsCSharp3Parser.T_interface_signal_declaration)
            {
                portDirection portDirect = portDirection.IN;
                string portType = "unknown";
                int pincount = 0;
                //zistim typ a mod daneho signalu(ov)
                for (int i = 0; i < tree.ChildCount; i++)
                {
                    if (tree.GetChild(i).Type == vamsCSharp3Parser.T_subtype_indication)
                    {
                        portType = tree.GetChild(i).ToMyString2();
                        if (tree.GetChild(i).GetChild(0).ChildCount == 4)
                            pincount = 1 + Math.Abs(int.Parse(tree.GetChild(i).GetChild(0).GetChild(2).GetChild(0).ToMyString2()) - int.Parse(tree.GetChild(i).GetChild(0).GetChild(2).GetChild(2).ToMyString2()));
                        else
                            pincount = 1;
                    }
                    //mode
                    else if (tree.GetChild(i).Type == vamsCSharp3Parser.T_mode)
                    {
                        int type = tree.GetChild(i).GetChild(0).Type;
                        if (type == vamsCSharp3Parser.IN)
                        {
                            portDirect = portDirection.IN;
                        }
                        else if (type == vamsCSharp3Parser.OUT)
                        {
                            portDirect = portDirection.OUT;
                        }
                        else if (type == vamsCSharp3Parser.INOUT)
                        {
                            portDirect = portDirection.INOUT;
                        }
                        else if (type == vamsCSharp3Parser.BUFFER)
                        {
                            portDirect = portDirection.BUFFER;
                        }
                        else if (type == vamsCSharp3Parser.LINKAGE)
                        {
                            portDirect = portDirection.LINKAGE;
                        }
                    }
                }
                //z listu signalov vytvorim instancie
                for (int i = 0; i < tree.ChildCount; i++)
                {
                    if (tree.GetChild(i).Type == vamsCSharp3Parser.T_identifier_list)
                    {
                        ITree pomtree = tree.GetChild(i);
                        for (int n = 0; n < pomtree.ChildCount; n++)
                        {
                            EntityPort port = new EntityPort();
                            port.Name = pomtree.GetChild(n).ToMyString();
                            port.Direction = portDirect;
                            port.Type = portType;
                            port.pinCount = pincount;
                            ports.Add(port);
                        }
                        break;
                    }
                }
            }
            else
            {
                for (int i = 0; i < tree.ChildCount; i++)
                {
                    ports.AddRange(ParseInterfaceSignal(tree.GetChild(i)));
                }
            }
            return ports;
        }



        /// <summary>
        /// Rozparsuje entitu - iba jej porty
        /// Nie je rekurzina, tree musi byt typu vamsCSharp3Parser.T_entity_declaration.
        /// </summary>
        private EntityDeclaration ParseEntity(string file, string library, string package, ITree tree)
        {
            if (tree.Type == vamsCSharp3Parser.T_entity_declaration)
            {
                EntityDeclaration entity = new EntityDeclaration();
                entity.Name = tree.GetChild(0).ToMyString();
                entity.LibraryName = library;
                entity.PackageName = package;
                entity.Ports = new List<EntityPort>();
                //T_entity_header
                ITree pomtree = tree.GetChild(1);
                for (int i = 0; i < pomtree.ChildCount; i++)
                {
                    if (pomtree.GetChild(i).Type == vamsCSharp3Parser.T_port_clause)
                    {
                        entity.Ports.AddRange(ParseInterfaceSignal(pomtree.GetChild(i)));
                    }
                    else if (pomtree.GetChild(i).Type == vamsCSharp3Parser.T_generic_clause)
                    {
                        //TODO: naimplementovat
                    }
                }
                return entity;
            }
            else return null;
        }


        private string _VHDLCode;
        /// <summary>
        /// VHDL kod, ktory sa ma transformovat. Obsahuje zdruzeny kod zo vsetkych otvorenych suborov
        /// </summary>
        public string VHDLCode
        {
            get
            {
                return _VHDLCode;
            }
            set
            {
                _VHDLCode = value;
            }
        }
        #endregion

        #region publicFunctions
        List<XElement> modules = new List<XElement>();
        List<XElement> ports = new List<XElement>();
        List<XElement> signals = new List<XElement>();
        /// <summary>
        /// Generuje XML strukturu z designu.
        /// </summary>
        //TODO: pridat aj typ portov do xml a vseobecne do vizualizacie (teraz to bolo, ze typ bol v ramci mena portu, avsak kedze budem pracovat aj s vektormi, treba zmenit system)
        //TODO: pridat library, package do xml
        public XDocument GenerateXML()
        {
            //List<XElement> entities = new List<XElement>();
            ////generovanie XML pre entity v navrhu
            //for (int i = 0; i < this.Entities.Count; i++)
            //{
            //    List<XElement> ports = new List<XElement>();
            //    //pre kazdu entity urobim zoznam jej portov
            //    for (int ii=0; ii<Entities[i].Ports.Count;ii++)
            //    {
            //        ports.Add(
            //            new XElement("Port",
            //                new XAttribute("Name",Entities[i].Ports[ii].Name),
            //                new XAttribute("Direction",Entities[i].Ports[ii].Direction.ToMyString()),
            //                new XAttribute("Type",Entities[i].Ports[ii].Type)
            //            )
            //        );
            //    }

            //    //ku kazdej entite vytvorim xml + pridam ku kazdej zoznam jej portov
            //    entities.Add(
            //          new XElement(
            //              "Entity",
            //              new XAttribute("Name",Entities[i].Name),
            //              new XAttribute("Package",Entities[i].PackageName),
            //              new XAttribute("Library",Entities[i].LibraryName),
            //              new XElement("Ports",ports.ToArray())
            //          )
            //    );
            //}

            List<XElement> archs = new List<XElement>();
            //generovanie XML pre architektury v navrhu
            for (int i = 0; i < this.Architectures.Count; i++)
            {
                if (Architectures[i].IsStructure)
                {
                    //VHDLPorts - porty entity - samostatne porty
                    List<XElement> vhdlPorts = new List<XElement>();
                    if (Architectures[i].EntityReference != null)
                    {
                        //pomocne graficke premmenne umiestnenia portov 
                        int yleft = 30;
                        int yright = 30;
                        int yposition;
                        int xposition;

                        for (int p = 0; p < Architectures[i].EntityReference.Ports.Count; p++)
                        {
                            //kedze su to samostatne porty - ak je in tak musi byt anchor doprava, ak je out, tak dam anchor dolava
                            portDirection tempDirection = Architectures[i].EntityReference.Ports[p].Direction;
                            EAnchor tempAnchor = Architectures[i].EntityReference.Ports[p].Anchor;
                            if (Architectures[i].EntityReference.Ports[p].Direction == portDirection.IN)
                            {
                                tempDirection = portDirection.IN;
                                tempAnchor = EAnchor.Right;
                            }
                            else if (Architectures[i].EntityReference.Ports[p].Direction == portDirection.OUT)
                            {
                                tempDirection = portDirection.OUT;
                                tempAnchor = EAnchor.Left;
                            }
                            else if (Architectures[i].EntityReference.Ports[p].Direction == portDirection.BUFFER)
                            {
                                tempAnchor = EAnchor.Left;
                            }
                            else if (Architectures[i].EntityReference.Ports[p].Direction == portDirection.INOUT)
                            {
                                tempAnchor = EAnchor.Left;
                            }
                            else if (Architectures[i].EntityReference.Ports[p].Direction == portDirection.LINKAGE)
                            {
                                tempAnchor = EAnchor.Left;
                            }

                            if (Architectures[i].EntityReference.Ports[p].Anchor == EAnchor.Left)
                            {
                                yleft += 20;
                                yposition = yleft;
                                xposition = 50;
                            }
                            else
                            {
                                yright += 20;
                                yposition = yright;
                                xposition = Architectures[i].MaxLeft;
                            }
                            XElement newport = new XElement("Port",
                                new XAttribute("anchor", tempAnchor.ToMyString()),
                                new XAttribute("x", xposition.ToString()),
                                new XAttribute("y", yposition.ToString()),
                                new XAttribute("width", "20"),
                                new XAttribute("height", "10"),
                                new XElement(
                                    "PortName", this.Architectures[i].EntityReference.Ports[p].Name
                                ),
                                new XElement(
                                    "PortMode", tempDirection.ToMyString()
                                ),
                                new XElement(
                                    "PortType", this.Architectures[i].EntityReference.Ports[p].Type
                                ),
                                new XElement(
                                    "PortPinCount", this.Architectures[i].EntityReference.Ports[p].pinCount
                                )
                            );
                            //          if (this.Architectures[i].EntityReference.Ports[p].Direction != portDirection.IN)
                            if (this.Architectures[i].EntityReference.Ports[p].pinCount > 1)
                            for (int j = 0; j < this.Architectures[i].EntityReference.Ports[p].pinCount; j++)
                                //newport.Add(new XElement("PortValue", new XAttribute("pin", j), this.Architectures[i].EntityReference.Ports[p].getValue(j)));
                                if (Architectures[i].opv.GetValueOf(this.Architectures[i].EntityReference.Ports[p].Name, j) != "")
                                    newport.Add(new XElement("PortValue", new XAttribute("pin", j), Architectures[i].opv.GetValueOf(this.Architectures[i].EntityReference.Ports[p].Name, j)));
                                else if (Architectures[i].opv.GetValueOf(this.Architectures[i].EntityReference.Ports[p].Name, -1) != "" && this.Architectures[i].EntityReference.Ports[p].Direction == portDirection.OUT)
                                {
                                    string buffer = Architectures[i].opv.GetValueOf(this.Architectures[i].EntityReference.Ports[p].Name, -1).Trim();
                                    if (buffer.Split(' ').Length == 1)
                                        newport.Add(new XElement("PortValue", new XAttribute("pin", j), " " + buffer + "(" + j.ToString() + ") "));
                                }
                            if (Architectures[i].opv.GetValueOf(this.Architectures[i].EntityReference.Ports[p].Name, -1) != "")
                                newport.Add(new XElement("PortValue", Architectures[i].opv.GetValueOf(this.Architectures[i].EntityReference.Ports[p].Name, -1)));
                            else if (this.Architectures[i].EntityReference.Ports[p].pinCount > 1 && this.Architectures[i].EntityReference.Ports[p].Direction == portDirection.OUT)
                            {
                                string buf = "";
                                string portname = this.Architectures[i].EntityReference.Ports[p].Name;
                                for (int pin = 0; pin < this.Architectures[i].EntityReference.Ports[p].pinCount; pin++)
                                {
                                    if (this.Architectures[i].EntityReference.Ports[p].Type.Contains("downto"))
                                        buf += portname + "(" + (this.Architectures[i].EntityReference.Ports[p].pinCount - pin -1).ToString() + ")";
                                    else
                                        buf += portname + "(" + pin.ToString() + ")";
                                    if (pin < this.Architectures[i].EntityReference.Ports[p].pinCount -1)
                                        buf += " & ";
                                }
                                newport.Add(new XElement("PortValue", " " + buf + " ")) ;
                            }
                            vhdlPorts.Add(newport);
                        }

                        //globalne signaly a konstantne hodnoty (napr. '1', '0',globalSignal)
                        foreach (Connection conn in Architectures[i].Connections)
                        {
                            if (conn.Port.Direction == portDirection.GLOBAL)
                            {
                                if (conn.Port.Anchor == EAnchor.Right)
                                {
                                    yleft += 20;
                                    yposition = yleft;
                                    xposition = 140;
                                }
                                else
                                {
                                    yright += 20;
                                    yposition = yright;
                                    xposition = Architectures[i].MaxLeft;
                                }

                                XElement newport = new XElement(
                                        "Port",
                                        new XAttribute("anchor", conn.Port.Anchor.ToMyString()),
                                        new XAttribute("x", xposition.ToString()),
                                        new XAttribute("y", yposition.ToString()),
                                        new XAttribute("width", "20"),
                                        new XAttribute("height", "10"),
                                        new XElement(
                                            "PortName", conn.Port.Name
                                        ),
                                        new XElement(
                                            "PortMode", conn.Port.Direction.ToMyString()
                                        )
                                    );
                                newport.Add(new XElement("PortValue", " " + conn.Port.Name.Replace("\'","").Replace("\"","").Trim() + " "));
                                vhdlPorts.Add(newport);
                            }
                        }
                    }

                    //VHDLInstances
                    List<XElement> vhdlInstances = new List<XElement>();
                    for (int p = 0; p < Architectures[i].Instances.Count; p++)
                    {
                        //porty instancie
                        List<XElement> instanceports = new List<XElement>();
                        foreach (EntityPort port in Architectures[i].Instances[p].EntityInstance.Ports)
                        {
                            XElement newport = new XElement("Port",
                                    new XAttribute("anchor", port.Anchor.ToMyString()),
                                    new XAttribute("x", port.Position.X.ToString()),
                                    new XAttribute("y", port.Position.Y.ToString()),
                                    new XAttribute("width", port.Position.Width.ToString()),
                                    new XAttribute("height", port.Position.Height.ToString()),
                                    new XElement("PortName", port.Name),
                                    new XElement("PortMode", port.Direction.ToMyString()),
                                    new XElement("PortType", port.Type),
                                    new XElement("PortPinCount", port.pinCount)
                                );
                            //if (port.Direction == portDirection.IN)
                            {
                                if (port.pinCount > 1)
                                for (int j = 0; j < port.pinCount; j++)
                                    if (Architectures[i].opv.GetValueOf(Architectures[i].Instances[p].InstanceName + "." + port.Name, j) != "")
                                        newport.Add(new XElement("PortValue", new XAttribute("pin", j), Architectures[i].opv.GetValueOf(Architectures[i].Instances[p].InstanceName + "." + port.Name, j)));
                                if (Architectures[i].opv.GetValueOf(Architectures[i].Instances[p].InstanceName + "." + port.Name, -1) != "")
                                    newport.Add(new XElement("PortValue", Architectures[i].opv.GetValueOf(Architectures[i].Instances[p].InstanceName + "." + port.Name, -1)));
                                else if (port.pinCount > 1)
                                {
                                    string buf = "";
                                    string portname = Architectures[i].Instances[p].InstanceName + "." + port.Name;
                                    for (int pin = 0; pin < port.pinCount; pin++)
                                    {
                                        if (port.Type.Contains("downto"))
                                            buf += portname + "(" + (port.pinCount - pin -1).ToString() + ")";
                                        else
                                            buf += portname + "(" + pin.ToString() + ")";
                                        if (pin < port.pinCount -1)
                                            buf += " & ";
                                    }
                                    newport.Add(new XElement("PortValue", " " + buf + " "));
                                }
                            }
                            /*else
                            {
                                for (int j = 0; j < port.pinCount; j++)
                                    if (Architectures[i].Instances[p].ArchitectureInstance != null && Architectures[i].Instances[p].ArchitectureInstance.opv.GetValueOf(port.Name, j) != "")
                                        newport.Add(new XElement("PortValue", new XAttribute("pin", j), Architectures[i].Instances[p].ArchitectureInstance.opv.GetValueOf(port.Name, j)));
                                if (Architectures[i].Instances[p].ArchitectureInstance != null && Architectures[i].Instances[p].ArchitectureInstance.opv.GetValueOf(port.Name, -1) != "")
                                    newport.Add(new XElement("PortValue", Architectures[i].Instances[p].ArchitectureInstance.opv.GetValueOf(port.Name, -1)));
                            }*/
                            instanceports.Add(newport);
                        }

                        vhdlInstances.Add(
                            new XElement(
                                "EntityInstance",
                                 new XAttribute("x", Architectures[i].Instances[p].Position.X.ToString()),
                                 new XAttribute("y", Architectures[i].Instances[p].Position.Y.ToString()),
                                 new XAttribute("width", Architectures[i].Instances[p].Position.Width.ToString()),
                                 new XAttribute("height", Architectures[i].Instances[p].Position.Height.ToString()),
                                 new XElement(
                                     "InstanceName",
                                     Architectures[i].Instances[p].InstanceName
                                 ),
                                 new XElement(
                                     "EntityName",
                                     new XAttribute("PackageName", ""),
                                     new XAttribute("LibraryName", Architectures[i].Instances[p].EntityLibraryName),
                                     Architectures[i].Instances[p].EntityName
                                 ),
                                 new XElement(
                                     "ArchitectureName",
                                     new XAttribute("PackageName", ""),
                                     new XAttribute("LibraryName", Architectures[i].Instances[p].ArchitectureLibraryName),
                                     Architectures[i].Instances[p].ArchitectureName
                                 ),
                                 new XElement(
                                     "VHDLPorts", instanceports.ToArray()
                                 )
                            )
                        );
                    }

                    //VHDLConnections
                    List<XElement> vhdlConnections = new List<XElement>();
                    //pre kazdy port entity danej architektury
                    foreach (Connection connection in Architectures[i].Connections)
                    {
                        if (connection.Port.Direction == portDirection.IN)
                        {
                            foreach (PortReference portref in connection.To)
                            {
                                //TODO: possible bug - portRef nemusi byt typu entitydeclaration, ak 
                                string tempInstanceName = "";
                                if (portref.ObjectReference != null)
                                {
                                    tempInstanceName = ((Entity_Component_Instance)portref.ObjectReference).InstanceName;
                                }
                                vhdlConnections.Add(
                                    new XElement("VHDLConnection",
                                    //new XAttribute("Name", VHDLToXML.AssignConcateName(connection.RangeName, portref.RangeName, "")),
                                        new XAttribute("Name", ""),
                                        new XElement("PortFrom",
                                            new XAttribute("InstanceEntityName", ""),
                                            new XAttribute("portName", connection.Port.Name),
                                            new XAttribute("portRangeName", connection.RangeName)
                                        ),
                                        new XElement("PortTo",
                                            new XAttribute("InstanceEntityName", tempInstanceName),
                                            new XAttribute("portName", portref.Port.Name),
                                            new XAttribute("portRangeName", portref.RangeName)
                                        )
                                    )
                                );
                            }
                        }
                        else
                        {
                            foreach (PortReference portref in connection.To)
                            {
                                //TODO: possible bug - portRef nemusi byt typu blockdeclaration - resp. moze byt null?? ak nebola najdena prislusna entita resp. komponent?? A pozriet aj portref.ObjectREference
                                string tempInstanceName = "";
                                if (portref.ObjectReference != null)
                                {
                                    tempInstanceName = ((Entity_Component_Instance)portref.ObjectReference).InstanceName;
                                }
                                vhdlConnections.Add(
                                    new XElement("VHDLConnection",
                                    //new XAttribute("Name", VHDLToXML.AssignConcateName(portref.RangeName, connection.RangeName, "")),
                                        new XAttribute("Name", ""),
                                        new XElement("PortFrom",
                                            new XAttribute("InstanceEntityName", tempInstanceName),
                                            new XAttribute("portName", portref.Port.Name),
                                            new XAttribute("portRangeName", portref.RangeName)
                                        ),
                                        new XElement("PortTo",
                                            new XAttribute("InstanceEntityName", ""),
                                            new XAttribute("portName", connection.Port.Name),
                                            new XAttribute("portRangeName", connection.RangeName)
                                        )
                                    )
                                );
                            }
                        }
                    }
                    
                    //VHDLSignals
                    List<XElement> vhdlSignals = new List<XElement>();
                    //pridame prepojenia, ale aj signaly pre potreby simulacie
                    foreach (Signal signal in Architectures[i].Signals)
                    {
                        //signal.from == null vtedy, ak je mu napr. v testbenchi priradzovany signal opisom spravania sa menosignalu<= 5ns,2ns,..
                        if (signal.From != null)
                        {
                            foreach (PortReference portref in signal.To)
                            {
                                string tempToInstanceName = "";
                                if (portref.ObjectReference != null)
                                {
                                    tempToInstanceName = ((Entity_Component_Instance)portref.ObjectReference).InstanceName;
                                }
                                string tempFromInstanceName = "";
                                if (signal.From.ObjectReference != null)
                                {
                                    tempFromInstanceName = ((Entity_Component_Instance)signal.From.ObjectReference).InstanceName;
                                }
                                vhdlConnections.Add(
                                    new XElement("VHDLConnection",
                                        new XAttribute("Name", signal.Name),//VHDLToXML.AssignConcateName(signal.From.RangeName, signal.Name + signal.RangeName, portref.RangeName)),
                                        new XElement("PortFrom",
                                            new XAttribute("InstanceEntityName", tempFromInstanceName),
                                            new XAttribute("portName", signal.From.Port.Name),
                                            new XAttribute("portRangeName", signal.From.RangeName)
                                        ),
                                        new XElement("PortTo",
                                            new XAttribute("InstanceEntityName", tempToInstanceName),
                                            new XAttribute("portName", portref.Port.Name),
                                            new XAttribute("portRangeName", portref.RangeName)
                                        )
                                    )
                                );
                            }
                        }
                        
                        //pridame signal do XML
                        if (signal.pinCount == 0) continue;
                        XElement newsignal = new XElement("Signal",
                                new XElement("Name", signal.Name),
                                new XElement("Type", signal.Type),
                                new XElement("Range", signal.pinCount),
                                new XElement("InitialValue", signal.signalvalue));
                        if (signal.pinCount > 1)
                        for (int j = 0; j < signal.pinCount; j++)
                            if (Architectures[i].opv.GetValueOf(signal.Name, j) != "")
                                newsignal.Add(new XElement("SignalValue", new XAttribute("pin", j), Architectures[i].opv.GetValueOf(signal.Name, j)));
                        if (Architectures[i].opv.GetValueOf(signal.Name, -1) != "")
                            newsignal.Add(new XElement("SignalValue", Architectures[i].opv.GetValueOf(signal.Name, -1)));
                        else if (signal.pinCount > 1)
                        {
                            string buf = "";
                            string portname = signal.Name;
                            for (int pin = 0; pin < signal.pinCount; pin++)
                            {
                                if (signal.Type.Contains("downto"))
                                    buf += portname + "(" + (signal.pinCount - pin - 1).ToString() + ")";
                                else
                                    buf += portname + "(" + pin.ToString() + ")";
                                if (pin < signal.pinCount - 1)
                                    buf += " & ";
                            }
                            newsignal.Add(new XElement("SignalValue", " " + buf + " "));
                        }

                        vhdlSignals.Add(newsignal);
                    }

                    archs.Add(

                            new XElement(
                                "Architecture",
                                    new XAttribute("IsStructure", this.Architectures[i].IsStructure.ToString()),
                                    new XAttribute("ArchitectureName", this.Architectures[i].ArchitectureName),
                                    new XAttribute("ArchitectureLibraryName", this.Architectures[i].ArchitectureLibraryName),
                        //new XAttribute("ArchitecturePackageName", this.Architectures[i].ArchitecturePackageName),
                                    new XAttribute("ArchitecturePackageName", ""),
                                    new XAttribute("EntityName", this.Architectures[i].EntityName),
                                    new XAttribute("EntityLibraryName", this.Architectures[i].EntityLibraryName),
                        //new XAttribute("EntityPackageName", this.Architectures[i].EntityPackageName),
                                    new XAttribute("EntityPackageName", ""),
                                    new XElement("VHDLPorts", vhdlPorts.ToArray()),
                                    new XElement("VHDLInstances", vhdlInstances.ToArray()),
                                    new XElement("VHDLConnections", vhdlConnections.ToArray()),
                                    new XElement("VHDLSignals", vhdlSignals.ToArray())
                            )

                    );

                }
                else//architektura spravania
                {
                    //VHDLPorts - porty entity - samostatne porty
                    List<XElement> vhdlPorts = new List<XElement>();
                    if (Architectures[i].EntityReference != null)
                    {
                        for (int p = 0; p < Architectures[i].EntityReference.Ports.Count; p++)
                        {
                            XElement newport = new XElement("Port",
                                new XElement(
                                    "PortName", this.Architectures[i].EntityReference.Ports[p].Name
                                ),
                                new XElement(
                                    "PortMode", this.Architectures[i].EntityReference.Ports[p].Direction.ToMyString()
                                ),
                                new XElement(
                                    "PortType", this.Architectures[i].EntityReference.Ports[p].Type
                                ),
                                new XElement(
                                    "PortPinCount", this.Architectures[i].EntityReference.Ports[p].pinCount
                                )
                            );
                            if (this.Architectures[i].EntityReference.Ports[p].pinCount > 1)
                            for (int j = 0; j < this.Architectures[i].EntityReference.Ports[p].pinCount; j++)
                                if (Architectures[i].opv.GetValueOf(this.Architectures[i].EntityReference.Ports[p].Name, j) != "")
                                    newport.Add(new XElement("PortValue", new XAttribute("pin", j), Architectures[i].opv.GetValueOf(this.Architectures[i].EntityReference.Ports[p].Name, j)));
                            if (Architectures[i].opv.GetValueOf(this.Architectures[i].EntityReference.Ports[p].Name, -1) != "")
                                newport.Add(new XElement("PortValue", Architectures[i].opv.GetValueOf(this.Architectures[i].EntityReference.Ports[p].Name, -1)));
                            else if (this.Architectures[i].EntityReference.Ports[p].pinCount > 1 && this.Architectures[i].EntityReference.Ports[p].Direction == portDirection.OUT)
                            {
                                string buf = "";
                                string portname = this.Architectures[i].EntityReference.Ports[p].Name;
                                for (int pin = 0; pin < this.Architectures[i].EntityReference.Ports[p].pinCount; pin++)
                                {
                                    if (this.Architectures[i].EntityReference.Ports[p].Type.Contains("downto"))
                                        buf += portname + "(" + (this.Architectures[i].EntityReference.Ports[p].pinCount - pin - 1).ToString() + ")";
                                    else
                                        buf += portname + "(" + pin.ToString() + ")";
                                    if (pin < this.Architectures[i].EntityReference.Ports[p].pinCount - 1)
                                        buf += " & ";
                                }
                                newport.Add(new XElement("PortValue", " " + buf + " "));
                            }
                            vhdlPorts.Add(newport);
                        }
                    }

                    //VHDLSignals
                    List<XElement> vhdlSignals = new List<XElement>();
                    foreach (Signal signal in Architectures[i].Signals)
                    {
                        if (signal.pinCount == 0) continue;
                        XElement newsignal = new XElement("Signal",
                                new XElement("Name", signal.Name),
                                new XElement("Type", signal.Type),
                                new XElement("Range", signal.pinCount),
                                new XElement("InitialValue", signal.signalvalue));
                        if (signal.pinCount > 1)
                        for (int j = 0; j < signal.pinCount; j++)
                            if (Architectures[i].opv.GetValueOf(signal.Name, j) != "")
                                newsignal.Add(new XElement("SignalValue", new XAttribute("pin", j), Architectures[i].opv.GetValueOf(signal.Name, j)));
                        if (Architectures[i].opv.GetValueOf(signal.Name, -1) != "")
                            newsignal.Add(new XElement("SignalValue", Architectures[i].opv.GetValueOf(signal.Name, -1)));
                        else if (signal.pinCount > 1)
                        {
                            string buf = "";
                            string portname = signal.Name;
                            for (int pin = 0; pin < signal.pinCount; pin++)
                            {
                                if (signal.Type.Contains("downto"))
                                    buf += portname + "(" + (signal.pinCount - pin - 1).ToString() + ")";
                                else
                                    buf += portname + "(" + pin.ToString() + ")";
                                if (pin < signal.pinCount - 1)
                                    buf += " & ";
                            }
                            newsignal.Add(new XElement("SignalValue", " " + buf + " "));
                        }

                        vhdlSignals.Add(newsignal);
                    }

                    XElement pozicia = new XElement("Position",
                                            new XAttribute("x", -1),
                                            new XAttribute("y", -1),
                                            new XAttribute("width", -1),
                                            new XAttribute("height", -1));
                    //pridam architekturu
                    archs.Add(new XElement("Architecture",
                                    new XAttribute("IsStructure", this.Architectures[i].IsStructure.ToString()),
                                    new XAttribute("ArchitectureName", this.Architectures[i].ArchitectureName),
                                    new XAttribute("ArchitectureLibraryName", this.Architectures[i].ArchitectureLibraryName),
                        //new XAttribute("ArchitecturePackageName", this.Architectures[i].ArchitecturePackageName),
                                    new XAttribute("ArchitecturePackageName", ""),
                                    new XAttribute("EntityName", this.Architectures[i].EntityName),
                                    new XAttribute("EntityLibraryName", this.Architectures[i].EntityLibraryName),
                        //new XAttribute("EntityPackageName", this.Architectures[i].EntityPackageName),
                                    new XAttribute("EntityPackageName", ""),
                                    pozicia,
                                    new XElement("VHDLPorts", vhdlPorts.ToArray()),
                                    new XElement("VHDLSignals", vhdlSignals.ToArray())
                            ));
                }
            }

            return new XDocument(
                                    new XElement("VHDLProject",
                                         new XElement(
                                             "Architectures",
                                            archs.ToArray()
                                        )
                                    )
                       );
        }

        public XDocument GenerateXML2()
        {
            //urcenie top architektur
            List<ArchitectureDeclaration> TopArch = new List<ArchitectureDeclaration>();
            int pom;
            for (int a = 0; a < Architectures.Count; a++) //pre kazdu architekruru
            {
                pom = 0;
                for (int b = 0; b < Architectures.Count; b++) //pozrie kazdu architekturu
                {
                    if (a != b) // nema zmysel pre tu istu architekturu
                    {
                        if (Architectures[b].IsStructure == true) // ma zmysel iba pre strukturu
                        {
                            for (int c = 0; c < Architectures[b].Instances.Count; c++) //pre kazdy komponent prezeranej architektury
                            {
                                if (Architectures[b].Instances[c].ArchitectureFullName == Architectures[a].FullName)
                                {
                                    pom++;
                                    break;
                                }
                            }
                        }
                    }
                    if (pom > 0)
                        break;
                }
                if (pom == 0)
                    TopArch.Add(Architectures[a]); // v zozname su teraz top architektury
            }
            // komponenty
            XElement map = new XElement("Map");
            for (int i = 0; i < TopArch.Count; i++) // pre kazdu top architekturu
            {
                if (TopArch[i].IsStructure == true) // pre arch struktury
                {
                    for (int k = 0; k < TopArch[i].Instances.Count; k++)
                    {
                        for (int l = 0; l<TopArch[i].Signals.Count;l++)
                        {
                            XElement signal = new XElement("Signal",
                                new XElement("Name", TopArch[i].Signals[l].Name),
                                new XElement("DataType", TopArch[i].Signals[l].Type),
                                new XElement("IsClock")
                            );
                            XElement portmap = new XElement("PortMap");
                            Entity_Component_Instance inst = new Entity_Component_Instance();
                            if (TopArch[i].Signals[l].From != null &&
                                (Entity_Component_Instance)TopArch[i].Signals[l].From.ObjectReference != null)
                            {
                                inst = (Entity_Component_Instance)TopArch[i].Signals[l].From.ObjectReference;
                                map = new XElement("Map",
                                    new XElement("PortName", TopArch[i].Signals[l].From.Port.Name),
                                    new XElement("ModulePath", inst.InstanceName)
                                );
                                portmap.Add(map);
                            }
                            for (int m = 0; m < TopArch[i].Signals[l].To.Count; m++)
                            {
                                if (TopArch[i].Signals[l].To[m] != null &&
                                    (Entity_Component_Instance)TopArch[i].Signals[l].To[m].ObjectReference != null)
                                {
                                    inst = (Entity_Component_Instance)TopArch[i].Signals[l].To[m].ObjectReference;
                                    map = new XElement("Map",
                                        new XElement("PortName", TopArch[i].Signals[l].To[m].Port.Name),
                                        new XElement("ModulePath", inst.InstanceName)
                                    );
                                    portmap.Add(map);
                                }
                            }
                            signal.Add(portmap);
                            signals.Add(signal);
                        }
                        Components(TopArch[i].Instances[k],TopArch[i].Instances[k].InstanceName);
                    }
                }
                else // pre arch spravania
                {
                }
            }

                return new XDocument(
                    new XElement("Root",
                        new XElement("Lang", "vhdl"),
                        new XElement("Modules", modules.ToArray()),
                        new XElement("Ports", ports.ToArray()),
                        new XElement("Signals", signals.ToArray())
                    )
                );
        }

        void Components(Entity_Component_Instance comp, string path)
        {
            //moduly
            XElement module = new XElement("Module",
                new XElement("Name",comp.InstanceName),
                new XElement("NameLang"),
                new XElement("Type",comp.EntityName+"("+comp.ArchitectureName+")"),
                new XElement("Path",path),
                new XElement("PosX1"),
                new XElement("PosY1"),
                new XElement("PosX2"),
                new XElement("PosY2")
            );
            modules.Add(module);

            //porty
            for (int j = 0; j < comp.EntityInstance.Ports.Count; j++)
            {
                XElement port = new XElement("Port",
                    new XElement("Name", comp.EntityInstance.Ports[j].Name),
                    new XElement("Orientation", comp.EntityInstance.Ports[j].Direction.ToString()),
                    new XElement("DataType", comp.EntityInstance.Ports[j].Type),
                    new XElement("Width", comp.EntityInstance.Ports[j].pinCount.ToString()),
                    new XElement("ModulePath", path),
                    new XElement("PosX1"),
                    new XElement("PosY1"),
                    new XElement("PosX2"),
                    new XElement("PosY2"),
                    new XElement("Rotate")
                );
                ports.Add(port);
            }

            if (comp.ArchitectureInstance.IsStructure == true)
            {
                string hlp="";
                //signaly
                for (int k = 0; k < comp.ArchitectureInstance.Signals.Count; k++)
                {
                    XElement signal = new XElement("Signal",
                        new XElement("Name", comp.ArchitectureInstance.Signals[k].Name),
                        new XElement("DataType", comp.ArchitectureInstance.Signals[k].Type),
                        new XElement("IsClock")
                    );
                    XElement portmap = new XElement("PortMap");
                    Entity_Component_Instance inst = new Entity_Component_Instance();
                    if ((Entity_Component_Instance)comp.ArchitectureInstance.Signals[k].From.ObjectReference != null)
                    {
                        inst = (Entity_Component_Instance)comp.ArchitectureInstance.Signals[k].From.ObjectReference;
                        hlp = inst.InstanceName;
                    }
                    XElement map = new XElement("Map",
                        new XElement("PortName", comp.ArchitectureInstance.Signals[k].From.Port.Name),
                        new XElement("ModulePath", path+hlp)
                    );
                    portmap.Add(map);
                    inst = null;
                    map = null;
                    for (int l = 0; l < comp.ArchitectureInstance.Signals[k].To.Count; l++)
                    {
                        hlp = "";
                        if ((Entity_Component_Instance)comp.ArchitectureInstance.Signals[k].To[l].ObjectReference != null)
                        {
                            inst = (Entity_Component_Instance)comp.ArchitectureInstance.Signals[k].To[l].ObjectReference;
                            hlp = "." + inst.InstanceName;
                        }
                        map = new XElement("Map",
                            new XElement("PortName", comp.ArchitectureInstance.Signals[k].To[l].Port.Name),
                            new XElement("ModulePath", path + hlp)
                        );
                        portmap.Add(map);
                    }
                    signal.Add(portmap);
                    signals.Add(signal);
                }

                for (int i = 0; i < comp.ArchitectureInstance.Instances.Count; i++)
                {
                    Components(comp.ArchitectureInstance.Instances[i], path + "." + comp.ArchitectureInstance.Instances[i].InstanceName);
                }
            }
        }
        #endregion

        #region parsingFunctions

        /// <summary>
        /// Rozparsuje meno:  "aaa.bbb.ccc" na a "aaa.bbb" a "ccc" 
        /// </summary>
        internal static string ParseName(ref string name)
        {
            string pomName = name.Trim();
            int index = pomName.LastIndexOf(".");
            if (index > 0)
            {
                name = pomName.Substring(index + 1);
                return pomName.Substring(0, index);
            }
            else
            {
                return "";
            }
        }

        /// <summary>
        /// Rozparsuje meno:  "aaa.bbb.ccc" na a "aaa.bbb" a "ccc" 
        /// returnIfNull sa vrati, ak bol napr. vstup iba "aaa"
        /// </summary>
        internal static string ParseName(ref string name, string returnIfNull)
        {
            string pomName = name.Trim();
            int index = pomName.LastIndexOf(".");
            if (index > 0)
            {
                name = pomName.Substring(index + 1);
                return pomName.Substring(0, index);
            }
            else
            {
                return returnIfNull;
            }
        }

        /// <summary>
        /// Spoji mena do: "nameA.nameB.nameC"  resp. ich kombinacie a oddeli ich bodkov
        /// Use napr:  "library.package.name"
        /// </summary>
        public static string DotConcateName(string nameA, string nameB, string nameC)
        {
            return SignConcateName(nameA, nameB, nameC, ".");
        }

        /// <summary>
        /// Spoji mena do: "nameA=>nameB=>nameC"  resp. ich kombinacie a oddeli ich znakom =>
        /// </summary>
        public static string AssignConcateName(string nameA, string nameB, string nameC)
        {
            return SignConcateName(nameA, nameB, nameC, "=>");
        }

        /// <summary>
        /// if e.g. sign == "."
        /// Spoji mena do: "nameA.nameB.nameC"  resp. ich kombinacie a oddeli ich znakom sign
        /// Use napr:  "library.package.name"
        /// </summary>
        public static string SignConcateName(string nameA, string nameB, string nameC, string sign)
        {
            string result = "";
            if (!String.IsNullOrEmpty(nameA))
            {
                result = nameA;
            }
            if (!String.IsNullOrEmpty(nameB))
            {
                if (!String.IsNullOrEmpty(result))
                {
                    result += sign;
                }
                result += nameB;
            }
            if (!String.IsNullOrEmpty(nameC))
            {
                if (!String.IsNullOrEmpty(result))
                {
                    result += sign;
                }
                result += nameC;
            }
            return result;
        }

        /// <summary>
        /// Devides string "absdef ( x)" to "absdef"  a  (x).
        /// Return "(x)" and update parameter to "absdef"
        /// </summary>
        public static string ExtractRangeName(ref string str)
        {
            string result = "";
            if (String.IsNullOrEmpty(str))
            {
                return "";
            }


            int index = str.IndexOf('(');

            if (index > 0)
            {
                result = str.Substring(index).Replace(" ", "");
                str = str.Substring(0, index).Replace(" ", "");
            }
            else
            {
                str = str.Replace(" ", "");
                return "";
            }
            return result;
        }


        #endregion

        #region graphicFunctionsLocations
        /// <summary>
        /// Graficky rozmiestni instancie v ramci kazdej architektury.
        /// </summary>
        //TODO: urobit aj rozmiestnenie portov a signalov. Teraz su porty rozmiestnene automaticky vizualizacnou triedou (a aj je nastavena vyska instancie automaticky)
        public void PositionDesign()
        {
            //kolko je lavych a pravych portov

            //rozmiestnenie portov v ramci entity
            foreach (EntityDeclaration entity in Entities)
            {
                int cleft = 0;
                int cright = 0;
                foreach (EntityPort port in entity.Ports)
                {
                    //ak je x or y -1, nastavia sa tieto suradnice automaticky graphenginom podla polohy entity. 
                    //width a height su dane typom vykreslovania portu a teda tu nemusia byt nastavene
                    port.Position = new Rectangle(-1, -1, -1, -1);
                    if (port.Direction == portDirection.IN)
                    {
                        port.Anchor = EAnchor.Left;
                        cleft++;
                    }
                    else
                    {
                        port.Anchor = EAnchor.Right;
                        cright++;
                    }
                }
                entity.LeftPortsCount = cleft;
                entity.RightPortsCount = cright;
            }

            foreach (ArchitectureDeclaration architecture in this.Architectures)
            {
                for (int i = 0; i < architecture.Instances.Count; i++)
                {
                    ////TODO: vyratam vysku instancie (podla poctu portov z toho kolko ich je na pravo a kolko na lavo)
                    ////sirku neratam, mam pevnu
                    ////Pozn. Engine teraz robi tak, ze automaticky nastavi vysku instancie podla pridavanych portov
                    //architecture.Instances[i].Position = new Rectangle(
                    //                                          (i * architecture.Instances[i].Position.Width) + ((i + 1) * 100),
                    //                                          (i * architecture.Instances[i].Position.Height) + ((i + 1) * 60),
                    //                                          250,
                    //                                          100
                    //                                     );

                    //odhad vysky instancie
                    architecture.Instances[i].SugestedHeight = Math.Max(architecture.Instances[i].EntityInstance.RightPortsCount, architecture.Instances[i].EntityInstance.LeftPortsCount) * 25;


                    //ak je to instancia komponentu
                    if (architecture.Instances[i].IsComponentInstance)
                    {
                        //TODO: ak portmap priradenim, inac, komponent by asi vobec nemal byt vizualizovany  ale jeho entita. Tym padom tento riadok by sa ma vymazat.
                        architecture.Instances[i].Ports = new List<EntityPort>(architecture.Instances[i].ComponentInstance.Ports);
                    }
                    else
                    {
                        architecture.Instances[i].Ports = new List<EntityPort>(architecture.Instances[i].EntityInstance.Ports);
                    }
                }

                //urcim instancie s indexom 1 (1. stlpec v zobrazeni)
                int top = 50;
                int order = 1;
                int left = 100;

                foreach (Connection con in architecture.Connections)
                {
                    if ((con.Port.Direction == portDirection.IN) ||
                         ((con.Port.Direction == portDirection.GLOBAL) && (con.Port.Anchor == EAnchor.Right)))
                    {
                        foreach (PortReference portref in con.To)
                        {
                            if (portref.Type == EntityType.Instance)
                            {

                                Entity_Component_Instance inst = ((Entity_Component_Instance)portref.ObjectReference);
                                if (inst.PositionIndex < 0)
                                {
                                    inst.PositionIndex = 1;
                                    inst.OrderIndex = order++;
                                    inst.Position = new Rectangle(300, top, 250, inst.SugestedHeight);
                                    top += inst.SugestedHeight + 50;
                                }
                            }
                        }
                        left = 650;
                    }
                }

                //urcim instancie s indexom >1 (dalsie stlpece v zobrazeni)
                bool found = true;
                int index = 1;

                while (found)
                {
                    found = false;
                    top = 50;
                    foreach (Signal signal in architecture.Signals)
                    {
                        if (signal.From != null)
                        {
                            if (signal.From.Type == EntityType.Instance)
                            {
                                if (((Entity_Component_Instance)signal.From.ObjectReference).PositionIndex == index)
                                {
                                    
                                    order = 1;
                                    foreach (PortReference portref in signal.To)
                                    {
                                        if (portref.Type == EntityType.Instance)
                                        {
                                            Entity_Component_Instance inst = ((Entity_Component_Instance)portref.ObjectReference);
                                            if (inst.PositionIndex <=2 )
                                            {
                                                found = true;
                                                inst.PositionIndex = index + 1;
                                                inst.OrderIndex = order++;
                                                if (inst.SugestedHeight < 0) inst.SugestedHeight = 100;
                                                inst.Position = new Rectangle(left, top, 250, inst.SugestedHeight);
                                                top += inst.SugestedHeight + 20;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    if (found)
                    {
                        index++;
                        left += 350 + 40;
                    }
                }

                top = 100;
                found = false;
                foreach (Entity_Component_Instance instance in architecture.Instances)
                {
                    if (instance.PositionIndex < 0)
                    {
                        instance.Position = new Rectangle(left, top, 250, instance.SugestedHeight);
                        top += instance.SugestedHeight + 20;
                        found = true;
                    }
                }

                if (found) left += 390;
                architecture.MaxLeft = left;
            }
        }
        #endregion

        #region DominoExtension

        //ulozenie pomocneho vypisu do suboru - AST vo formate string a XML
        private void ToXMLtree(CommonTree t, String file)
        {
            TextWriter tw = new StreamWriter(file);
            tw.Write(t.ToStringTree());

            XDocument x = new XDocument();
            x.Add(new XElement(t.Text));
            help(x.Element(t.Text), t);
            x.Save(file + ".xml");
            tw.Close();
        }

        //rekurzivna funkcia na prevod do XML
        private void help(XNode xe, ITree t)
        {
            for (int i = 0; i < t.ChildCount; i++)
            {
                XNode n;
                if (t.GetChild(i).Text.StartsWith("T_"))
                    n = new XElement(t.GetChild(i).Text);
                else
                    //n = new XText(t.GetChild(i).Text);
                    n = new XElement("value", t.GetChild(i).Text);
                ((XElement)xe).Add(n);
                help(n, t.GetChild(i));
            }
        }

        //funkcia na vypocet vyrazu pomocou parsera podla urceneho typu
        static public string Calculate(string expr, string type)
        {
            if (expr.Replace(" ", "") == "" || type.Replace(" ", "") == "") return "";
            vamsCSharp3Parser parser = new vamsCSharp3Parser(new CommonTokenStream(new vamsCSharp3Lexer(new ANTLRStringStream(expr.ToLower()))));
            vamsCSharp3Parser.expression_return e = parser.expression();
            return RecurzCalc((ITree)e.Tree, type);
        }

        //rekurzivna funkcia na vypocet vyrazu podla typu
        static private string RecurzCalc(ITree tree, string type)
        {
            int value = 0;
            string buf = "";
            if (tree.ChildCount == 0 && (tree.Text != "(" || tree.Text != ")"))
                return tree.Text;

            if (tree.GetChild(0).Type == vamsCSharp3Parser.T_operator)
                if (tree.GetChild(0).GetChild(0).Type == vamsCSharp3Parser.ABS && type != "bit" && type != "bit_vector" && type != "std_logic" && type != "std_logic_vector")
                    return (Math.Abs(int.Parse(RecurzCalc(tree.GetChild(1), type)))).ToString();
                else if (tree.GetChild(0).GetChild(0).Type == vamsCSharp3Parser.NOT && (type == "bit" || type == "boolean" || type == "std_logic"))
                {
                    buf = RecurzCalc(tree.GetChild(1), type);
                    if (buf == "0" || buf == "false") return "1";
                    else if (buf == "1" || buf == "true") return "0";
                }

            for (int i = 0; i < tree.ChildCount; i++)
            {
                if (tree.GetChild(i).Text == "(" || tree.GetChild(i).Text == ")") continue;
                /*if (tree.GetChild(i).Text == "-" && i+1 < tree.ChildCount)
                {
                    i++;
                    value = -1 * RecurzCalc(tree.GetChild(i), type);
                }*/
                if (tree.GetChild(i).Type != vamsCSharp3Parser.T_operator)
                    buf = RecurzCalc(tree.GetChild(i), type);

                if (tree.GetChild(i).Type == vamsCSharp3Parser.T_operator && i + 1 < tree.ChildCount)
                {
                    int op = tree.GetChild(i).GetChild(0).Type;
                    i++;
                    if (type == "bit" || type == "boolean" || type == "std_logic")
                    {
                        if (buf != "1" && buf != "0" && buf != "true" && buf != "false") return null;
                        switch (op)
                        {
                            case vamsCSharp3Parser.AND: buf = toS(toB(buf) && toB(RecurzCalc(tree.GetChild(i), type))); break;
                            case vamsCSharp3Parser.OR: buf = toS(toB(buf) || toB(RecurzCalc(tree.GetChild(i), type))); break;
                            case vamsCSharp3Parser.NAND: buf = toS(!(toB(buf) && toB(RecurzCalc(tree.GetChild(i), type)))); break;
                            case vamsCSharp3Parser.NOR: buf = toS(!(toB(buf) || toB(RecurzCalc(tree.GetChild(i), type)))); break;
                            case vamsCSharp3Parser.XOR: buf = toS(toB(buf) ^ toB(RecurzCalc(tree.GetChild(i), type))); break;
                            case vamsCSharp3Parser.XNOR: buf = toS(!(toB(buf) ^ toB(RecurzCalc(tree.GetChild(i), type)))); break;
                            case vamsCSharp3Parser.EQ: buf = toS(toB(buf) == toB(RecurzCalc(tree.GetChild(i), type))); break;
                            case vamsCSharp3Parser.NEQ: buf = toS(toB(buf) != toB(RecurzCalc(tree.GetChild(i), type))); break;
                        }
                    }
                    else if (type == "bit_vector" || type == "std_logic_vector")
                        switch (op)
                        {
                            case vamsCSharp3Parser.AMPERSAND: buf += RecurzCalc(tree.GetChild(i), type); break;
                        }
                    /*switch (op)
                    {
                        //todo:pre viacbitove signaly
                        case vamsCSharp3Parser.AND: value = toI(toB(value) && toB(RecurzCalc(tree.GetChild(i), type))); break;
                        case vamsCSharp3Parser.OR: value = toI(toB(value) || toB(RecurzCalc(tree.GetChild(i), type))); break;
                        case vamsCSharp3Parser.NAND: value = toI(!(toB(value) && toB(RecurzCalc(tree.GetChild(i), type)))); break;
                        case vamsCSharp3Parser.NOR: value = toI(!(toB(value) || toB(RecurzCalc(tree.GetChild(i), type)))); break;
                        case vamsCSharp3Parser.XOR: value = toI(toB(value) ^ toB(RecurzCalc(tree.GetChild(i), type))); break;
                        case vamsCSharp3Parser.XNOR: value = toI(!(toB(value) ^ toB(RecurzCalc(tree.GetChild(i), type)))); break;
                        case vamsCSharp3Parser.EQ: value = toI(toB(value) == toB(RecurzCalc(tree.GetChild(i), type))); break;
                        case vamsCSharp3Parser.NEQ: value = toI(toB(value) != toB(RecurzCalc(tree.GetChild(i), type))); break;
                        case vamsCSharp3Parser.LOWERTHAN: value = toI(value < RecurzCalc(tree.GetChild(i), type)); break;
                        case vamsCSharp3Parser.LE: value = toI(value <= RecurzCalc(tree.GetChild(i), type)); break;
                        case vamsCSharp3Parser.GREATERTHAN: value = toI(value > RecurzCalc(tree.GetChild(i), type)); break;
                        case vamsCSharp3Parser.GE: value = toI(value >= RecurzCalc(tree.GetChild(i), type)); break;
                        //todo:shift operators
                        case vamsCSharp3Parser.PLUS: value += RecurzCalc(tree.GetChild(i), type); break;
                        case vamsCSharp3Parser.MINUS: value -= RecurzCalc(tree.GetChild(i), type); break;
                        //case vamsCSharp3Parser.AMPERSAND: value += RecurzCalc(tree.GetChild(i)); break;
                        case vamsCSharp3Parser.MUL: value *= RecurzCalc(tree.GetChild(i), type); break;
                        case vamsCSharp3Parser.DIV: value /= RecurzCalc(tree.GetChild(i), type); break;
                        case vamsCSharp3Parser.MOD: value %= RecurzCalc(tree.GetChild(i), type); break;
                        case vamsCSharp3Parser.REM: value %= RecurzCalc(tree.GetChild(i), type); break;//mod a rem sa trosku lisia vo vhdl v zapornych hodnotach
                        case vamsCSharp3Parser.DOUBLESTAR: value = (int)Math.Pow(value, RecurzCalc(tree.GetChild(i), type)); break;//nemusi byt int -> prerobit
                    }*/
                }
            }
            return buf;
        }

        //prevod stringu na boolean
        static private bool toB(string s)
        {
            if (s == "1" || s == "true")
                return true;
            else
                return false;
        }

        //prevod boolean na string
        static private string toS(bool b)
        {
            if (b) return "1";
            else return "0";
        }

        //prevod boolean na integer
        static private int toI(bool b)
        {
            return (b) ? 1 : 0;
        }

        //prevod waveform z vhdl formy na zoznam retazcov jednotlivych zmien
        static public List<string> waveform(string s)//waveform
        {
            List<string> str = new List<string>();
            if (s.Trim() == "") return null;
            vamsCSharp3Parser parser = new vamsCSharp3Parser(new CommonTokenStream(new vamsCSharp3Lexer(new ANTLRStringStream(s.ToLower()))));
            vamsCSharp3Parser.waveform_return w = parser.waveform();

            for (int i = 0; i < ((ITree)w.Tree).ChildCount; i++)
                if (((ITree)w.Tree).GetChild(i).Text != ",")
                    str.Add(((ITree)w.Tree).GetChild(i).ToMyString2());

            return str;
        }

        #endregion
    }
}
