﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;

namespace VerilogToXML
{
    enum PortType { input, output, inout };
    enum PortAnchor { No, Left, Right, Top, Bottom};

    class Port : PortInterface
    {
        public Port(){}
        public Port(Module mod) // kazdy port musi byt na nejakom module
        {
            _module = mod;
            Position = new Rectangle(-1, -1, 20, 10);
        }

        private PortAnchor _anchor;

        public void SetAnchor(PortAnchor anch)
        {
            _anchor = anch;
        }
        public String GetAnchor()
        {
            switch (_anchor)
            {
                case PortAnchor.No: return "no";
                case PortAnchor.Left: return "left";
                case PortAnchor.Right: return "right";
                case PortAnchor.Top: return "top";
                case PortAnchor.Bottom: return "bottom";
                default: return "";
            }
        }

        public Rectangle Position;
        private PortType _type; // typ portu, oproti VHDL verilog pozna len tri typy: input, output a inout
        public PortType GetPortType()
        {
            return _type;
        }
        public String GetPortXVHDLmode()
        {
            switch (_type)
            {
                case PortType.input: return "IN";
                case PortType.output: return "OUT";
                case PortType.inout: return "INOUT";
                default: return "";
            }
        }
        public String GetPortXVerilogMode()
        {
            switch (_type)
            {
                case PortType.input: return "input";
                case PortType.output: return "output";
                case PortType.inout: return "inout";
                default: return "";
            }
        }
        public void SetPortType(PortType value)
        {
            _type = value;
        }

        private String _name; // meno portu, ktore ho jednoznacne identifikuje v rámci modulu
        public String GetName()
        {
            return _name;
        }
        public void SetName(String value)
        {
            _name = value;
        }
        public String GetModuleInstanceName()
        {
            return "";
        }

        private Module _module;   // kazdy port musi byt na nejakom moduli
        public Module GetModule()
        {
            return _module;
        }

        private int _down=0;
        private int _up=0;
        public int GetRange()
        {
            return Math.Abs(_up-_down)+1 ;
        }
        public String GetPortXVHDLtype()
        {
            if (GetRange() == 1)
                return "bit";
            else
                return "bit_vector(" + _down + "to" + _up + ")";
        }
        public String GetPortXVerilogType()
        {
            if (GetRange() == 1)
                return "bit";
            else
                return "bit_vector(" + _down + "to" + _up + ")";
        }
        public int GetDown()
        {
            return _down;
        }
        public int GetUp()
        {
            return _up;
        }
        public void SetRange(int d, int u)
        {
            _down = d;
            _up = u;
        }
    }
}
