﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using System.Xml.XPath;
using System.Xml.Linq;
using AntlrVerilogParser;
using Antlr;
using Antlr.Runtime;
using Antlr.Runtime.Tree;
using System.Diagnostics;

namespace VerilogToXML
{
    public class Verilog2XML
    {
        public Verilog2XML(String filename)
        {
            _verilogcode = new VerilogCode();
            _verilogcode.AddVerilogFile(filename);
            _modules = new List<Module>();
            _primitives = new List<Primitive>();
        }
        public Verilog2XML()
        {
            _verilogcode = new VerilogCode();
            _modules = new List<Module>();
            _primitives = new List<Primitive>();
        }

        public event EventHandler ProgressChanged;

        private List<Module> _modules;
        private List<Primitive> _primitives;
        private VerilogCode _verilogcode;

        public void AddVerilogFile(String file)
        {
            _verilogcode.AddVerilogFile(file);
        }

        public void Parse()
        {
            try
            {
                _modules.Clear();
                _primitives.Clear();
                if (ProgressChanged != null)
                    ProgressChanged.Invoke(this, new EventArgs());
                VerilogLexer lex = new VerilogLexer(new ANTLRStringStream(_verilogcode.GetVerilogCode()));
                if (ProgressChanged != null)
                    ProgressChanged.Invoke(this, new EventArgs());
                CommonTokenStream tokens = new CommonTokenStream(lex);
                if (ProgressChanged != null)
                    ProgressChanged.Invoke(this, new EventArgs());
                VerilogParser parser = new VerilogParser(tokens);

                CommonTree tree = (CommonTree)parser.source_text().Tree;

                for (int i = 0; i < tree.Children.Count; i++)
                {
                    if (tree.GetChild(i).Type == VerilogParser.T__156) while (tree.GetChild(++i).Type != VerilogParser.T__157) ;
                    if (tree.GetChild(i).Type == VerilogParser.T__158) while (tree.GetChild(++i).Type != VerilogParser.T__159) ;
                    if (tree.GetChild(i).Type == VerilogParser.T__79) while (tree.GetChild(++i).Type != VerilogParser.T__80) ;
                    if (tree.GetChild(i).Type == VerilogParser.T__81) while (tree.GetChild(++i).Type != VerilogParser.T__82) ;
                    if (tree.GetChild(i).Type == VerilogParser.T__83) while (tree.GetChild(++i).Type != VerilogParser.T__84) ;
                    if (tree.GetChild(i).Type == 0)
                        throw new Exception("Error: Unsupported Verilog language construction!\r\n:" + tree.GetChild(i).ToString() + "\r\nText: " + tree.GetChild(i).Text);
                    if (tree.GetChild(i).Type == VerilogParser.T__67)  // T__67 - 'module'
                    {
                        AddModule(tree, tree.GetChild(i).ChildIndex);
                    }
                    if (tree.GetChild(i).Type == VerilogParser.T__73)  // T__73 - 'primitive'
                    {
                        AddPrimitive(tree, tree.GetChild(i).ChildIndex);
                    }
                    if (ProgressChanged != null)
                        ProgressChanged.Invoke(this, new EventArgs());
                }
                // prve parsovanie je za nami, nasleduje druhe,
                // pri tomto parsovani je uz znamy pocet modulov, ich porty, registre, siete nets
                // este je potrebne naparsovat instancie modulov a primitiv.
                foreach (Module module in _modules)
                    ParseModule(module);

            }
            catch (Exception ex)
            {
                throw ex;
            }
        }

        private int AddModule(CommonTree tree, int index) // Vytvori objekt modul a da mu meno, tree je strom
        // Predstavuje prve parsovanie, v ktorom sa naparsuju portlist, netlist, reglist, mozno neskor tu bude doimplementovane parsovanie assign ...
        {
            Module module;
            module = new Module(tree.GetChild(index + 1).Text, index, tree);
            try
            {
                module.ParsePortRegNet(tree, index + 3);
            }
            catch (Exception ex)
            {
                throw ex;
            }
            //module.SetModuleTree(mtree);
            _modules.Add(module);
            return _modules.Count;

        }
        
        private int AddPrimitive(CommonTree tree, int index) // Vytvori objekt primitive a da mu meno, tree je strom
        // Predstavuje prve parsovanie, v ktorom sa naparsuju portlist, netlist, reglist, mozno neskor tu bude doimplementovane parsovanie assign ...
        {
            Primitive primitive;
            primitive = new Primitive(tree.GetChild(index + 1).Text, index, tree);
            try
            {
                primitive.ParsePort(tree, index + 3);
            }
            catch (Exception ex)
            {
                throw ex;
            }
            //primitive.SetModuleTree(ptree);
            _primitives.Add(primitive);
            return _primitives.Count;

        }

        private Module GetModule(String name)    // prehlada moduly v _modules a vrati modul s danym menom
        // v pripade neuspechu vracia null
        {
            foreach (Module mod in _modules)
                if (name == mod.Name)
                    return mod;
            return null;
        }
        
        private Module GetPrimitive(String name)    // prehlada primitivy v _primitivess a vrati primitivu s danym menom
        // v pripade neuspechu vracia null
        {
            foreach (Primitive prim in _primitives)
                if (name == prim.Name)
                    return prim;
            return null;
        }

        private void ParseModule(Module module)
        // Predstavuje druhe parsovanie, v ktorom sa vytvaraju instancie modulov
        {
            CommonTree tree;
            int index;
            tree = module.ModuleTree;
            index = module.TreeIndex;
            if (tree.GetChild(index + 1).Text != module.Name)
                throw new Exception("---- Symstem error! ----"); //  -*-
            for (int i = index + 2; ((ITree)tree.Children[i]).Type != VerilogParser.T__69 && i + 2 < tree.ChildCount; i++)
            {
                if (tree.GetChild(i).Type == VerilogParser.IDENTIFIER
                    && tree.GetChild(i + 1).Type == VerilogParser.IDENTIFIER
                    && tree.GetChild(i + 2).Type == VerilogParser.LPAREN
                    && GetModule(tree.GetChild(i).Text) != null // ak tento modul existuje, tak vrati jeho meno, inak vrati null
                    )
                {
                    module.ParseModuleInstance(tree, i, GetModule(tree.GetChild(i).Text));
                }
                if (tree.GetChild(i).Type == VerilogParser.IDENTIFIER
                    && GetPrimitive(tree.GetChild(i).Text) != null// ak tento modul existuje, tak vrati jeho meno, inak vrati null
                    )
                {
                    module.ParsePrimitiveInstance(tree, i, GetPrimitive(tree.GetChild(i).Text));
                }
            }
        }

        List<ModuleInstance> mi = new List<ModuleInstance>();
        List<string> mipath = new List<string>();


        private List<Module> getTopModules()
        {
            List<Module> mod = new List<Module>();
            bool top = true;

            foreach (Module m1 in _modules)
            {
                top = true;
                foreach (Module m2 in _modules)
                {
                    foreach (ModuleInstance mi in m2.ModuleInstanceList)
                    {
                        if (mi.Module.Name == m1.Name)
                        {
                            top = false;
                            break;
                        }
                    }
                }
                if (top)// && m1.ModuleInstanceList.Count > 0)
                    mod.Add(m1);
            }
            return mod;
        }

        public XDocument GenerateXVerilogXML()
        {
            List<XElement> mods = new List<XElement>();
            List<XElement> ports = new List<XElement>();
            List<XElement> signals = new List<XElement>();

            List<Module> TopModules = getTopModules();


            //pridavanie topmodulov do XML
            foreach (Module m in TopModules)
            {






                int x = 0;
                foreach (Net net in m.NetList)
                {
                    if (net.Connections.Count > 0)
                    {
                        List<XElement> PortMap = new List<XElement>();

                        List<bool> In = new List<bool>();
                        List<bool> Out = new List<bool>();

                        bool b = false;
                        for (int k = 0; k < net.Connections.Count; k++)
                        {
                            if (b)
                            {
                                b = false;
                                for (int c = 0; c < net.Connections[k].NetBitList.Count; c++)
                                {
                                    if (net.Connections[k].NetBitList[c] != net.Connections[k - 1].NetBitList[c])
                                        b = true;
                                }
                            }
                            if (b)
                            {
                                signals.Add(new XElement("Signal",
                            new XElement("Name", net.Name),
                            new XElement("DataType", net.getType()),
                            new XElement("IsClock", "false"),
                            new XElement("PortMap", PortMap.ToArray()),
                            new XElement("Route")));
                                PortMap = new List<XElement>();
                                In = new List<bool>();
                                Out = new List<bool>();
                            }

                            string path = m.Name;

                            if (net.Connections[k].GetModuleInstanceName() != "")
                                path = path + "." + net.Connections[k].GetModuleInstanceName();

                            if (In.Count == 0)
                            {
                                In = net.Connections[k].NetBitList;
                            }
                            else
                            {
                                if (Out.Count == 0)
                                    for (int c = 0; c < net.Connections[c].NetBitList.Count; c++)
                                        Out.Add(false);

                                for (int c = 0; c < net.Connections[k].NetBitList.Count; c++)
                                {
                                    if (Out[c] || net.Connections[k].NetBitList[c])
                                        Out[c] = true;
                                }
                            }

                            PortMap.Add(new XElement("Map",
                                new XElement("Portname", net.Connections[k].Port.GetName()),
                                new XElement("ModulePath", path)));

                            b = false;
                            if (In.Count != 0 && Out.Count != 0)
                            {
                                b = true;
                                for (int c = 0; c < In.Count; c++)
                                {
                                    if (In[c] != Out[c])
                                    {
                                        b = false;
                                    }
                                }
                            }

                        }

                        signals.Add(new XElement("Signal",
                            new XElement("Name", net.Name),
                            new XElement("DataType", net.getType()),
                            new XElement("IsClock", "false"),
                            new XElement("PortMap", PortMap.ToArray()),
                            new XElement("Route")));
                    }
                }

                foreach (Port pi in m.PortList)
                {
                    ports.Add(new XElement("Port",
                            new XElement("Name", pi.GetName()),
                            new XElement("Orientation", pi.GetPortXVerilogMode()),
                            new XElement("DataType", pi.GetPortXVerilogType()),
                            new XElement("Width", pi.GetRange()),
                            new XElement("ModulePath", m.Name),
                            new XElement("PosX1"),
                            new XElement("PosY1"),
                            new XElement("PosX2"),
                            new XElement("PosY2"),
                            new XElement("Rotate", "1")));
                }
                mods.Add(new XElement("Module",
                            new XElement("Name", m.Name),
                            new XElement("NameLang", m.Name),
                            new XElement("Type", m.Name),
                            new XElement("Path", m.Name),
                            new XElement("PosX1"),
                            new XElement("PosY1"),
                            new XElement("PosX2"),
                            new XElement("PosY2")));





                


                AddInstances(m, m.Name);
            }


            //pridavanie instancii modulov do xml
            int j = 0;
            foreach (ModuleInstance i in mi)
            {

                int x = 0;
                foreach (Net net in i.Module.NetList)
                {
                    if (net.Connections.Count > 0)
                    {
                        List<XElement> PortMap = new List<XElement>();

                        List<bool> In = new List<bool>();
                        List<bool> Out = new List<bool>();

                        bool b = false;
                        for (int k = 0; k < net.Connections.Count; k++)
                        {
                            if (b)
                            {
                                b = false;
                                for (int c = 0; c < net.Connections[k].NetBitList.Count; c++)
                                {
                                    if (net.Connections[k].NetBitList[c] != net.Connections[k - 1].NetBitList[c])
                                        b = true;
                                }
                            }
                            if (b)
                            {
                                signals.Add(new XElement("Signal",
                            new XElement("Name", net.Name),
                            new XElement("DataType", net.getType()),
                            new XElement("IsClock", "false"),
                            new XElement("PortMap", PortMap.ToArray()),
                            new XElement("Route")));
                                PortMap = new List<XElement>();
                                In = new List<bool>();
                                Out = new List<bool>();
                            }

                            string path = mipath[j];

                            if (net.Connections[k].GetModuleInstanceName() != "")
                                path = path + "." + net.Connections[k].GetModuleInstanceName();

                            if (In.Count == 0)
                            {
                                In = net.Connections[k].NetBitList;
                            }
                            else
                            {
                                if (Out.Count == 0)
                                    for (int c = 0; c < net.Connections[c].NetBitList.Count; c++)
                                        Out.Add(false);

                                for (int c = 0; c < net.Connections[k].NetBitList.Count; c++)
                                {
                                    if (Out[c] || net.Connections[k].NetBitList[c])
                                        Out[c] = true;
                                }
                            }

                            PortMap.Add(new XElement("Map",
                                new XElement("Portname", net.Connections[k].Port.GetName()),
                                new XElement("ModulePath", path)));

                            b = false;
                            if (In.Count != 0 && Out.Count != 0)
                            {
                                b = true;
                                for (int c = 0; c < In.Count; c++)
                                {
                                    if (In[c] != Out[c])
                                    {
                                        b = false;
                                    }
                                }
                            }

                        }

                        signals.Add(new XElement("Signal",
                            new XElement("Name", net.Name),
                            new XElement("DataType", net.getType()),
                            new XElement("IsClock", "false"),
                            new XElement("PortMap", PortMap.ToArray()),
                            new XElement("Route")));
                    }
                }
                
                foreach (PortInstance pi in i.PortInstanceList)
                {
                    ports.Add(new XElement("Port",
                            new XElement("Name", pi.GetName()),
                            new XElement("Orientation",pi.GetPortXVerilogMode()),
                            new XElement("DataType",pi.GetPortXVerilogType()),
                            new XElement("Width",pi.GetRange()),
                            new XElement("ModulePath", mipath[j]),
                            new XElement("PosX1"),
                            new XElement("PosY1"),
                            new XElement("PosX2"),
                            new XElement("PosY2"),
                            new XElement("Rotate","1")));
                }
                mods.Add(new XElement("Module",
                            new XElement("Name", i.InstanceName),
                            new XElement("NameLang", i.InstanceName),
                            new XElement("Type", i.Module.Name),
                            new XElement("Path", mipath[j]),
                            new XElement("PosX1"),
                            new XElement("PosY1"),
                            new XElement("PosX2"),
                            new XElement("PosY2")));
                j++;
            }
            
            return new XDocument(new XElement("Root", new XElement("Lang", "Verilog"), new XElement("Modules", mods.ToArray()), new XElement("Ports", ports.ToArray()), new XElement("Signals", signals.ToArray())));
        }

        private void AddInstances(Module m, string path)
        {
            string p;
            foreach (ModuleInstance i in m.ModuleInstanceList)
            {
                mi.Add(i);
                p = path + "." + i.InstanceName;
                mipath.Add(p);

                if (i.Module.ModuleInstanceList.Count > 0)
                    AddInstances(i.Module,p);
            }
        }

        public bool CheckSyntax(string file, string iverilog_path)
        {
            Process exeProcess = new Process();
            exeProcess.StartInfo.CreateNoWindow = true;
            exeProcess.StartInfo.UseShellExecute = false;
            exeProcess.StartInfo.FileName = iverilog_path;
            exeProcess.StartInfo.WindowStyle = ProcessWindowStyle.Hidden;
            exeProcess.StartInfo.Arguments = file;
            exeProcess.StartInfo.RedirectStandardError = true;

            exeProcess.Start();
            string str;

            while ((str = exeProcess.StandardError.ReadLine()) != null)
            {
                if (str.Length > 0)
                {
                    return false;
                }
            }
            exeProcess.WaitForExit(5000);

            return true;
        }
    }
}
